
#include <string.h>
#include <stdio.h>

#include "php.h"
#include "zend_extensions.h"
#include "zend_compile.h"
#include "zend_API.h"
#include "zend_ini.h"

#include "xcache.h"
#include "align.h"
#include "const_string.h"
#include "processor.h"
#include "stack.h"
#include "xcache_globals.h"

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
extern unsigned int zend_hash_canary;
#endif













#undef NDEBUG
#include <assert.h>





















































































































































































































































































































































































































































































































#ifndef NDEBUG
#	undef inline
#define inline
#endif

typedef zval *zval_ptr;
typedef zend_uchar zval_data_type;
#ifdef IS_UNICODE
typedef UChar zstr_uchar;
#endif
typedef char  zstr_char;

#define MAX_DUP_STR_LEN 256
/* export: typedef struct _xc_processor_t xc_processor_t; :export {{{ */
struct _xc_processor_t {
	char *p;
	zend_uint size;
	HashTable strings;
	HashTable zvalptrs;
	zend_bool reference; /* enable if to deal with reference */
	zend_bool have_references;
	const xc_entry_t *xce_src;
	const xc_entry_t *xce_dst;
	const zend_class_entry *cache_ce;
	zend_uint cache_class_num;

	const zend_op          *active_opcodes_src;
	zend_op                *active_opcodes_dst;
	const zend_class_entry *active_class_entry_src;
	zend_class_entry       *active_class_entry_dst;
	zend_uint               active_class_num;

	zend_bool readonly_protection; /* wheather it's present */

#ifndef NDEBUG
		xc_stack_t allocsizes;
#endif

};
/* }}} */
#ifdef HAVE_XCACHE_DPRINT
static void xc_dprint_indent(int indent) /* {{{ */
{
	int i;
	for (i = 0; i < indent; i ++) {
		fprintf(stderr, "  ");
	}
}
/* }}} */
static void xc_dprint_str_len(const char *str, int len) /* {{{ */
{
	const unsigned char *p = (const unsigned char *) str;
	int i;
	for (i = 0; i < len; i ++) {
		if (p[i] < 32 || p[i] == 127) {
			fprintf(stderr, "\\%03o", (unsigned int) p[i]);
		}
		else {
			fputc(p[i], stderr);
		}
	}
}
/* }}} */
#endif
/* {{{ xc_zstrlen_char */
static inline int xc_zstrlen_char(zstr s)
{
	return strlen(ZSTR_S(s));
}
/* }}} */
#ifdef IS_UNICODE
/* {{{ xc_zstrlen_uchar */
static inline int xc_zstrlen_uchar(zstr s)
{
	int i;
	UChar *p = ZSTR_U(s);
	for (i = 0; *p; i ++, p++) {
		/* empty */
	}
	return i;
}
/* }}} */
/* {{{ xc_zstrlen */
static inline int xc_zstrlen(int type, zstr s)
{
	return type == IS_UNICODE ? xc_zstrlen_uchar(s) : xc_zstrlen_char(s);
}
/* }}} */
#else
/* {{{ xc_zstrlen */
#define xc_zstrlen(dummy, s) xc_zstrlen_char(s)
/* }}} */
#endif
/* {{{ xc_calc_string_n */
 
static inline void xc_calc_string_n(xc_processor_t *processor, zend_uchar type, zstr str, long size 
#ifndef NDEBUG
		, int relayline
#endif
) {
	
	int realsize = UNISW(size, (type == IS_UNICODE) ? UBYTES(size) : size);
	long dummy = 1;

	if (realsize > MAX_DUP_STR_LEN) {
		
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(char) * realsize));
			xc_stack_push(&processor->allocsizes, (void *) (long) (relayline));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(char) * realsize;
	
	
	
	
	

	}
	else if (zend_u_hash_add(&processor->strings, type, str, size, (void *) &dummy, sizeof(dummy), NULL) == SUCCESS) {
		/* new string */
		
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(char) * realsize));
			xc_stack_push(&processor->allocsizes, (void *) (long) (relayline));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(char) * realsize;
	
	
	
	
	

	} 
	
#ifndef NDEBUG
		
		else {
					}
	
#endif

	
}
/* }}} */
/* {{{ xc_store_string_n */
 
static inline zstr xc_store_string_n(xc_processor_t *processor, zend_uchar type, zstr str, long size 
#ifndef NDEBUG
		, int relayline
#endif
) {
	
	int realsize = UNISW(size, (type == IS_UNICODE) ? UBYTES(size) : size);
	zstr ret, *pret;

	if (realsize > MAX_DUP_STR_LEN) {
		
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `ZSTR_V(ret)',`char',`realsize' at line %d\n", relayline);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(char) * realsize;
				if (expect != real) {
					fprintf(stderr, "mismatch `ZSTR_V(ret)',`char',`realsize' at line %d(was %lu): real %lu - expect %lu = %lu\n", relayline, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		ZSTR_V(ret) = (char *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(ZSTR_V(ret), -1, sizeof(char) * realsize);
#endif

			
		processor->p += sizeof(char) * realsize;

		
	
	
	
	

		memcpy(ZSTR_V(ret), ZSTR_V(str), realsize);
		return ret;
	}

	if (zend_u_hash_find(&processor->strings, type, str, size, (void **) &pret) == SUCCESS) {
		return *pret;
	}

	/* new string */
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `ZSTR_V(ret)',`char',`realsize' at line %d\n", relayline);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(char) * realsize;
				if (expect != real) {
					fprintf(stderr, "mismatch `ZSTR_V(ret)',`char',`realsize' at line %d(was %lu): real %lu - expect %lu = %lu\n", relayline, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		ZSTR_V(ret) = (char *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(ZSTR_V(ret), -1, sizeof(char) * realsize);
#endif

			
		processor->p += sizeof(char) * realsize;

		
	
	
	
	

	memcpy(ZSTR_V(ret), ZSTR_V(str), realsize);
	zend_u_hash_add(&processor->strings, type, str, size, (void *) &ret, sizeof(zstr), NULL);
	return ret;

	
}
/* }}} */
/* {{{ xc_get_class_num
 * return class_index + 1
 */
static zend_ulong xc_get_class_num(xc_processor_t *processor, zend_class_entry *ce) {
	zend_ulong i;
	const xc_entry_t *xce = processor->xce_src;
	zend_class_entry *ceptr;

	if (processor->cache_ce == ce) {
		return processor->cache_class_num;
	}
	for (i = 0; i < xce->data.php->classinfo_cnt; i ++) {
		ceptr = CestToCePtr(xce->data.php->classinfos[i].cest);
		if (ZCEP_REFCOUNT_PTR(ceptr) == ZCEP_REFCOUNT_PTR(ce)) {
			processor->cache_ce = ceptr;
			processor->cache_class_num = i + 1;
			return i + 1;
		}
	}
	assert(0);
	return (zend_ulong) -1;
}
/* }}} */
/* {{{ xc_get_class */
#ifdef ZEND_ENGINE_2
static zend_class_entry *xc_get_class(xc_processor_t *processor, zend_ulong class_num) {
	/* must be parent or currrent class */
	assert(class_num <= processor->active_class_num);
	return CestToCePtr(processor->xce_dst->data.php->classinfos[class_num - 1].cest);
}
#endif
/* }}} */
#ifdef ZEND_ENGINE_2
/* fix method on store */
static void xc_fix_method(xc_processor_t *processor, zend_op_array *dst) /* {{{ */
{
	zend_function *zf = (zend_function *) dst;
	zend_class_entry *ce = processor->active_class_entry_dst;

	/* Fixing up the default functions for objects here since
	 * we need to compare with the newly allocated functions
	 *
	 * caveat: a sub-class method can have the same name as the
	 * parent~s constructor and create problems.
	 */

	if (zf->common.fn_flags & ZEND_ACC_CTOR) {
		if (!ce->constructor) {
			ce->constructor = zf;
		}
	}
	else if (zf->common.fn_flags & ZEND_ACC_DTOR) {
		ce->destructor = zf;
	}
	else if (zf->common.fn_flags & ZEND_ACC_CLONE) {
		ce->clone = zf;
	}
	else {
#define SET_IF_SAME_NAME(member) \
		do { \
			if (!strcasecmp(ZSTR_S(zf->common.function_name), #member)) { \
				ce->member = zf; \
			} \
		} \
		while(0)
		/* if(ce->member && !strcmp(zf->common.function_name, ce->member->common.function_name)) { \ */

		SET_IF_SAME_NAME(__get);
		SET_IF_SAME_NAME(__set);
#ifdef ZEND_ENGINE_2_1
		SET_IF_SAME_NAME(__unset);
		SET_IF_SAME_NAME(__isset);
#endif
		SET_IF_SAME_NAME(__call);
#ifdef ZEND_CALLSTATIC_FUNC_NAME
		SET_IF_SAME_NAME(__callstatic);
#endif
#if defined(ZEND_ENGINE_2_2) || PHP_MAJOR_VERSION >= 6
		SET_IF_SAME_NAME(__tostring);
#endif

#undef SET_IF_SAME_NAME
	}
}
/* }}} */
#endif
/* {{{ call op_array ctor handler */
extern zend_bool xc_have_op_array_ctor;
static void xc_zend_extension_op_array_ctor_handler(zend_extension *extension, zend_op_array *op_array TSRMLS_DC)
{
	if (extension->op_array_ctor) {
		extension->op_array_ctor(op_array);
	}
}
/* }}} */
/* export: xc_entry_t *xc_processor_store_xc_entry_t(xc_entry_t *src TSRMLS_DC); :export {{{ */
xc_entry_t *xc_processor_store_xc_entry_t(xc_entry_t *src TSRMLS_DC) {
	xc_entry_t *dst;
	xc_processor_t processor;

	memset(&processor, 0, sizeof(processor));
	processor.reference = 1;

	
#ifndef NDEBUG
		xc_stack_init(&processor.allocsizes);
#endif


	/* calc size */ {
		zend_hash_init(&processor.strings, 0, NULL, NULL, 0);
		if (processor.reference) {
			zend_hash_init(&processor.zvalptrs, 0, NULL, NULL, 0);
		}

		processor.size = 0;
		/* allocate */
		processor.size = ALIGN(processor.size + sizeof(src[0]));

		xc_calc_xc_entry_t(&processor, src TSRMLS_CC);
		if (processor.reference) {
			zend_hash_destroy(&processor.zvalptrs);
		}
		zend_hash_destroy(&processor.strings);
	}
	src->size = processor.size;
	src->have_references = processor.have_references;

	
#ifndef NDEBUG
		xc_stack_reverse(&processor.allocsizes);
#endif

	/* store {{{ */
	{
		
#ifndef NDEBUG
		char *oldp;
#endif

		zend_hash_init(&processor.strings, 0, NULL, NULL, 0);
		if (processor.reference) {
			zend_hash_init(&processor.zvalptrs, 0, NULL, NULL, 0);
		}

		/* mem :) */
		processor.p = (char *) src->cache->mem->handlers->malloc(src->cache->mem, processor.size);
		if (processor.p == NULL) {
			dst = NULL;
			goto err_alloc;
		}
		
#ifndef NDEBUG
		oldp = processor.p;
#endif

		assert(processor.p == (char *) ALIGN(processor.p));

		/* allocate */
		dst = (xc_entry_t *) processor.p;
		processor.p = (char *) ALIGN(processor.p + sizeof(dst[0]));

		xc_store_xc_entry_t(&processor, dst, src TSRMLS_CC);
		
#ifndef NDEBUG
		 {
			int real = processor.p - oldp;
			int should = processor.size;
			if (real != processor.size) {
				fprintf(stderr, "real %d - should %d = %d\n", real, should, real - should);
				abort();
			}
		}
#endif

err_alloc:
		if (processor.reference) {
			zend_hash_destroy(&processor.zvalptrs);
		}
		zend_hash_destroy(&processor.strings);
	}
	/* }}} */

	
#ifndef NDEBUG
		xc_stack_destroy(&processor.allocsizes);
#endif


	return dst;
}
/* }}} */
/* export: xc_entry_t *xc_processor_restore_xc_entry_t(xc_entry_t *dst, const xc_entry_t *src, zend_bool readonly_protection TSRMLS_DC); :export {{{ */
xc_entry_t *xc_processor_restore_xc_entry_t(xc_entry_t *dst, const xc_entry_t *src, zend_bool readonly_protection TSRMLS_DC) {
	xc_processor_t processor;

	memset(&processor, 0, sizeof(processor));
	processor.readonly_protection = readonly_protection;
	if (src->have_references) {
		processor.reference = 1;
	}

	if (processor.reference) {
		zend_hash_init(&processor.zvalptrs, 0, NULL, NULL, 0);
	}
	xc_restore_xc_entry_t(&processor, dst, src TSRMLS_CC);
	if (processor.reference) {
		zend_hash_destroy(&processor.zvalptrs);
	}
	return dst;
}
/* }}} */
/* export: zval *xc_processor_restore_zval(zval *dst, const zval *src, zend_bool have_references TSRMLS_DC); :export {{{ */
zval *xc_processor_restore_zval(zval *dst, const zval *src, zend_bool have_references TSRMLS_DC) {
	xc_processor_t processor;

	memset(&processor, 0, sizeof(processor));
	processor.reference = have_references;

	if (processor.reference) {
		zend_hash_init(&processor.zvalptrs, 0, NULL, NULL, 0);
				zend_hash_add(&processor.zvalptrs, (char *)src, sizeof(src), (void*)&dst, sizeof(dst), NULL);
	}
	xc_restore_zval(&processor, dst, src TSRMLS_CC);
	if (processor.reference) {
		zend_hash_destroy(&processor.zvalptrs);
	}

	return dst;
}
/* }}} */
/* export: void xc_dprint(xc_entry_t *src, int indent TSRMLS_DC); :export {{{ */
#ifdef HAVE_XCACHE_DPRINT
void xc_dprint(xc_entry_t *src, int indent TSRMLS_DC) {
	
	xc_dprint_xc_entry_t(src, indent TSRMLS_CC);
}
#endif
/* }}} */



  /* {{{ Pre-declare */
 	  	 	void 	xc_calc_zval( 		 		xc_processor_t *processor, const zval * const src 		 		 		 		 		TSRMLS_DC 	); 	
 	  	 	static void inline 	xc_calc_zval_ptr( 		 		xc_processor_t *processor, const zval_ptr * const src 		 		 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_calc_zend_op_array( 		 		xc_processor_t *processor, const zend_op_array * const src 		 		 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_calc_zend_class_entry( 		 		xc_processor_t *processor, const zend_class_entry * const src 		 		 		 		 		TSRMLS_DC 	); 	
#ifdef HAVE_XCACHE_CONSTANT
 	  	 	static void inline 	xc_calc_zend_constant( 		 		xc_processor_t *processor, const zend_constant * const src 		 		 		 		 		TSRMLS_DC 	); 	
#endif
 	  	 	void 	xc_calc_zend_function( 		 		xc_processor_t *processor, const zend_function * const src 		 		 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_calc_xc_entry_t( 		 		xc_processor_t *processor, const xc_entry_t * const src 		 		 		 		 		TSRMLS_DC 	); 	
#ifdef ZEND_ENGINE_2
 	  	 	static void inline 	xc_calc_zend_property_info( 		 		xc_processor_t *processor, const zend_property_info * const src 		 		 		 		 		TSRMLS_DC 	); 	
#endif
/* }}} */
#ifdef IS_CV

	
	

/* {{{ xc_calc_zend_compiled_variable */
	
 	  	 	static void inline 	xc_calc_zend_compiled_variable( 		 		xc_processor_t *processor, const zend_compiled_variable * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_compiled_variable*)NULL)->name) + sizeof(((zend_compiled_variable*)NULL)->name_len) + sizeof(((zend_compiled_variable*)NULL)->hash_value)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(int,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,hash_value) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hash_value);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_calc_zend_uint */
	
 	  	 	static void inline 	xc_calc_zend_uint( 		 		xc_processor_t *processor, const zend_uint * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_uint), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifndef ZEND_ENGINE_2

	
	

/* {{{ xc_calc_int */
	
 	  	 	static void inline 	xc_calc_int( 		 		xc_processor_t *processor, const int * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(int), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_calc_zend_try_catch_element */
	
 	  	 	static void inline 	xc_calc_zend_try_catch_element( 		 		xc_processor_t *processor, const zend_try_catch_element * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_try_catch_element*)NULL)->try_op) + sizeof(((zend_try_catch_element*)NULL)->catch_op)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(zend_uint,try_op) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_op);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,catch_op) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->catch_op);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif /* ifdef ZEND_ENGINE_2 */

	
	

/* {{{ xc_calc_zend_brk_cont_element */
	
 	  	 	static void inline 	xc_calc_zend_brk_cont_element( 		 		xc_processor_t *processor, const zend_brk_cont_element * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_brk_cont_element*)NULL)->start) + sizeof(((zend_brk_cont_element*)NULL)->cont) + sizeof(((zend_brk_cont_element*)NULL)->brk) + sizeof(((zend_brk_cont_element*)NULL)->parent)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
#ifdef ZEND_ENGINE_2_2
#ifndef IS_UNICODE
	
	
	/* DISPATCH(int,start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start);
	done_count ++;

#endif



	

#endif
#endif
	
	
	/* DISPATCH(int,cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,brk) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,parent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	
	

/* {{{ xc_calc_HashTable_zval_ptr */
	
 	  	 	static void inline 	xc_calc_HashTable_zval_ptr( 		 		xc_processor_t *processor, const HashTable * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(Bucket*) * src->nTableSize));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(Bucket*) * src->nTableSize;
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(char) * bucketsize));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(char) * bucketsize;
	
	
	
	
	

			
			
			
			if (sizeof(void *) == sizeof(zval_ptr)) {
				
								
	
	/* STRUCT_P_EX(zval_ptr,pnew->pData,((zval_ptr*)b->pData),,, ) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof((  ((zval_ptr*)b->pData))[0]));
	
	
	xc_calc_zval_ptr(
		
		processor,   ((zval_ptr*)b->pData)
		
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zval_ptr,pnew->pData,((zval_ptr*)b->pData),,) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof(( ((zval_ptr*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zval_ptr)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zval_ptr);
	
	
	
	
	

	
	xc_calc_zval_ptr(
		
		processor,  ((zval_ptr*)b->pData)
		
		
		
		
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	
	

/* {{{ xc_calc_HashTable_zend_constant */
	
 	  	 	static void inline 	xc_calc_HashTable_zend_constant( 		 		xc_processor_t *processor, const HashTable * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(Bucket*) * src->nTableSize));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(Bucket*) * src->nTableSize;
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(char) * bucketsize));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(char) * bucketsize;
	
	
	
	
	

			
			
			
			if (sizeof(void *) == sizeof(zend_constant)) {
				
								
	
	/* STRUCT_P_EX(zend_constant,pnew->pData,((zend_constant*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_constant) == sizeof((  ((zend_constant*)b->pData))[0]));
	
	
	xc_calc_zend_constant(
		
		processor,   ((zend_constant*)b->pData)
		
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_constant,pnew->pData,((zend_constant*)b->pData),,) */

	
	

	
	assert(sizeof(zend_constant) == sizeof(( ((zend_constant*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_constant)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_constant);
	
	
	
	
	

	
	xc_calc_zend_constant(
		
		processor,  ((zend_constant*)b->pData)
		
		
		
		
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	
	

/* {{{ xc_calc_HashTable_zend_function */
	
		/* export:  	  	 	void 	xc_calc_HashTable_zend_function( 		 		xc_processor_t *processor, const HashTable * const src 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_calc_HashTable_zend_function( 		 		xc_processor_t *processor, const HashTable * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(Bucket*) * src->nTableSize));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(Bucket*) * src->nTableSize;
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(char) * bucketsize));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(char) * bucketsize;
	
	
	
	
	

			
			
			
			if (sizeof(void *) == sizeof(zend_function)) {
				
								
	
	/* STRUCT_P_EX(zend_function,pnew->pData,((zend_function*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_function) == sizeof((  ((zend_function*)b->pData))[0]));
	
	
	xc_calc_zend_function(
		
		processor,   ((zend_function*)b->pData)
		
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_function,pnew->pData,((zend_function*)b->pData),,) */

	
	

	
	assert(sizeof(zend_function) == sizeof(( ((zend_function*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_function)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_function);
	
	
	
	
	

	
	xc_calc_zend_function(
		
		processor,  ((zend_function*)b->pData)
		
		
		
		
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef ZEND_ENGINE_2

	
	
	

/* {{{ xc_calc_HashTable_zend_property_info */
	
 	  	 	static void inline 	xc_calc_HashTable_zend_property_info( 		 		xc_processor_t *processor, const HashTable * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(Bucket*) * src->nTableSize));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(Bucket*) * src->nTableSize;
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(char) * bucketsize));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(char) * bucketsize;
	
	
	
	
	

			
			
			
			if (sizeof(void *) == sizeof(zend_property_info)) {
				
								
	
	/* STRUCT_P_EX(zend_property_info,pnew->pData,((zend_property_info*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_property_info) == sizeof((  ((zend_property_info*)b->pData))[0]));
	
	
	xc_calc_zend_property_info(
		
		processor,   ((zend_property_info*)b->pData)
		
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_property_info,pnew->pData,((zend_property_info*)b->pData),,) */

	
	

	
	assert(sizeof(zend_property_info) == sizeof(( ((zend_property_info*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_property_info)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_property_info);
	
	
	
	
	

	
	xc_calc_zend_property_info(
		
		processor,  ((zend_property_info*)b->pData)
		
		
		
		
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	

/* {{{ xc_calc_zval */
	
		/* export:  	  	 	void 	xc_calc_zval( 		 		xc_processor_t *processor, const zval * const src 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_calc_zval( 		 		xc_processor_t *processor, const zval * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zval*)NULL)->value) + sizeof(((zval*)NULL)->refcount) + sizeof(((zval*)NULL)->type) + sizeof(((zval*)NULL)->is_ref)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
				/* Variable information */
		
	
	

		switch (src->type & ~IS_CONSTANT_INDEX) {
			case IS_LONG:
			case IS_RESOURCE:
			case IS_BOOL:
				
	
	/* DISPATCH(long,value.lval) */

	
	
	
	
	

	

				break;
			case IS_DOUBLE:
				
	
	/* DISPATCH(double,value.dval) */

	
	
	
	
	

	

				break;
			case IS_NULL:
				
				break;

			case IS_CONSTANT:
#ifdef IS_UNICODE
				if (UG(unicode)) {
					goto proc_unicode;
				}
#endif
			case IS_STRING:
#ifdef FLAG_IS_BC
			case FLAG_IS_BC:
#endif
				
	
	/* DISPATCH(int,value.str.len) */

	
	
	
	
	

	

				
	/* PROC_STRING_L(value.str.val,value.str.len) */
 
	/* PROC_STRING_N(value.str.val,value.str.len + 1) */
 
	
	
	
	
	
	
	
	if (src->value.str.val == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, ZSTR(src->value.str.val), src->value.str.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

				break;
#ifdef IS_UNICODE
			case IS_UNICODE:
proc_unicode:
				
	
	/* DISPATCH(int32_t,value.uni.len) */

	
	
	
	
	

	

				
	/* PROC_ZSTRING_L(1,value.uni.val,value.uni.len) */
 
	
	/* PROC_ZSTRING_N(1,value.uni.val,value.uni.len,len) */

#ifdef IS_UNICODE
	
	
	
	
	
	
	
	
	
	
	if (ZSTR_U(src->value.uni.val) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->value.uni.val, src->value.uni.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

	
#else
	
	
	
	
	
	
	
	
	
	if (ZSTR_S(src->value.uni.val) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->value.uni.val, src->value.uni.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

				break;
#endif

			case IS_ARRAY:
			case IS_CONSTANT_ARRAY:
				
	
	/* STRUCT_P(HashTable,value.ht,HashTable_zval_ptr) */

	if (src->value.ht) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->value.ht,src->value.ht,value.ht,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->value.ht)[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(HashTable)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(HashTable);
	
	
	
	
	

	
	xc_calc_HashTable_zval_ptr(
		
		processor,  src->value.ht
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

				break;

			case IS_OBJECT:
				
				#ifndef ZEND_ENGINE_2
				
	
	/* STRUCT_P(zend_class_entry,value.obj.ce) */

	if (src->value.obj.ce) {
		
		
	
	/* STRUCT_P_EX(zend_class_entry,dst->value.obj.ce,src->value.obj.ce,value.obj.ce,) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof(( src->value.obj.ce)[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_class_entry)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_class_entry);
	
	
	
	
	

	
	xc_calc_zend_class_entry(
		
		processor,  src->value.obj.ce
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

				
	
	/* STRUCT_P(HashTable,value.obj.properties,HashTable_zval_ptr) */

	if (src->value.obj.properties) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->value.obj.properties,src->value.obj.properties,value.obj.properties,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->value.obj.properties)[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(HashTable)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(HashTable);
	
	
	
	
	

	
	xc_calc_HashTable_zval_ptr(
		
		processor,  src->value.obj.properties
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

#endif
				break;

			default:
				assert(0);
		}
		
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif


		
	
	/* DISPATCH(zval_data_type,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_uchar,is_ref) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->is_ref);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_ushort,refcount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_3
		
	
	/* DISPATCH(zend_uchar,idx_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->idx_type);
	done_count ++;

#endif



	

#endif
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_calc_zval_ptr */
	
 	  	 	static void inline 	xc_calc_zval_ptr( 		 		xc_processor_t *processor, const zval_ptr * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zval_ptr), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
		do {
			
				if (processor->reference) {
					zval_ptr *ppzv;
					if (zend_hash_find(&processor->zvalptrs, (char *) &src[0], sizeof(src[0]), (void **) &ppzv) == SUCCESS) {
						
						processor->have_references = 1;
						
						
						break;
					}
				}
			
			
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zval)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zval);
	
	
	
	
	

			
				if (processor->reference) {
					
						/* make dummy */
						zval_ptr pzv = (zval_ptr)-1;
					
					if (zend_hash_add(&processor->zvalptrs, (char *) &src[0], sizeof(src[0]), (void *) &pzv, sizeof(pzv), NULL) == SUCCESS) {
						/* first add, go on */
											}
					else {
						assert(0);
					}
				}
			
			
			
			
	
	/* STRUCT_P_EX(zval,dst[0],src[0],[0],, ) */

	
	

	
	assert(sizeof(zval) == sizeof((  src[0])[0]));
	
	
	xc_calc_zval(
		
		processor,   src[0]
		
		
		
		
		TSRMLS_CC
	);
	
	
	

			
		} while (0);
	
	
#ifndef NDEBUG
		
	done_size += sizeof(zval_ptr);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_calc_zend_arg_info */
	
 	  	 	static void inline 	xc_calc_zend_arg_info( 		 		xc_processor_t *processor, const zend_arg_info * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_arg_info*)NULL)->name) + sizeof(((zend_arg_info*)NULL)->name_len) + sizeof(((zend_arg_info*)NULL)->class_name) + sizeof(((zend_arg_info*)NULL)->class_name_len) + sizeof(((zend_arg_info*)NULL)->array_type_hint) + sizeof(((zend_arg_info*)NULL)->allow_null) + sizeof(((zend_arg_info*)NULL)->pass_by_reference) + sizeof(((zend_arg_info*)NULL)->return_reference) + sizeof(((zend_arg_info*)NULL)->required_num_args)  ), assert_count = 9;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(zend_uint,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_uint,class_name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,class_name,class_name_len) */
 
	
	/* PROC_ZSTRING_N(,class_name,class_name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->class_name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->class_name, src->class_name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->class_name, src->class_name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->class_name, src->class_name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_bool,array_type_hint) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->array_type_hint);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,allow_null) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->allow_null);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_by_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_by_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,return_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,required_num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_calc_zend_constant */
	
 	  	 	static void inline 	xc_calc_zend_constant( 		 		xc_processor_t *processor, const zend_constant * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_constant*)NULL)->value) + sizeof(((zend_constant*)NULL)->flags) + sizeof(((zend_constant*)NULL)->name) + sizeof(((zend_constant*)NULL)->name_len) + sizeof(((zend_constant*)NULL)->module_number)  ), assert_count = 5;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
	
	/* STRUCT(zval,value) */

	assert(sizeof(zval) == sizeof(src->value));
	
	
	
	/* STRUCT_P_EX(zval,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->value)[0]));
	
	
	xc_calc_zval(
		
		processor, & src->value
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



	
	
	/* DISPATCH(int,flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	
	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	
	/* DISPATCH(int,module_number) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module_number);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_calc_zend_function */
	
		/* export:  	  	 	void 	xc_calc_zend_function( 		 		xc_processor_t *processor, const zend_function * const src 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_calc_zend_function( 		 		xc_processor_t *processor, const zend_function * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_function), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
	
	

	switch (src->type) {
	case ZEND_INTERNAL_FUNCTION:
	case ZEND_OVERLOADED_FUNCTION:
		
		break;

	case ZEND_USER_FUNCTION:
	case ZEND_EVAL_CODE:
		
		
	
	/* STRUCT(zend_op_array,op_array) */

	assert(sizeof(zend_op_array) == sizeof(src->op_array));
	
	
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,,&) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof((& src->op_array)[0]));
	
	
	xc_calc_zend_op_array(
		
		processor, & src->op_array
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

		break;

	default:
		assert(0);
	}
	
	
	

	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_calc_zend_property_info */
	
 	  	 	static void inline 	xc_calc_zend_property_info( 		 		xc_processor_t *processor, const zend_property_info * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_property_info*)NULL)->flags) + sizeof(((zend_property_info*)NULL)->name) + sizeof(((zend_property_info*)NULL)->name_length) + sizeof(((zend_property_info*)NULL)->h) + sizeof(((zend_property_info*)NULL)->doc_comment) + sizeof(((zend_property_info*)NULL)->doc_comment_len) + sizeof(((zend_property_info*)NULL)->ce)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(zend_uint,flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,name_length) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,h) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->h);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif
	#if defined(ZEND_ENGINE_2_2) && !defined(IS_UNICODE)
	
	if (src->ce) {
		
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce);
	done_count ++;

#endif


#endif

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_calc_zend_class_entry */
	
		/* export:  	  	 	void 	xc_calc_zend_class_entry( 		 		xc_processor_t *processor, const zend_class_entry * const src 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_calc_zend_class_entry( 		 		xc_processor_t *processor, const zend_class_entry * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_class_entry*)NULL)->type) + sizeof(((zend_class_entry*)NULL)->name) + sizeof(((zend_class_entry*)NULL)->name_length) + sizeof(((zend_class_entry*)NULL)->parent) + sizeof(((zend_class_entry*)NULL)->refcount) + sizeof(((zend_class_entry*)NULL)->constants_updated) + sizeof(((zend_class_entry*)NULL)->ce_flags) + sizeof(((zend_class_entry*)NULL)->function_table) + sizeof(((zend_class_entry*)NULL)->default_properties) + sizeof(((zend_class_entry*)NULL)->properties_info) + sizeof(((zend_class_entry*)NULL)->default_static_members) + sizeof(((zend_class_entry*)NULL)->static_members) + sizeof(((zend_class_entry*)NULL)->constants_table) + sizeof(((zend_class_entry*)NULL)->builtin_functions) + sizeof(((zend_class_entry*)NULL)->constructor) + sizeof(((zend_class_entry*)NULL)->destructor) + sizeof(((zend_class_entry*)NULL)->clone) + sizeof(((zend_class_entry*)NULL)->__get) + sizeof(((zend_class_entry*)NULL)->__set) + sizeof(((zend_class_entry*)NULL)->__unset) + sizeof(((zend_class_entry*)NULL)->__isset) + sizeof(((zend_class_entry*)NULL)->__call) + sizeof(((zend_class_entry*)NULL)->__tostring) + sizeof(((zend_class_entry*)NULL)->serialize_func) + sizeof(((zend_class_entry*)NULL)->unserialize_func) + sizeof(((zend_class_entry*)NULL)->iterator_funcs) + sizeof(((zend_class_entry*)NULL)->create_object) + sizeof(((zend_class_entry*)NULL)->get_iterator) + sizeof(((zend_class_entry*)NULL)->interface_gets_implemented) + sizeof(((zend_class_entry*)NULL)->serialize) + sizeof(((zend_class_entry*)NULL)->unserialize) + sizeof(((zend_class_entry*)NULL)->interfaces) + sizeof(((zend_class_entry*)NULL)->num_interfaces) + sizeof(((zend_class_entry*)NULL)->filename) + sizeof(((zend_class_entry*)NULL)->line_start) + sizeof(((zend_class_entry*)NULL)->line_end) + sizeof(((zend_class_entry*)NULL)->doc_comment) + sizeof(((zend_class_entry*)NULL)->doc_comment_len) + sizeof(((zend_class_entry*)NULL)->module)  ), assert_count = 39;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
		processor->active_class_entry_src = src;
		
	
	
	
	/* DISPATCH(char,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,name_length) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
		
	if (src->parent) {
		
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif


	
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(int,refcount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	

#else
	
	
	/* STRUCT_P(int,refcount) */

	if (src->refcount) {
		
		
	
	/* STRUCT_P_EX(int,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(int) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(int)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(int);
	
	
	
	
	

	
	xc_calc_int(
		
		processor,  src->refcount
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(zend_bool,constants_updated) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_updated);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,ce_flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce_flags);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT(HashTable,default_properties,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_properties));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->default_properties,src->default_properties,default_properties,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_properties)[0]));
	
	
	xc_calc_HashTable_zval_ptr(
		
		processor, & src->default_properties
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_properties);
	done_count ++;

#endif



	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->builtin_functions);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT(HashTable,properties_info,HashTable_zend_property_info) */

	assert(sizeof(HashTable) == sizeof(src->properties_info));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->properties_info,src->properties_info,properties_info,HashTable_zend_property_info,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->properties_info)[0]));
	
	
	xc_calc_HashTable_zend_property_info(
		
		processor, & src->properties_info
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->properties_info);
	done_count ++;

#endif



#	ifdef ZEND_ENGINE_2_1
	
	
	/* STRUCT(HashTable,default_static_members,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_static_members));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->default_static_members,src->default_static_members,default_static_members,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_static_members)[0]));
	
	
	xc_calc_HashTable_zval_ptr(
		
		processor, & src->default_static_members
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_static_members);
	done_count ++;

#endif



	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif


#	else
	
	
	/* STRUCT_P(HashTable,static_members,HashTable_zval_ptr) */

	if (src->static_members) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->static_members,src->static_members,static_members,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_members)[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(HashTable)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(HashTable);
	
	
	
	
	

	
	xc_calc_HashTable_zval_ptr(
		
		processor,  src->static_members
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif



#	endif
	
	
	/* STRUCT(HashTable,constants_table,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->constants_table));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->constants_table,src->constants_table,constants_table,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->constants_table)[0]));
	
	
	xc_calc_HashTable_zval_ptr(
		
		processor, & src->constants_table
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_table);
	done_count ++;

#endif




		
	
	
		
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interfaces);
	done_count ++;

#endif


		
	
	
	
	/* DISPATCH(zend_uint,num_interfaces) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_interfaces);
	done_count ++;

#endif



	


	
	/* PROC_STRING(filename) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->filename == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, ZSTR(src->filename), strlen(src->filename) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->iterator_funcs);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->create_object);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_iterator);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interface_gets_implemented);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2_3
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_static_method);
	done_count ++;

#endif


#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize);
	done_count ++;

#endif


	/* deal with it inside xc_fix_method */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->destructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->clone);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__set);
	done_count ++;

#endif


/* should be >5.1 */
#ifdef ZEND_ENGINE_2_1
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__unset);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__isset);
	done_count ++;

#endif


# if defined(ZEND_ENGINE_2_2) || PHP_MAJOR_VERSION >= 6
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__tostring);
	done_count ++;

#endif


# endif
#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__call);
	done_count ++;

#endif


#ifdef ZEND_CALLSTATIC_FUNC_NAME
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__callstatic);
	done_count ++;

#endif


#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module);
	done_count ++;

#endif


#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_function_call);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_set);
	done_count ++;

#endif


#endif
		
	
	/* STRUCT(HashTable,function_table,HashTable_zend_function) */

	assert(sizeof(HashTable) == sizeof(src->function_table));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->function_table,src->function_table,function_table,HashTable_zend_function,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->function_table)[0]));
	
	
	xc_calc_HashTable_zend_function(
		
		processor, & src->function_table
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_table);
	done_count ++;

#endif



	
	
		processor->active_class_entry_src = NULL;
		
	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_calc_znode */
	
 	  	 	static void inline 	xc_calc_znode( 		 		xc_processor_t *processor, const znode * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((znode*)NULL)->op_type) + sizeof(((znode*)NULL)->u)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
	
	/* DISPATCH(int,op_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_type);
	done_count ++;

#endif



	


#ifdef IS_CV
#	define XCACHE_IS_CV IS_CV
#else
/* compatible with zend optimizer */
#	define XCACHE_IS_CV 16
#endif
	assert(src->op_type == IS_CONST ||
		src->op_type == IS_VAR ||
		src->op_type == XCACHE_IS_CV ||
		src->op_type == IS_TMP_VAR ||
		src->op_type == IS_UNUSED);
#undef XCACHE_IS_CV
		
	
	

	switch (src->op_type) {
		case IS_CONST:
			
	
	/* STRUCT(zval,u.constant) */

	assert(sizeof(zval) == sizeof(src->u.constant));
	
	
	
	/* STRUCT_P_EX(zval,dst->u.constant,src->u.constant,u.constant,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->u.constant)[0]));
	
	
	xc_calc_zval(
		
		processor, & src->u.constant
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

			break;
		
		case IS_VAR:
		case IS_TMP_VAR:
#ifdef IS_CV
		case IS_CV:
#else
		case 16:
#endif
			
	
	/* DISPATCH(zend_uint,u.var) */

	
	
	
	
	

	

			
	
	/* DISPATCH(zend_uint,u.EA.type) */

	
	
	
	
	

	

			break;
		case IS_UNUSED:
			
			
	
	/* DISPATCH(zend_uint,u.opline_num) */

	
	
	
	
	

	

#ifndef ZEND_ENGINE_2
			
	
	/* DISPATCH(zend_uint,u.fetch_type) */

	
	
	
	
	

	

#endif
			
	
	/* DISPATCH(zend_uint,u.EA.type) */

	
	
	
	
	

	

			break;
		
	}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->u);
	done_count ++;

#endif



		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_calc_zend_op */
	
		/* export:  	  	 	void 	xc_calc_zend_op( 		 		xc_processor_t *processor, const zend_op * const src 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_calc_zend_op( 		 		xc_processor_t *processor, const zend_op * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op*)NULL)->handler) + sizeof(((zend_op*)NULL)->result) + sizeof(((zend_op*)NULL)->op1) + sizeof(((zend_op*)NULL)->op2) + sizeof(((zend_op*)NULL)->extended_value) + sizeof(((zend_op*)NULL)->lineno) + sizeof(((zend_op*)NULL)->opcode)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uchar,opcode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcode);
	done_count ++;

#endif



	

	
	
	/* STRUCT(znode,result) */

	assert(sizeof(znode) == sizeof(src->result));
	
	
	
	/* STRUCT_P_EX(znode,dst->result,src->result,result,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->result)[0]));
	
	
	xc_calc_znode(
		
		processor, & src->result
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->result);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op1) */

	assert(sizeof(znode) == sizeof(src->op1));
	
	
	
	/* STRUCT_P_EX(znode,dst->op1,src->op1,op1,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op1)[0]));
	
	
	xc_calc_znode(
		
		processor, & src->op1
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op1);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op2) */

	assert(sizeof(znode) == sizeof(src->op2));
	
	
	
	/* STRUCT_P_EX(znode,dst->op2,src->op2,op2,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op2)[0]));
	
	
	xc_calc_znode(
		
		processor, & src->op2
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op2);
	done_count ++;

#endif



	
	
	/* DISPATCH(ulong,extended_value) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->extended_value);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,lineno) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->lineno);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	
	/* DISPATCH(opcode_handler_t,handler) */

	/* is copying enough? */
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handler);
	done_count ++;

#endif


	

#endif

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_calc_zend_op_array */
	
		/* export:  	  	 	void 	xc_calc_zend_op_array( 		 		xc_processor_t *processor, const zend_op_array * const src 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_calc_zend_op_array( 		 		xc_processor_t *processor, const zend_op_array * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op_array*)NULL)->type) + sizeof(((zend_op_array*)NULL)->function_name) + sizeof(((zend_op_array*)NULL)->scope) + sizeof(((zend_op_array*)NULL)->fn_flags) + sizeof(((zend_op_array*)NULL)->prototype) + sizeof(((zend_op_array*)NULL)->num_args) + sizeof(((zend_op_array*)NULL)->required_num_args) + sizeof(((zend_op_array*)NULL)->arg_info) + sizeof(((zend_op_array*)NULL)->pass_rest_by_reference) + sizeof(((zend_op_array*)NULL)->return_reference) + sizeof(((zend_op_array*)NULL)->refcount) + sizeof(((zend_op_array*)NULL)->opcodes) + sizeof(((zend_op_array*)NULL)->last) + sizeof(((zend_op_array*)NULL)->size) + sizeof(((zend_op_array*)NULL)->vars) + sizeof(((zend_op_array*)NULL)->last_var) + sizeof(((zend_op_array*)NULL)->size_var) + sizeof(((zend_op_array*)NULL)->T) + sizeof(((zend_op_array*)NULL)->brk_cont_array) + sizeof(((zend_op_array*)NULL)->last_brk_cont) + sizeof(((zend_op_array*)NULL)->current_brk_cont) + sizeof(((zend_op_array*)NULL)->try_catch_array) + sizeof(((zend_op_array*)NULL)->last_try_catch) + sizeof(((zend_op_array*)NULL)->static_variables) + sizeof(((zend_op_array*)NULL)->start_op) + sizeof(((zend_op_array*)NULL)->backpatch_count) + sizeof(((zend_op_array*)NULL)->done_pass_two) + sizeof(((zend_op_array*)NULL)->uses_this) + sizeof(((zend_op_array*)NULL)->filename) + sizeof(((zend_op_array*)NULL)->line_start) + sizeof(((zend_op_array*)NULL)->line_end) + sizeof(((zend_op_array*)NULL)->doc_comment) + sizeof(((zend_op_array*)NULL)->doc_comment_len) + sizeof(((zend_op_array*)NULL)->reserved)  ), assert_count = 34;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
	do {
		zend_uint ii;
	int i;

	/* Common elements */
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING(,function_name) */
 
	
	/* PROC_ZSTRING_N(,function_name,,strlen) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->function_name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->function_name, xc_zstrlen_uchar (src->function_name) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->function_name, xc_zstrlen_char (src->function_name) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->function_name, xc_zstrlen_char (src->function_name) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,fn_flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->fn_flags);
	done_count ++;

#endif



	

	


	if (src->arg_info) {
		
	

		
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_arg_info) * src->num_args));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_arg_info) * src->num_args;
	
	
	
	
	

			
			for (ii = 0; ii < src->num_args; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_arg_info,arg_info[i],) */

	assert(sizeof(zend_arg_info) == sizeof(src->arg_info[ii]));
	
	
	
	/* STRUCT_P_EX(zend_arg_info,dst->arg_info[i],src->arg_info[i],arg_info[i],,&) */

	
	

	
	assert(sizeof(zend_arg_info) == sizeof((& src->arg_info[ii])[0]));
	
	
	xc_calc_zend_arg_info(
		
		processor, & src->arg_info[ii]
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,required_num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_rest_by_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_rest_by_reference);
	done_count ++;

#endif



	

#else
	if (src->arg_types) {
		
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_uchar) * src->arg_types[0] + 1));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_uchar) * src->arg_types[0] + 1;
	
	
	
	
	

		
		
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif


	}
	else {
		
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif



		
	}
#endif
	
	
	/* DISPATCH(unsigned char,return_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	/* END of common elements */
#ifdef IS_UNICODE
	#endif

	
	
	/* STRUCT_P(zend_uint,refcount) */

	if (src->refcount) {
		
		
	
	/* STRUCT_P_EX(zend_uint,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(zend_uint) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_uint)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_uint);
	
	
	
	
	

	
	xc_calc_zend_uint(
		
		processor,  src->refcount
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	
	

	
	


	if (src->opcodes) {
		
	

		
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_op) * src->last));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_op) * src->last;
	
	
	
	
	

			
			for (ii = 0; ii < src->last; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_op,opcodes[i],) */

	assert(sizeof(zend_op) == sizeof(src->opcodes[ii]));
	
	
	
	/* STRUCT_P_EX(zend_op,dst->opcodes[i],src->opcodes[i],opcodes[i],,&) */

	
	

	
	assert(sizeof(zend_op) == sizeof((& src->opcodes[ii])[0]));
	
	
	xc_calc_zend_op(
		
		processor, & src->opcodes[ii]
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif



		
	}



	
	
	
	/* DISPATCH(zend_uint,last) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



	


#ifdef IS_CV
	
	if (src->vars) {
		
	

		
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_compiled_variable) * src->last_var));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_compiled_variable) * src->last_var;
	
	
	
	
	

			
			for (i = 0; i < src->last_var; i ++) {
				
	
	

					
					
	
	/* STRUCT(zend_compiled_variable,vars[i],) */

	assert(sizeof(zend_compiled_variable) == sizeof(src->vars[i]));
	
	
	
	/* STRUCT_P_EX(zend_compiled_variable,dst->vars[i],src->vars[i],vars[i],,&) */

	
	

	
	assert(sizeof(zend_compiled_variable) == sizeof((& src->vars[i])[0]));
	
	
	xc_calc_zend_compiled_variable(
		
		processor, & src->vars[i]
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_var) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_var);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,size_var) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size_var);
	done_count ++;

#endif



	

#else
		
#endif

	
	
	/* DISPATCH(zend_uint,T) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->T);
	done_count ++;

#endif



	


	


	if (src->brk_cont_array) {
		
	

		
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_brk_cont_element) * src->last_brk_cont));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_brk_cont_element) * src->last_brk_cont;
	
	
	
	
	

			
			for (ii = 0; ii < src->last_brk_cont; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_brk_cont_element,brk_cont_array[i],) */

	assert(sizeof(zend_brk_cont_element) == sizeof(src->brk_cont_array[ii]));
	
	
	
	/* STRUCT_P_EX(zend_brk_cont_element,dst->brk_cont_array[i],src->brk_cont_array[i],brk_cont_array[i],,&) */

	
	

	
	assert(sizeof(zend_brk_cont_element) == sizeof((& src->brk_cont_array[ii])[0]));
	
	
	xc_calc_zend_brk_cont_element(
		
		processor, & src->brk_cont_array[ii]
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,last_brk_cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_brk_cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,current_brk_cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->current_brk_cont);
	done_count ++;

#endif



	

#ifndef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_globals) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_globals);
	done_count ++;

#endif



	

#endif

#ifdef ZEND_ENGINE_2
	
	if (src->try_catch_array) {
		
	

		
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_try_catch_element) * src->last_try_catch));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_try_catch_element) * src->last_try_catch;
	
	
	
	
	

			
			for (i = 0; i < src->last_try_catch; i ++) {
				
	
	

					
					
	
	/* STRUCT(zend_try_catch_element,try_catch_array[i],) */

	assert(sizeof(zend_try_catch_element) == sizeof(src->try_catch_array[i]));
	
	
	
	/* STRUCT_P_EX(zend_try_catch_element,dst->try_catch_array[i],src->try_catch_array[i],try_catch_array[i],,&) */

	
	

	
	assert(sizeof(zend_try_catch_element) == sizeof((& src->try_catch_array[i])[0]));
	
	
	xc_calc_zend_try_catch_element(
		
		processor, & src->try_catch_array[i]
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_try_catch) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_try_catch);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT_P(HashTable,static_variables,HashTable_zval_ptr) */

	if (src->static_variables) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->static_variables,src->static_variables,static_variables,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_variables)[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(HashTable)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(HashTable);
	
	
	
	
	

	
	xc_calc_HashTable_zval_ptr(
		
		processor,  src->static_variables
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_variables);
	done_count ++;

#endif




	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start_op);
	done_count ++;

#endif


	
	
	/* DISPATCH(int,backpatch_count) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->backpatch_count);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(zend_bool,done_pass_two) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->done_pass_two);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_this) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_this);
	done_count ++;

#endif



	

#endif

	
	/* PROC_STRING(filename) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->filename == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, ZSTR(src->filename), strlen(src->filename) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#ifdef IS_UNICODE
	
		
	/* PROC_STRING(script_encoding) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->script_encoding);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->script_encoding == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, ZSTR(src->script_encoding), strlen(src->script_encoding) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

	
#endif
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif

	/* reserved */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->reserved);
	done_count ++;

#endif


#if defined(HARDENING_PATCH) && HARDENING_PATCH
	
	
	/* DISPATCH(zend_bool,created_by_eval) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->created_by_eval);
	done_count ++;

#endif



	

#endif
	} while (0);

#ifdef ZEND_ENGINE_2
		
			
				
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->prototype);
	done_count ++;

#endif



			
	
#endif

	
#ifdef ZEND_ENGINE_2
		
	if (src->scope) {
		
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->scope);
	done_count ++;

#endif


#endif
	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_zend_op_array : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_zend_op_array : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_calc_xc_constinfo_t */
	
 	  	 	static void inline 	xc_calc_xc_constinfo_t( 		 		xc_processor_t *processor, const xc_constinfo_t * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_constinfo_t*)NULL)->key) + sizeof(((xc_constinfo_t*)NULL)->key_size) + sizeof(((xc_constinfo_t*)NULL)->constant)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	/* STRUCT(zend_constant,constant) */

	assert(sizeof(zend_constant) == sizeof(src->constant));
	
	
	
	/* STRUCT_P_EX(zend_constant,dst->constant,src->constant,constant,,&) */

	
	

	
	assert(sizeof(zend_constant) == sizeof((& src->constant)[0]));
	
	
	xc_calc_zend_constant(
		
		processor, & src->constant
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constant);
	done_count ++;

#endif




		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_calc_xc_funcinfo_t */
	
		/* export:  	  	 	void 	xc_calc_xc_funcinfo_t( 		 		xc_processor_t *processor, const xc_funcinfo_t * const src 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_calc_xc_funcinfo_t( 		 		xc_processor_t *processor, const xc_funcinfo_t * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_funcinfo_t*)NULL)->key) + sizeof(((xc_funcinfo_t*)NULL)->key_size) + sizeof(((xc_funcinfo_t*)NULL)->func)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	/* STRUCT(zend_function,func) */

	assert(sizeof(zend_function) == sizeof(src->func));
	
	
	
	/* STRUCT_P_EX(zend_function,dst->func,src->func,func,,&) */

	
	

	
	assert(sizeof(zend_function) == sizeof((& src->func)[0]));
	
	
	xc_calc_zend_function(
		
		processor, & src->func
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->func);
	done_count ++;

#endif




		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_calc_xc_classinfo_t */
	
		/* export:  	  	 	void 	xc_calc_xc_classinfo_t( 		 		xc_processor_t *processor, const xc_classinfo_t * const src 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_calc_xc_classinfo_t( 		 		xc_processor_t *processor, const xc_classinfo_t * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_classinfo_t*)NULL)->key) + sizeof(((xc_classinfo_t*)NULL)->key_size) + sizeof(((xc_classinfo_t*)NULL)->cest) + sizeof(((xc_classinfo_t*)NULL)->oplineno)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT_P(zend_class_entry,cest) */

	if (src->cest) {
		
		
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof(( src->cest)[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_class_entry)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_class_entry);
	
	
	
	
	

	
	xc_calc_zend_class_entry(
		
		processor,  src->cest
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#else
	
	
	/* STRUCT(zend_class_entry,cest) */

	assert(sizeof(zend_class_entry) == sizeof(src->cest));
	
	
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,,&) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof((& src->cest)[0]));
	
	
	xc_calc_zend_class_entry(
		
		processor, & src->cest
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(int,oplineno) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->oplineno);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2_1

	
	

/* {{{ xc_calc_xc_autoglobal_t */
	
 	  	 	static void inline 	xc_calc_xc_autoglobal_t( 		 		xc_processor_t *processor, const xc_autoglobal_t * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_autoglobal_t*)NULL)->key) + sizeof(((xc_autoglobal_t*)NULL)->key_len)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_len);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	/* PROC_ZSTRING_L(type,key,key_len) */
 
	
	/* PROC_ZSTRING_N(type,key,key_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->key, src->key_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->key, src->key_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->key, src->key_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_calc_xc_entry_data_php_t */
	
 	  	 	static void inline 	xc_calc_xc_entry_data_php_t( 		 		xc_processor_t *processor, const xc_entry_data_php_t * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_php_t*)NULL)->sourcesize) + sizeof(((xc_entry_data_php_t*)NULL)->device) + sizeof(((xc_entry_data_php_t*)NULL)->inode) + sizeof(((xc_entry_data_php_t*)NULL)->mtime) + sizeof(((xc_entry_data_php_t*)NULL)->op_array) + sizeof(((xc_entry_data_php_t*)NULL)->constinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->constinfos) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfos) + sizeof(((xc_entry_data_php_t*)NULL)->classinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->classinfos) + sizeof(((xc_entry_data_php_t*)NULL)->have_early_binding) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobal_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobals)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			zend_uint i;

#ifdef HAVE_INODE
	
	
	/* DISPATCH(int,device) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->device);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,inode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inode);
	done_count ++;

#endif



	

#endif
	
	
	/* DISPATCH(size_t,sourcesize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->sourcesize);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(time_t,mtime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->mtime);
	done_count ++;

#endif



	


	
	
	/* STRUCT_P(zend_op_array,op_array) */

	if (src->op_array) {
		
		
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof(( src->op_array)[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(zend_op_array)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(zend_op_array);
	
	
	
	
	

	
	xc_calc_zend_op_array(
		
		processor,  src->op_array
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_array);
	done_count ++;

#endif




#ifdef HAVE_XCACHE_CONSTANT
	
	
	/* DISPATCH(zend_uint,constinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->constinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(xc_constinfo_t) * src->constinfo_cnt));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(xc_constinfo_t) * src->constinfo_cnt;
	
	
	
	
	

			
			for (i = 0; i < src->constinfo_cnt; i ++) {
				
	
	

					
					
	
	/* STRUCT(xc_constinfo_t,constinfos[i],) */

	assert(sizeof(xc_constinfo_t) == sizeof(src->constinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_constinfo_t,dst->constinfos[i],src->constinfos[i],constinfos[i],,&) */

	
	

	
	assert(sizeof(xc_constinfo_t) == sizeof((& src->constinfos[i])[0]));
	
	
	xc_calc_xc_constinfo_t(
		
		processor, & src->constinfos[i]
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif



		
	}

#endif

	
	
	/* DISPATCH(zend_uint,funcinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->funcinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(xc_funcinfo_t) * src->funcinfo_cnt));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(xc_funcinfo_t) * src->funcinfo_cnt;
	
	
	
	
	

			
			for (i = 0; i < src->funcinfo_cnt; i ++) {
				
	
	

					
					
	
	/* STRUCT(xc_funcinfo_t,funcinfos[i],) */

	assert(sizeof(xc_funcinfo_t) == sizeof(src->funcinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_funcinfo_t,dst->funcinfos[i],src->funcinfos[i],funcinfos[i],,&) */

	
	

	
	assert(sizeof(xc_funcinfo_t) == sizeof((& src->funcinfos[i])[0]));
	
	
	xc_calc_xc_funcinfo_t(
		
		processor, & src->funcinfos[i]
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif



		
	}


	
	
	/* DISPATCH(zend_uint,classinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfo_cnt);
	done_count ++;

#endif



	

	
	
	if (src->classinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(xc_classinfo_t) * src->classinfo_cnt));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(xc_classinfo_t) * src->classinfo_cnt;
	
	
	
	
	

			
			for (i = 0; i < src->classinfo_cnt; i ++) {
				
	
	

					
		
	
					
	
	/* STRUCT(xc_classinfo_t,classinfos[i],) */

	assert(sizeof(xc_classinfo_t) == sizeof(src->classinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_classinfo_t,dst->classinfos[i],src->classinfos[i],classinfos[i],,&) */

	
	

	
	assert(sizeof(xc_classinfo_t) == sizeof((& src->classinfos[i])[0]));
	
	
	xc_calc_xc_classinfo_t(
		
		processor, & src->classinfos[i]
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif



		
	}

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,autoglobal_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobal_cnt);
	done_count ++;

#endif



	

	
	if (src->autoglobals) {
		
	

		
			
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(xc_autoglobal_t) * src->autoglobal_cnt));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(xc_autoglobal_t) * src->autoglobal_cnt;
	
	
	
	
	

			
			for (i = 0; i < src->autoglobal_cnt; i ++) {
				
	
	

					
		
	
					
	
	/* STRUCT(xc_autoglobal_t,autoglobals[i],) */

	assert(sizeof(xc_autoglobal_t) == sizeof(src->autoglobals[i]));
	
	
	
	/* STRUCT_P_EX(xc_autoglobal_t,dst->autoglobals[i],src->autoglobals[i],autoglobals[i],,&) */

	
	

	
	assert(sizeof(xc_autoglobal_t) == sizeof((& src->autoglobals[i])[0]));
	
	
	xc_calc_xc_autoglobal_t(
		
		processor, & src->autoglobals[i]
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif



		
	}

#endif
	
	
	/* DISPATCH(zend_bool,have_early_binding) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_early_binding);
	done_count ++;

#endif



	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_calc_xc_entry_data_var_t */
	
 	  	 	static void inline 	xc_calc_xc_entry_data_var_t( 		 		xc_processor_t *processor, const xc_entry_data_var_t * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_var_t*)NULL)->value)  ), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
			
	
	
	/* STRUCT_P_EX(zval_ptr,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof((& src->value)[0]));
	
	
	xc_calc_zval_ptr(
		
		processor, & src->value
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_calc_xc_entry_t */
	
		/* export:  	  	 	void 	xc_calc_xc_entry_t( 		 		xc_processor_t *processor, const xc_entry_t * const src 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_calc_xc_entry_t( 		 		xc_processor_t *processor, const xc_entry_t * const src 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_t*)NULL)->type) + sizeof(((xc_entry_t*)NULL)->hvalue) + sizeof(((xc_entry_t*)NULL)->next) + sizeof(((xc_entry_t*)NULL)->cache) + sizeof(((xc_entry_t*)NULL)->size) + sizeof(((xc_entry_t*)NULL)->refcount) + sizeof(((xc_entry_t*)NULL)->hits) + sizeof(((xc_entry_t*)NULL)->ctime) + sizeof(((xc_entry_t*)NULL)->atime) + sizeof(((xc_entry_t*)NULL)->dtime) + sizeof(((xc_entry_t*)NULL)->ttl) + sizeof(((xc_entry_t*)NULL)->name) + sizeof(((xc_entry_t*)NULL)->data) + sizeof(((xc_entry_t*)NULL)->have_references)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		

		
		
	
	
	
	/* DISPATCH(xc_entry_type_t,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(size_t,size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(xc_hash_value_t,hvalue) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hvalue);
	done_count ++;

#endif



	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cache);
	done_count ++;

#endif


	/* skip */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->next);
	done_count ++;

#endif



	
	
	/* DISPATCH(long,refcount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(time_t,ctime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ctime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,atime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->atime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,dtime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->dtime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(long,ttl) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ttl);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_ulong,hits) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hits);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,name_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_type);
	done_count ++;

#endif



	

#endif
		
	
	

#ifdef IS_UNICODE
		if (src->name_type == IS_UNICODE) {
			
	
	/* DISPATCH(int32_t,name.ustr.len) */

	
	
	
	
	

	

		}
		else {
			
	
	/* DISPATCH(int,name.str.len) */

	
	
	
	
	

	

		}
#else
		
	
	/* DISPATCH(int,name.str.len) */

	
	
	
	
	

	

#endif
		
#ifdef IS_UNICODE
			
	/* PROC_ZSTRING_L(name_type,name.uni.val,name.uni.len) */
 
	
	/* PROC_ZSTRING_N(name_type,name.uni.val,name.uni.len,len) */

#ifdef IS_UNICODE
	
	
	
		if (src->name_type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name.uni.val) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_UNICODE, src->name.uni.val, src->name.uni.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name.uni.val) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name.uni.val, src->name.uni.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name.uni.val) == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, src->name.uni.val, src->name.uni.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#else
			
	/* PROC_STRING_L(name.str.val,name.str.len) */
 
	/* PROC_STRING_N(name.str.val,name.str.len + 1) */
 
	
	
	
	
	
	
	
	if (src->name.str.val == NULL) {
		
		
	}
	else {
		
		xc_calc_string_n(processor, IS_STRING, ZSTR(src->name.str.val), src->name.str.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
);
		
		
		
		
	}
	
	
	
	
	
	

#endif
		
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		
	
	

		switch (src->type) {
		case XC_TYPE_PHP:
			
	
	/* STRUCT_P(xc_entry_data_php_t,data.php) */

	if (src->data.php) {
		
		
	
	/* STRUCT_P_EX(xc_entry_data_php_t,dst->data.php,src->data.php,data.php,) */

	
	

	
	assert(sizeof(xc_entry_data_php_t) == sizeof(( src->data.php)[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(xc_entry_data_php_t)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(xc_entry_data_php_t);
	
	
	
	
	

	
	xc_calc_xc_entry_data_php_t(
		
		processor,  src->data.php
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

			break;
		case XC_TYPE_VAR:
			
	
	/* STRUCT_P(xc_entry_data_var_t,data.var) */

	if (src->data.var) {
		
		
	
	/* STRUCT_P_EX(xc_entry_data_var_t,dst->data.var,src->data.var,data.var,) */

	
	

	
	assert(sizeof(xc_entry_data_var_t) == sizeof(( src->data.var)[0]));
	
	
	
	
	/* allocate */
	
		
#ifndef NDEBUG
		
			xc_stack_push(&processor->allocsizes, (void *) (long) (sizeof(xc_entry_data_var_t)));
			xc_stack_push(&processor->allocsizes, (void *) (long) (__LINE__));
		
#endif

		processor->size = (size_t) ALIGN(processor->size);
		processor->size += sizeof(xc_entry_data_var_t);
	
	
	
	
	

	
	xc_calc_xc_entry_data_var_t(
		
		processor,  src->data.var
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

			break;
		default:
			assert(0);
		}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->data);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,have_references) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_references);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_calc_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_calc_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


  /* {{{ Pre-declare */
 	  	 	void 	xc_store_zval( 		 		 		xc_processor_t *processor, zval *dst, const zval * const src 		 		 		 		TSRMLS_DC 	); 	
 	  	 	static void inline 	xc_store_zval_ptr( 		 		 		xc_processor_t *processor, zval_ptr *dst, const zval_ptr * const src 		 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_store_zend_op_array( 		 		 		xc_processor_t *processor, zend_op_array *dst, const zend_op_array * const src 		 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_store_zend_class_entry( 		 		 		xc_processor_t *processor, zend_class_entry *dst, const zend_class_entry * const src 		 		 		 		TSRMLS_DC 	); 	
#ifdef HAVE_XCACHE_CONSTANT
 	  	 	static void inline 	xc_store_zend_constant( 		 		 		xc_processor_t *processor, zend_constant *dst, const zend_constant * const src 		 		 		 		TSRMLS_DC 	); 	
#endif
 	  	 	void 	xc_store_zend_function( 		 		 		xc_processor_t *processor, zend_function *dst, const zend_function * const src 		 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_store_xc_entry_t( 		 		 		xc_processor_t *processor, xc_entry_t *dst, const xc_entry_t * const src 		 		 		 		TSRMLS_DC 	); 	
#ifdef ZEND_ENGINE_2
 	  	 	static void inline 	xc_store_zend_property_info( 		 		 		xc_processor_t *processor, zend_property_info *dst, const zend_property_info * const src 		 		 		 		TSRMLS_DC 	); 	
#endif
/* }}} */
#ifdef IS_CV

	
	

/* {{{ xc_store_zend_compiled_variable */
	
 	  	 	static void inline 	xc_store_zend_compiled_variable( 		 		 		xc_processor_t *processor, zend_compiled_variable *dst, const zend_compiled_variable * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_compiled_variable*)NULL)->name) + sizeof(((zend_compiled_variable*)NULL)->name_len) + sizeof(((zend_compiled_variable*)NULL)->hash_value)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_compiled_variable));
			do {
		

		
		
	
	
	/* DISPATCH(int,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->name) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->name) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->name));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name));

		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,hash_value) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hash_value);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_store_zend_uint */
	
 	  	 	static void inline 	xc_store_zend_uint( 		 		 		xc_processor_t *processor, zend_uint *dst, const zend_uint * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_uint), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_uint));
			do {
		

		
		
	dst[0] = src[0];
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifndef ZEND_ENGINE_2

	
	

/* {{{ xc_store_int */
	
 	  	 	static void inline 	xc_store_int( 		 		 		xc_processor_t *processor, int *dst, const int * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(int), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(int));
			do {
		

		
		
	*dst = *src;
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_store_zend_try_catch_element */
	
 	  	 	static void inline 	xc_store_zend_try_catch_element( 		 		 		xc_processor_t *processor, zend_try_catch_element *dst, const zend_try_catch_element * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_try_catch_element*)NULL)->try_op) + sizeof(((zend_try_catch_element*)NULL)->catch_op)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_try_catch_element));
			do {
		

		
		
	
	
	/* DISPATCH(zend_uint,try_op) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_op);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,catch_op) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->catch_op);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif /* ifdef ZEND_ENGINE_2 */

	
	

/* {{{ xc_store_zend_brk_cont_element */
	
 	  	 	static void inline 	xc_store_zend_brk_cont_element( 		 		 		xc_processor_t *processor, zend_brk_cont_element *dst, const zend_brk_cont_element * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_brk_cont_element*)NULL)->start) + sizeof(((zend_brk_cont_element*)NULL)->cont) + sizeof(((zend_brk_cont_element*)NULL)->brk) + sizeof(((zend_brk_cont_element*)NULL)->parent)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_brk_cont_element));
			do {
		

		
		
#ifdef ZEND_ENGINE_2_2
#ifndef IS_UNICODE
	
	
	/* DISPATCH(int,start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start);
	done_count ++;

#endif



	

#endif
#endif
	
	
	/* DISPATCH(int,cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,brk) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,parent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	
	

/* {{{ xc_store_HashTable_zval_ptr */
	
 	  	 	static void inline 	xc_store_HashTable_zval_ptr( 		 		 		xc_processor_t *processor, HashTable *dst, const HashTable * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(HashTable));
			do {
		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				uint n;
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		dst->pInternalPointer = NULL;	/* Used for element traversal */ 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		dst->pListHead = NULL; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->arBuckets',`Bucket*',`src->nTableSize',`1' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(Bucket*) * src->nTableSize;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->arBuckets',`Bucket*',`src->nTableSize',`1' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->arBuckets = (Bucket* *) (processor->p = (char *) ALIGN(processor->p));
		
				memset(dst->arBuckets, 0, sizeof(Bucket*) * src->nTableSize);
		
		processor->p += sizeof(Bucket*) * src->nTableSize;

		
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `pnew',`char',`bucketsize',`',`Bucket' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(char) * bucketsize;
				if (expect != real) {
					fprintf(stderr, "mismatch `pnew',`char',`bucketsize',`',`Bucket' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		pnew = (Bucket *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(pnew, -1, sizeof(char) * bucketsize);
#endif

			
		processor->p += sizeof(char) * bucketsize;

		
	
	
	
	

			memcpy(pnew, b, bucketsize);
			
				n = b->h & src->nTableMask;
				/* pnew into hash node chain */
				pnew->pLast = NULL;
				if (dst->arBuckets[n]) {
					pnew->pNext = dst->arBuckets[n];
					pnew->pNext->pLast = pnew;
				}
				else {
					pnew->pNext = NULL;
				}
				dst->arBuckets[n] = pnew;
			
			
			if (sizeof(void *) == sizeof(zval_ptr)) {
				pnew->pData = &pnew->pDataPtr;
								
	
	/* STRUCT_P_EX(zval_ptr,pnew->pData,((zval_ptr*)b->pData),,, ) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof((  ((zval_ptr*)b->pData))[0]));
	
	
	xc_store_zval_ptr(
		
		
		processor,   pnew->pData,   ((zval_ptr*)b->pData)
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zval_ptr,pnew->pData,((zval_ptr*)b->pData),,) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof(( ((zval_ptr*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `pnew->pData',`zval_ptr' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zval_ptr);
				if (expect != real) {
					fprintf(stderr, "mismatch `pnew->pData',`zval_ptr' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		pnew->pData = (zval_ptr *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(pnew->pData, -1, sizeof(zval_ptr));
#endif

			
		processor->p += sizeof(zval_ptr);

		
	
	
	
	

	
	xc_store_zval_ptr(
		
		
		processor,  pnew->pData,  ((zval_ptr*)b->pData)
		
		
		
		TSRMLS_CC
	);
	
	
	
	pnew->pData = (zval_ptr *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)pnew->pData);


				pnew->pDataPtr = NULL;
			}

			if (first) {
				dst->pListHead = pnew;
				first = 0;
			}

			
				/* flat link */
				pnew->pListLast = prev;
				pnew->pListNext = NULL;
				if (prev) {
					prev->pListNext = pnew;
				}
			
			prev = pnew;
		}
		
	
	

		dst->pListTail = pnew; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		dst->pDestructor = src->pDestructor; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	
	

/* {{{ xc_store_HashTable_zend_constant */
	
 	  	 	static void inline 	xc_store_HashTable_zend_constant( 		 		 		xc_processor_t *processor, HashTable *dst, const HashTable * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(HashTable));
			do {
		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				uint n;
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		dst->pInternalPointer = NULL;	/* Used for element traversal */ 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		dst->pListHead = NULL; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->arBuckets',`Bucket*',`src->nTableSize',`1' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(Bucket*) * src->nTableSize;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->arBuckets',`Bucket*',`src->nTableSize',`1' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->arBuckets = (Bucket* *) (processor->p = (char *) ALIGN(processor->p));
		
				memset(dst->arBuckets, 0, sizeof(Bucket*) * src->nTableSize);
		
		processor->p += sizeof(Bucket*) * src->nTableSize;

		
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `pnew',`char',`bucketsize',`',`Bucket' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(char) * bucketsize;
				if (expect != real) {
					fprintf(stderr, "mismatch `pnew',`char',`bucketsize',`',`Bucket' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		pnew = (Bucket *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(pnew, -1, sizeof(char) * bucketsize);
#endif

			
		processor->p += sizeof(char) * bucketsize;

		
	
	
	
	

			memcpy(pnew, b, bucketsize);
			
				n = b->h & src->nTableMask;
				/* pnew into hash node chain */
				pnew->pLast = NULL;
				if (dst->arBuckets[n]) {
					pnew->pNext = dst->arBuckets[n];
					pnew->pNext->pLast = pnew;
				}
				else {
					pnew->pNext = NULL;
				}
				dst->arBuckets[n] = pnew;
			
			
			if (sizeof(void *) == sizeof(zend_constant)) {
				pnew->pData = &pnew->pDataPtr;
								
	
	/* STRUCT_P_EX(zend_constant,pnew->pData,((zend_constant*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_constant) == sizeof((  ((zend_constant*)b->pData))[0]));
	
	
	xc_store_zend_constant(
		
		
		processor,   pnew->pData,   ((zend_constant*)b->pData)
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_constant,pnew->pData,((zend_constant*)b->pData),,) */

	
	

	
	assert(sizeof(zend_constant) == sizeof(( ((zend_constant*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `pnew->pData',`zend_constant' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_constant);
				if (expect != real) {
					fprintf(stderr, "mismatch `pnew->pData',`zend_constant' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		pnew->pData = (zend_constant *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(pnew->pData, -1, sizeof(zend_constant));
#endif

			
		processor->p += sizeof(zend_constant);

		
	
	
	
	

	
	xc_store_zend_constant(
		
		
		processor,  pnew->pData,  ((zend_constant*)b->pData)
		
		
		
		TSRMLS_CC
	);
	
	
	
	pnew->pData = (zend_constant *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)pnew->pData);


				pnew->pDataPtr = NULL;
			}

			if (first) {
				dst->pListHead = pnew;
				first = 0;
			}

			
				/* flat link */
				pnew->pListLast = prev;
				pnew->pListNext = NULL;
				if (prev) {
					prev->pListNext = pnew;
				}
			
			prev = pnew;
		}
		
	
	

		dst->pListTail = pnew; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		dst->pDestructor = src->pDestructor; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	
	

/* {{{ xc_store_HashTable_zend_function */
	
		/* export:  	  	 	void 	xc_store_HashTable_zend_function( 		 		 		xc_processor_t *processor, HashTable *dst, const HashTable * const src 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_store_HashTable_zend_function( 		 		 		xc_processor_t *processor, HashTable *dst, const HashTable * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(HashTable));
			do {
		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				uint n;
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		dst->pInternalPointer = NULL;	/* Used for element traversal */ 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		dst->pListHead = NULL; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->arBuckets',`Bucket*',`src->nTableSize',`1' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(Bucket*) * src->nTableSize;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->arBuckets',`Bucket*',`src->nTableSize',`1' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->arBuckets = (Bucket* *) (processor->p = (char *) ALIGN(processor->p));
		
				memset(dst->arBuckets, 0, sizeof(Bucket*) * src->nTableSize);
		
		processor->p += sizeof(Bucket*) * src->nTableSize;

		
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `pnew',`char',`bucketsize',`',`Bucket' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(char) * bucketsize;
				if (expect != real) {
					fprintf(stderr, "mismatch `pnew',`char',`bucketsize',`',`Bucket' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		pnew = (Bucket *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(pnew, -1, sizeof(char) * bucketsize);
#endif

			
		processor->p += sizeof(char) * bucketsize;

		
	
	
	
	

			memcpy(pnew, b, bucketsize);
			
				n = b->h & src->nTableMask;
				/* pnew into hash node chain */
				pnew->pLast = NULL;
				if (dst->arBuckets[n]) {
					pnew->pNext = dst->arBuckets[n];
					pnew->pNext->pLast = pnew;
				}
				else {
					pnew->pNext = NULL;
				}
				dst->arBuckets[n] = pnew;
			
			
			if (sizeof(void *) == sizeof(zend_function)) {
				pnew->pData = &pnew->pDataPtr;
								
	
	/* STRUCT_P_EX(zend_function,pnew->pData,((zend_function*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_function) == sizeof((  ((zend_function*)b->pData))[0]));
	
	
	xc_store_zend_function(
		
		
		processor,   pnew->pData,   ((zend_function*)b->pData)
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_function,pnew->pData,((zend_function*)b->pData),,) */

	
	

	
	assert(sizeof(zend_function) == sizeof(( ((zend_function*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `pnew->pData',`zend_function' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_function);
				if (expect != real) {
					fprintf(stderr, "mismatch `pnew->pData',`zend_function' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		pnew->pData = (zend_function *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(pnew->pData, -1, sizeof(zend_function));
#endif

			
		processor->p += sizeof(zend_function);

		
	
	
	
	

	
	xc_store_zend_function(
		
		
		processor,  pnew->pData,  ((zend_function*)b->pData)
		
		
		
		TSRMLS_CC
	);
	
	
	
	pnew->pData = (zend_function *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)pnew->pData);


				pnew->pDataPtr = NULL;
			}

			if (first) {
				dst->pListHead = pnew;
				first = 0;
			}

			
				/* flat link */
				pnew->pListLast = prev;
				pnew->pListNext = NULL;
				if (prev) {
					prev->pListNext = pnew;
				}
			
			prev = pnew;
		}
		
	
	

		dst->pListTail = pnew; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		dst->pDestructor = src->pDestructor; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef ZEND_ENGINE_2

	
	
	

/* {{{ xc_store_HashTable_zend_property_info */
	
 	  	 	static void inline 	xc_store_HashTable_zend_property_info( 		 		 		xc_processor_t *processor, HashTable *dst, const HashTable * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(HashTable));
			do {
		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				uint n;
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		dst->pInternalPointer = NULL;	/* Used for element traversal */ 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		dst->pListHead = NULL; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->arBuckets',`Bucket*',`src->nTableSize',`1' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(Bucket*) * src->nTableSize;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->arBuckets',`Bucket*',`src->nTableSize',`1' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->arBuckets = (Bucket* *) (processor->p = (char *) ALIGN(processor->p));
		
				memset(dst->arBuckets, 0, sizeof(Bucket*) * src->nTableSize);
		
		processor->p += sizeof(Bucket*) * src->nTableSize;

		
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `pnew',`char',`bucketsize',`',`Bucket' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(char) * bucketsize;
				if (expect != real) {
					fprintf(stderr, "mismatch `pnew',`char',`bucketsize',`',`Bucket' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		pnew = (Bucket *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(pnew, -1, sizeof(char) * bucketsize);
#endif

			
		processor->p += sizeof(char) * bucketsize;

		
	
	
	
	

			memcpy(pnew, b, bucketsize);
			
				n = b->h & src->nTableMask;
				/* pnew into hash node chain */
				pnew->pLast = NULL;
				if (dst->arBuckets[n]) {
					pnew->pNext = dst->arBuckets[n];
					pnew->pNext->pLast = pnew;
				}
				else {
					pnew->pNext = NULL;
				}
				dst->arBuckets[n] = pnew;
			
			
			if (sizeof(void *) == sizeof(zend_property_info)) {
				pnew->pData = &pnew->pDataPtr;
								
	
	/* STRUCT_P_EX(zend_property_info,pnew->pData,((zend_property_info*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_property_info) == sizeof((  ((zend_property_info*)b->pData))[0]));
	
	
	xc_store_zend_property_info(
		
		
		processor,   pnew->pData,   ((zend_property_info*)b->pData)
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_property_info,pnew->pData,((zend_property_info*)b->pData),,) */

	
	

	
	assert(sizeof(zend_property_info) == sizeof(( ((zend_property_info*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `pnew->pData',`zend_property_info' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_property_info);
				if (expect != real) {
					fprintf(stderr, "mismatch `pnew->pData',`zend_property_info' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		pnew->pData = (zend_property_info *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(pnew->pData, -1, sizeof(zend_property_info));
#endif

			
		processor->p += sizeof(zend_property_info);

		
	
	
	
	

	
	xc_store_zend_property_info(
		
		
		processor,  pnew->pData,  ((zend_property_info*)b->pData)
		
		
		
		TSRMLS_CC
	);
	
	
	
	pnew->pData = (zend_property_info *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)pnew->pData);


				pnew->pDataPtr = NULL;
			}

			if (first) {
				dst->pListHead = pnew;
				first = 0;
			}

			
				/* flat link */
				pnew->pListLast = prev;
				pnew->pListNext = NULL;
				if (prev) {
					prev->pListNext = pnew;
				}
			
			prev = pnew;
		}
		
	
	

		dst->pListTail = pnew; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		dst->pDestructor = src->pDestructor; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	

/* {{{ xc_store_zval */
	
		/* export:  	  	 	void 	xc_store_zval( 		 		 		xc_processor_t *processor, zval *dst, const zval * const src 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_store_zval( 		 		 		xc_processor_t *processor, zval *dst, const zval * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zval*)NULL)->value) + sizeof(((zval*)NULL)->refcount) + sizeof(((zval*)NULL)->type) + sizeof(((zval*)NULL)->is_ref)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zval));
			do {
		

		
			
				/* Variable information */
		
	
	

		switch (src->type & ~IS_CONSTANT_INDEX) {
			case IS_LONG:
			case IS_RESOURCE:
			case IS_BOOL:
				
	
	/* DISPATCH(long,value.lval) */

	
	
	
	
	

	

				break;
			case IS_DOUBLE:
				
	
	/* DISPATCH(double,value.dval) */

	
	
	
	
	

	

				break;
			case IS_NULL:
				
				break;

			case IS_CONSTANT:
#ifdef IS_UNICODE
				if (UG(unicode)) {
					goto proc_unicode;
				}
#endif
			case IS_STRING:
#ifdef FLAG_IS_BC
			case FLAG_IS_BC:
#endif
				
	
	/* DISPATCH(int,value.str.len) */

	
	
	
	
	

	

				
	/* PROC_STRING_L(value.str.val,value.str.len) */
 
	/* PROC_STRING_N(value.str.val,value.str.len + 1) */
 
	
	
	
	
	
	
	
	if (src->value.str.val == NULL) {
		
		
	}
	else {
		
		
		dst->value.str.val = ZSTR_S(xc_store_string_n(processor, IS_STRING, ZSTR(src->value.str.val), src->value.str.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	dst->value.str.val = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->value.str.val);

		
	}
	
	
	
	
	
	

				break;
#ifdef IS_UNICODE
			case IS_UNICODE:
proc_unicode:
				
	
	/* DISPATCH(int32_t,value.uni.len) */

	
	
	
	
	

	

				
	/* PROC_ZSTRING_L(1,value.uni.val,value.uni.len) */
 
	
	/* PROC_ZSTRING_N(1,value.uni.val,value.uni.len,len) */

#ifdef IS_UNICODE
	
	
	
	
	
	
	
	
	
	
	if (ZSTR_U(src->value.uni.val) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->value.uni.val) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->value.uni.val, src->value.uni.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->value.uni.val) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->value.uni.val));

		
	}
	
	
	
	
	
	

	
#else
	
	
	
	
	
	
	
	
	
	if (ZSTR_S(src->value.uni.val) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->value.uni.val) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->value.uni.val, src->value.uni.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->value.uni.val) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->value.uni.val));

		
	}
	
	
	
	
	
	

#endif
	

				break;
#endif

			case IS_ARRAY:
			case IS_CONSTANT_ARRAY:
				
	
	/* STRUCT_P(HashTable,value.ht,HashTable_zval_ptr) */

	if (src->value.ht) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->value.ht,src->value.ht,value.ht,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->value.ht)[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->value.ht',`HashTable' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(HashTable);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->value.ht',`HashTable' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->value.ht = (HashTable *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->value.ht, -1, sizeof(HashTable));
#endif

			
		processor->p += sizeof(HashTable);

		
	
	
	
	

	
	xc_store_HashTable_zval_ptr(
		
		
		processor,  dst->value.ht,  src->value.ht
		
		
		
		TSRMLS_CC
	);
	
	
	
	dst->value.ht = (HashTable *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->value.ht);


	}
	else {
		
		
	
	

	}
	

				break;

			case IS_OBJECT:
				
				#ifndef ZEND_ENGINE_2
				
	
	/* STRUCT_P(zend_class_entry,value.obj.ce) */

	if (src->value.obj.ce) {
		
		
	
	/* STRUCT_P_EX(zend_class_entry,dst->value.obj.ce,src->value.obj.ce,value.obj.ce,) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof(( src->value.obj.ce)[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->value.obj.ce',`zend_class_entry' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_class_entry);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->value.obj.ce',`zend_class_entry' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->value.obj.ce = (zend_class_entry *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->value.obj.ce, -1, sizeof(zend_class_entry));
#endif

			
		processor->p += sizeof(zend_class_entry);

		
	
	
	
	

	
	xc_store_zend_class_entry(
		
		
		processor,  dst->value.obj.ce,  src->value.obj.ce
		
		
		
		TSRMLS_CC
	);
	
	
	
	dst->value.obj.ce = (zend_class_entry *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->value.obj.ce);


	}
	else {
		
		
	
	

	}
	

				
	
	/* STRUCT_P(HashTable,value.obj.properties,HashTable_zval_ptr) */

	if (src->value.obj.properties) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->value.obj.properties,src->value.obj.properties,value.obj.properties,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->value.obj.properties)[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->value.obj.properties',`HashTable' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(HashTable);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->value.obj.properties',`HashTable' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->value.obj.properties = (HashTable *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->value.obj.properties, -1, sizeof(HashTable));
#endif

			
		processor->p += sizeof(HashTable);

		
	
	
	
	

	
	xc_store_HashTable_zval_ptr(
		
		
		processor,  dst->value.obj.properties,  src->value.obj.properties
		
		
		
		TSRMLS_CC
	);
	
	
	
	dst->value.obj.properties = (HashTable *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->value.obj.properties);


	}
	else {
		
		
	
	

	}
	

#endif
				break;

			default:
				assert(0);
		}
		
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif


		
	
	/* DISPATCH(zval_data_type,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_uchar,is_ref) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->is_ref);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_ushort,refcount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_3
		
	
	/* DISPATCH(zend_uchar,idx_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->idx_type);
	done_count ++;

#endif



	

#endif
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_store_zval_ptr */
	
 	  	 	static void inline 	xc_store_zval_ptr( 		 		 		xc_processor_t *processor, zval_ptr *dst, const zval_ptr * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zval_ptr), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zval_ptr));
			do {
		

		
			
		do {
			
				if (processor->reference) {
					zval_ptr *ppzv;
					if (zend_hash_find(&processor->zvalptrs, (char *) &src[0], sizeof(src[0]), (void **) &ppzv) == SUCCESS) {
						
							dst[0] = *ppzv;
							/* *dst is updated */
													
						processor->have_references = 1;
						assert(xc_is_shm(dst[0]));
						
						break;
					}
				}
			
			
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst[0]',`zval' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zval);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst[0]',`zval' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst[0] = (zval *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst[0], -1, sizeof(zval));
#endif

			
		processor->p += sizeof(zval);

		
	
	
	
	

			
				if (processor->reference) {
					
						zval_ptr pzv = dst[0];
						
	pzv = (zval *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)pzv);

					
					if (zend_hash_add(&processor->zvalptrs, (char *) &src[0], sizeof(src[0]), (void *) &pzv, sizeof(pzv), NULL) == SUCCESS) {
						/* first add, go on */
											}
					else {
						assert(0);
					}
				}
			
			
							
			
			
	
	/* STRUCT_P_EX(zval,dst[0],src[0],[0],, ) */

	
	

	
	assert(sizeof(zval) == sizeof((  src[0])[0]));
	
	
	xc_store_zval(
		
		
		processor,   dst[0],   src[0]
		
		
		
		TSRMLS_CC
	);
	
	
	

			
	dst[0] = (zval *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst[0]);

		} while (0);
	
	
#ifndef NDEBUG
		
	done_size += sizeof(zval_ptr);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_store_zend_arg_info */
	
 	  	 	static void inline 	xc_store_zend_arg_info( 		 		 		xc_processor_t *processor, zend_arg_info *dst, const zend_arg_info * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_arg_info*)NULL)->name) + sizeof(((zend_arg_info*)NULL)->name_len) + sizeof(((zend_arg_info*)NULL)->class_name) + sizeof(((zend_arg_info*)NULL)->class_name_len) + sizeof(((zend_arg_info*)NULL)->array_type_hint) + sizeof(((zend_arg_info*)NULL)->allow_null) + sizeof(((zend_arg_info*)NULL)->pass_by_reference) + sizeof(((zend_arg_info*)NULL)->return_reference) + sizeof(((zend_arg_info*)NULL)->required_num_args)  ), assert_count = 9;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_arg_info));
			do {
		

		
		
	
	
	/* DISPATCH(zend_uint,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->name) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->name) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->name));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name));

		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_uint,class_name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,class_name,class_name_len) */
 
	
	/* PROC_ZSTRING_N(,class_name,class_name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->class_name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->class_name) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->class_name, src->class_name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->class_name) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->class_name));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->class_name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->class_name, src->class_name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->class_name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->class_name));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->class_name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->class_name, src->class_name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->class_name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->class_name));

		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_bool,array_type_hint) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->array_type_hint);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,allow_null) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->allow_null);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_by_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_by_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,return_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,required_num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_store_zend_constant */
	
 	  	 	static void inline 	xc_store_zend_constant( 		 		 		xc_processor_t *processor, zend_constant *dst, const zend_constant * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_constant*)NULL)->value) + sizeof(((zend_constant*)NULL)->flags) + sizeof(((zend_constant*)NULL)->name) + sizeof(((zend_constant*)NULL)->name_len) + sizeof(((zend_constant*)NULL)->module_number)  ), assert_count = 5;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_constant));
			do {
		

		
			
	
	/* STRUCT(zval,value) */

	assert(sizeof(zval) == sizeof(src->value));
	
	
	
	/* STRUCT_P_EX(zval,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->value)[0]));
	
	
	xc_store_zval(
		
		
		processor, & dst->value, & src->value
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



	
	
	/* DISPATCH(int,flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	
	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->name) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->name) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->name));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name, src->name_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name));

		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	
	/* DISPATCH(int,module_number) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module_number);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_store_zend_function */
	
		/* export:  	  	 	void 	xc_store_zend_function( 		 		 		xc_processor_t *processor, zend_function *dst, const zend_function * const src 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_store_zend_function( 		 		 		xc_processor_t *processor, zend_function *dst, const zend_function * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_function), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_function));
			do {
		

		
			
	
	

	switch (src->type) {
	case ZEND_INTERNAL_FUNCTION:
	case ZEND_OVERLOADED_FUNCTION:
		
		break;

	case ZEND_USER_FUNCTION:
	case ZEND_EVAL_CODE:
		
		
	
	/* STRUCT(zend_op_array,op_array) */

	assert(sizeof(zend_op_array) == sizeof(src->op_array));
	
	
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,,&) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof((& src->op_array)[0]));
	
	
	xc_store_zend_op_array(
		
		
		processor, & dst->op_array, & src->op_array
		
		
		
		TSRMLS_CC
	);
	
	
	

	

		break;

	default:
		assert(0);
	}
	
	
	

	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_store_zend_property_info */
	
 	  	 	static void inline 	xc_store_zend_property_info( 		 		 		xc_processor_t *processor, zend_property_info *dst, const zend_property_info * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_property_info*)NULL)->flags) + sizeof(((zend_property_info*)NULL)->name) + sizeof(((zend_property_info*)NULL)->name_length) + sizeof(((zend_property_info*)NULL)->h) + sizeof(((zend_property_info*)NULL)->doc_comment) + sizeof(((zend_property_info*)NULL)->doc_comment_len) + sizeof(((zend_property_info*)NULL)->ce)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_property_info));
			do {
		

		
		
	
	
	/* DISPATCH(zend_uint,flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,name_length) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->name) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->name) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->name));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name));

		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,h) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->h);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->doc_comment) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->doc_comment) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->doc_comment));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->doc_comment) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->doc_comment) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->doc_comment));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->doc_comment) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->doc_comment) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->doc_comment));

		
	}
	
	
	
	
	
	

#endif
	

#endif
	#if defined(ZEND_ENGINE_2_2) && !defined(IS_UNICODE)
	
	if (src->ce) {
		dst->ce = (zend_class_entry *) xc_get_class_num(processor, src->ce);
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce);
	done_count ++;

#endif


#endif

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_store_zend_class_entry */
	
		/* export:  	  	 	void 	xc_store_zend_class_entry( 		 		 		xc_processor_t *processor, zend_class_entry *dst, const zend_class_entry * const src 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_store_zend_class_entry( 		 		 		xc_processor_t *processor, zend_class_entry *dst, const zend_class_entry * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_class_entry*)NULL)->type) + sizeof(((zend_class_entry*)NULL)->name) + sizeof(((zend_class_entry*)NULL)->name_length) + sizeof(((zend_class_entry*)NULL)->parent) + sizeof(((zend_class_entry*)NULL)->refcount) + sizeof(((zend_class_entry*)NULL)->constants_updated) + sizeof(((zend_class_entry*)NULL)->ce_flags) + sizeof(((zend_class_entry*)NULL)->function_table) + sizeof(((zend_class_entry*)NULL)->default_properties) + sizeof(((zend_class_entry*)NULL)->properties_info) + sizeof(((zend_class_entry*)NULL)->default_static_members) + sizeof(((zend_class_entry*)NULL)->static_members) + sizeof(((zend_class_entry*)NULL)->constants_table) + sizeof(((zend_class_entry*)NULL)->builtin_functions) + sizeof(((zend_class_entry*)NULL)->constructor) + sizeof(((zend_class_entry*)NULL)->destructor) + sizeof(((zend_class_entry*)NULL)->clone) + sizeof(((zend_class_entry*)NULL)->__get) + sizeof(((zend_class_entry*)NULL)->__set) + sizeof(((zend_class_entry*)NULL)->__unset) + sizeof(((zend_class_entry*)NULL)->__isset) + sizeof(((zend_class_entry*)NULL)->__call) + sizeof(((zend_class_entry*)NULL)->__tostring) + sizeof(((zend_class_entry*)NULL)->serialize_func) + sizeof(((zend_class_entry*)NULL)->unserialize_func) + sizeof(((zend_class_entry*)NULL)->iterator_funcs) + sizeof(((zend_class_entry*)NULL)->create_object) + sizeof(((zend_class_entry*)NULL)->get_iterator) + sizeof(((zend_class_entry*)NULL)->interface_gets_implemented) + sizeof(((zend_class_entry*)NULL)->serialize) + sizeof(((zend_class_entry*)NULL)->unserialize) + sizeof(((zend_class_entry*)NULL)->interfaces) + sizeof(((zend_class_entry*)NULL)->num_interfaces) + sizeof(((zend_class_entry*)NULL)->filename) + sizeof(((zend_class_entry*)NULL)->line_start) + sizeof(((zend_class_entry*)NULL)->line_end) + sizeof(((zend_class_entry*)NULL)->doc_comment) + sizeof(((zend_class_entry*)NULL)->doc_comment_len) + sizeof(((zend_class_entry*)NULL)->module)  ), assert_count = 39;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_class_entry));
			do {
		

		
			
		processor->active_class_entry_src = src;
		processor->active_class_entry_dst = dst;
	
	
	
	/* DISPATCH(char,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,name_length) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->name) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->name) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->name));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name, src->name_length + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name));

		
	}
	
	
	
	
	
	

#endif
	

	
		
	if (src->parent) {
		dst->parent = (zend_class_entry *) xc_get_class_num(processor, src->parent);
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif


	
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(int,refcount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	

#else
	
	
	/* STRUCT_P(int,refcount) */

	if (src->refcount) {
		
		
	
	/* STRUCT_P_EX(int,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(int) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->refcount',`int' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(int);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->refcount',`int' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->refcount = (int *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->refcount, -1, sizeof(int));
#endif

			
		processor->p += sizeof(int);

		
	
	
	
	

	
	xc_store_int(
		
		
		processor,  dst->refcount,  src->refcount
		
		
		
		TSRMLS_CC
	);
	
	
	
	dst->refcount = (int *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->refcount);


	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(zend_bool,constants_updated) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_updated);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,ce_flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce_flags);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT(HashTable,default_properties,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_properties));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->default_properties,src->default_properties,default_properties,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_properties)[0]));
	
	
	xc_store_HashTable_zval_ptr(
		
		
		processor, & dst->default_properties, & src->default_properties
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_properties);
	done_count ++;

#endif



	dst->builtin_functions = src->builtin_functions;
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->builtin_functions);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT(HashTable,properties_info,HashTable_zend_property_info) */

	assert(sizeof(HashTable) == sizeof(src->properties_info));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->properties_info,src->properties_info,properties_info,HashTable_zend_property_info,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->properties_info)[0]));
	
	
	xc_store_HashTable_zend_property_info(
		
		
		processor, & dst->properties_info, & src->properties_info
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->properties_info);
	done_count ++;

#endif



#	ifdef ZEND_ENGINE_2_1
	
	
	/* STRUCT(HashTable,default_static_members,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_static_members));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->default_static_members,src->default_static_members,default_static_members,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_static_members)[0]));
	
	
	xc_store_HashTable_zval_ptr(
		
		
		processor, & dst->default_static_members, & src->default_static_members
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_static_members);
	done_count ++;

#endif



	dst->static_members = &dst->default_static_members;
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif


#	else
	
	
	/* STRUCT_P(HashTable,static_members,HashTable_zval_ptr) */

	if (src->static_members) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->static_members,src->static_members,static_members,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_members)[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->static_members',`HashTable' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(HashTable);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->static_members',`HashTable' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->static_members = (HashTable *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->static_members, -1, sizeof(HashTable));
#endif

			
		processor->p += sizeof(HashTable);

		
	
	
	
	

	
	xc_store_HashTable_zval_ptr(
		
		
		processor,  dst->static_members,  src->static_members
		
		
		
		TSRMLS_CC
	);
	
	
	
	dst->static_members = (HashTable *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->static_members);


	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif



#	endif
	
	
	/* STRUCT(HashTable,constants_table,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->constants_table));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->constants_table,src->constants_table,constants_table,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->constants_table)[0]));
	
	
	xc_store_HashTable_zval_ptr(
		
		
		processor, & dst->constants_table, & src->constants_table
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_table);
	done_count ++;

#endif




		
	
	
		
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interfaces);
	done_count ++;

#endif


		
	
	
	
	/* DISPATCH(zend_uint,num_interfaces) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_interfaces);
	done_count ++;

#endif



	


	
	/* PROC_STRING(filename) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->filename == NULL) {
		
		
	}
	else {
		
		
		dst->filename = ZSTR_S(xc_store_string_n(processor, IS_STRING, ZSTR(src->filename), strlen(src->filename) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	dst->filename = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->filename);

		
	}
	
	
	
	
	
	

	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->doc_comment) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->doc_comment) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->doc_comment));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->doc_comment) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->doc_comment) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->doc_comment));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->doc_comment) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->doc_comment) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->doc_comment));

		
	}
	
	
	
	
	
	

#endif
	

#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->iterator_funcs);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->create_object);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_iterator);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interface_gets_implemented);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2_3
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_static_method);
	done_count ++;

#endif


#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize);
	done_count ++;

#endif


	/* deal with it inside xc_fix_method */
	dst->constructor = NULL;
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->destructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->clone);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__set);
	done_count ++;

#endif


/* should be >5.1 */
#ifdef ZEND_ENGINE_2_1
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__unset);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__isset);
	done_count ++;

#endif


# if defined(ZEND_ENGINE_2_2) || PHP_MAJOR_VERSION >= 6
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__tostring);
	done_count ++;

#endif


# endif
#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__call);
	done_count ++;

#endif


#ifdef ZEND_CALLSTATIC_FUNC_NAME
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__callstatic);
	done_count ++;

#endif


#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module);
	done_count ++;

#endif


#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_function_call);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_set);
	done_count ++;

#endif


#endif
		
	
	/* STRUCT(HashTable,function_table,HashTable_zend_function) */

	assert(sizeof(HashTable) == sizeof(src->function_table));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->function_table,src->function_table,function_table,HashTable_zend_function,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->function_table)[0]));
	
	
	xc_store_HashTable_zend_function(
		
		
		processor, & dst->function_table, & src->function_table
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_table);
	done_count ++;

#endif



	
	
		processor->active_class_entry_src = NULL;
		processor->active_class_entry_dst = NULL;
	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_store_znode */
	
 	  	 	static void inline 	xc_store_znode( 		 		 		xc_processor_t *processor, znode *dst, const znode * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((znode*)NULL)->op_type) + sizeof(((znode*)NULL)->u)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(znode));
			do {
		

		
			
	
	/* DISPATCH(int,op_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_type);
	done_count ++;

#endif



	


#ifdef IS_CV
#	define XCACHE_IS_CV IS_CV
#else
/* compatible with zend optimizer */
#	define XCACHE_IS_CV 16
#endif
	assert(src->op_type == IS_CONST ||
		src->op_type == IS_VAR ||
		src->op_type == XCACHE_IS_CV ||
		src->op_type == IS_TMP_VAR ||
		src->op_type == IS_UNUSED);
#undef XCACHE_IS_CV
		
	
	

	switch (src->op_type) {
		case IS_CONST:
			
	
	/* STRUCT(zval,u.constant) */

	assert(sizeof(zval) == sizeof(src->u.constant));
	
	
	
	/* STRUCT_P_EX(zval,dst->u.constant,src->u.constant,u.constant,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->u.constant)[0]));
	
	
	xc_store_zval(
		
		
		processor, & dst->u.constant, & src->u.constant
		
		
		
		TSRMLS_CC
	);
	
	
	

	

			break;
		
			
		
	}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->u);
	done_count ++;

#endif



		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_store_zend_op */
	
		/* export:  	  	 	void 	xc_store_zend_op( 		 		 		xc_processor_t *processor, zend_op *dst, const zend_op * const src 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_store_zend_op( 		 		 		xc_processor_t *processor, zend_op *dst, const zend_op * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op*)NULL)->handler) + sizeof(((zend_op*)NULL)->result) + sizeof(((zend_op*)NULL)->op1) + sizeof(((zend_op*)NULL)->op2) + sizeof(((zend_op*)NULL)->extended_value) + sizeof(((zend_op*)NULL)->lineno) + sizeof(((zend_op*)NULL)->opcode)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_op));
			do {
		

		
			
	
	/* DISPATCH(zend_uchar,opcode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcode);
	done_count ++;

#endif



	

	
	
	/* STRUCT(znode,result) */

	assert(sizeof(znode) == sizeof(src->result));
	
	
	
	/* STRUCT_P_EX(znode,dst->result,src->result,result,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->result)[0]));
	
	
	xc_store_znode(
		
		
		processor, & dst->result, & src->result
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->result);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op1) */

	assert(sizeof(znode) == sizeof(src->op1));
	
	
	
	/* STRUCT_P_EX(znode,dst->op1,src->op1,op1,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op1)[0]));
	
	
	xc_store_znode(
		
		
		processor, & dst->op1, & src->op1
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op1);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op2) */

	assert(sizeof(znode) == sizeof(src->op2));
	
	
	
	/* STRUCT_P_EX(znode,dst->op2,src->op2,op2,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op2)[0]));
	
	
	xc_store_znode(
		
		
		processor, & dst->op2, & src->op2
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op2);
	done_count ++;

#endif



	
	
	/* DISPATCH(ulong,extended_value) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->extended_value);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,lineno) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->lineno);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
		switch (src->opcode) {
			case ZEND_JMP:
				dst->op1.u.jmp_addr = processor->active_opcodes_dst + (src->op1.u.jmp_addr - processor->active_opcodes_src);
				break;

			case ZEND_JMPZ:
			case ZEND_JMPNZ:
			case ZEND_JMPZ_EX:
			case ZEND_JMPNZ_EX:
				dst->op2.u.jmp_addr = processor->active_opcodes_dst + (src->op2.u.jmp_addr - processor->active_opcodes_src);
				break;

			default:
				break;
		}
	
	
	
	/* DISPATCH(opcode_handler_t,handler) */

	/* is copying enough? */
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handler);
	done_count ++;

#endif


	

#endif

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_store_zend_op_array */
	
		/* export:  	  	 	void 	xc_store_zend_op_array( 		 		 		xc_processor_t *processor, zend_op_array *dst, const zend_op_array * const src 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_store_zend_op_array( 		 		 		xc_processor_t *processor, zend_op_array *dst, const zend_op_array * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op_array*)NULL)->type) + sizeof(((zend_op_array*)NULL)->function_name) + sizeof(((zend_op_array*)NULL)->scope) + sizeof(((zend_op_array*)NULL)->fn_flags) + sizeof(((zend_op_array*)NULL)->prototype) + sizeof(((zend_op_array*)NULL)->num_args) + sizeof(((zend_op_array*)NULL)->required_num_args) + sizeof(((zend_op_array*)NULL)->arg_info) + sizeof(((zend_op_array*)NULL)->pass_rest_by_reference) + sizeof(((zend_op_array*)NULL)->return_reference) + sizeof(((zend_op_array*)NULL)->refcount) + sizeof(((zend_op_array*)NULL)->opcodes) + sizeof(((zend_op_array*)NULL)->last) + sizeof(((zend_op_array*)NULL)->size) + sizeof(((zend_op_array*)NULL)->vars) + sizeof(((zend_op_array*)NULL)->last_var) + sizeof(((zend_op_array*)NULL)->size_var) + sizeof(((zend_op_array*)NULL)->T) + sizeof(((zend_op_array*)NULL)->brk_cont_array) + sizeof(((zend_op_array*)NULL)->last_brk_cont) + sizeof(((zend_op_array*)NULL)->current_brk_cont) + sizeof(((zend_op_array*)NULL)->try_catch_array) + sizeof(((zend_op_array*)NULL)->last_try_catch) + sizeof(((zend_op_array*)NULL)->static_variables) + sizeof(((zend_op_array*)NULL)->start_op) + sizeof(((zend_op_array*)NULL)->backpatch_count) + sizeof(((zend_op_array*)NULL)->done_pass_two) + sizeof(((zend_op_array*)NULL)->uses_this) + sizeof(((zend_op_array*)NULL)->filename) + sizeof(((zend_op_array*)NULL)->line_start) + sizeof(((zend_op_array*)NULL)->line_end) + sizeof(((zend_op_array*)NULL)->doc_comment) + sizeof(((zend_op_array*)NULL)->doc_comment_len) + sizeof(((zend_op_array*)NULL)->reserved)  ), assert_count = 34;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_op_array));
			do {
		

		
			
	do {
		zend_uint ii;
	int i;

	/* Common elements */
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING(,function_name) */
 
	
	/* PROC_ZSTRING_N(,function_name,,strlen) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->function_name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->function_name) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->function_name, xc_zstrlen_uchar (src->function_name) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->function_name) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->function_name));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->function_name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->function_name, xc_zstrlen_char (src->function_name) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->function_name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->function_name));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->function_name) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->function_name, xc_zstrlen_char (src->function_name) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->function_name) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->function_name));

		
	}
	
	
	
	
	
	

#endif
	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,fn_flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->fn_flags);
	done_count ++;

#endif



	

	


	if (src->arg_info) {
		
	

		
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->arg_info',`zend_arg_info',`src->num_args' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_arg_info) * src->num_args;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->arg_info',`zend_arg_info',`src->num_args' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->arg_info = (zend_arg_info *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->arg_info, -1, sizeof(zend_arg_info) * src->num_args);
#endif

			
		processor->p += sizeof(zend_arg_info) * src->num_args;

		
	
	
	
	

			
			for (ii = 0; ii < src->num_args; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_arg_info,arg_info[i],) */

	assert(sizeof(zend_arg_info) == sizeof(src->arg_info[ii]));
	
	
	
	/* STRUCT_P_EX(zend_arg_info,dst->arg_info[i],src->arg_info[i],arg_info[i],,&) */

	
	

	
	assert(sizeof(zend_arg_info) == sizeof((& src->arg_info[ii])[0]));
	
	
	xc_store_zend_arg_info(
		
		
		processor, & dst->arg_info[ii], & src->arg_info[ii]
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,required_num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_rest_by_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_rest_by_reference);
	done_count ++;

#endif



	

#else
	if (src->arg_types) {
		
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->arg_types',`zend_uchar',`src->arg_types[0] + 1' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_uchar) * src->arg_types[0] + 1;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->arg_types',`zend_uchar',`src->arg_types[0] + 1' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->arg_types = (zend_uchar *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->arg_types, -1, sizeof(zend_uchar) * src->arg_types[0] + 1);
#endif

			
		processor->p += sizeof(zend_uchar) * src->arg_types[0] + 1;

		
	
	
	
	

		memcpy(dst->arg_types, src->arg_types, sizeof(src->arg_types[0]) * (src->arg_types[0]+1));
		
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif


	}
	else {
		
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif



		
	}
#endif
	
	
	/* DISPATCH(unsigned char,return_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	/* END of common elements */
#ifdef IS_UNICODE
	#endif

	
	
	/* STRUCT_P(zend_uint,refcount) */

	if (src->refcount) {
		
		
	
	/* STRUCT_P_EX(zend_uint,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(zend_uint) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->refcount',`zend_uint' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_uint);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->refcount',`zend_uint' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->refcount = (zend_uint *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->refcount, -1, sizeof(zend_uint));
#endif

			
		processor->p += sizeof(zend_uint);

		
	
	
	
	

	
	xc_store_zend_uint(
		
		
		processor,  dst->refcount,  src->refcount
		
		
		
		TSRMLS_CC
	);
	
	
	
	dst->refcount = (zend_uint *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->refcount);


	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	
	dst->refcount = (zend_uint *) processor->xce_src->cache->shm->handlers->to_readwrite(processor->xce_src->cache->shm, (char *)dst->refcount);

	dst->refcount[0] = 1;

	
	


	if (src->opcodes) {
		
	

		
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->opcodes',`zend_op',`src->last' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_op) * src->last;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->opcodes',`zend_op',`src->last' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->opcodes = (zend_op *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->opcodes, -1, sizeof(zend_op) * src->last);
#endif

			
		processor->p += sizeof(zend_op) * src->last;

		
	
	
	
	

			
		processor->active_opcodes_dst = dst->opcodes;
		processor->active_opcodes_src = src->opcodes;
	
			for (ii = 0; ii < src->last; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_op,opcodes[i],) */

	assert(sizeof(zend_op) == sizeof(src->opcodes[ii]));
	
	
	
	/* STRUCT_P_EX(zend_op,dst->opcodes[i],src->opcodes[i],opcodes[i],,&) */

	
	

	
	assert(sizeof(zend_op) == sizeof((& src->opcodes[ii])[0]));
	
	
	xc_store_zend_op(
		
		
		processor, & dst->opcodes[ii], & src->opcodes[ii]
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif



		
		processor->active_opcodes_dst = dst->opcodes;
		processor->active_opcodes_src = src->opcodes;
	
	}



	
	
	
	/* DISPATCH(zend_uint,last) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last);
	done_count ++;

#endif



	

	dst->size = src->last;
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



#ifdef IS_CV
	
	if (src->vars) {
		
	

		
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->vars',`zend_compiled_variable',`src->last_var' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_compiled_variable) * src->last_var;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->vars',`zend_compiled_variable',`src->last_var' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->vars = (zend_compiled_variable *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->vars, -1, sizeof(zend_compiled_variable) * src->last_var);
#endif

			
		processor->p += sizeof(zend_compiled_variable) * src->last_var;

		
	
	
	
	

			
			for (i = 0; i < src->last_var; i ++) {
				
	
	

					
					
	
	/* STRUCT(zend_compiled_variable,vars[i],) */

	assert(sizeof(zend_compiled_variable) == sizeof(src->vars[i]));
	
	
	
	/* STRUCT_P_EX(zend_compiled_variable,dst->vars[i],src->vars[i],vars[i],,&) */

	
	

	
	assert(sizeof(zend_compiled_variable) == sizeof((& src->vars[i])[0]));
	
	
	xc_store_zend_compiled_variable(
		
		
		processor, & dst->vars[i], & src->vars[i]
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_var) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_var);
	done_count ++;

#endif



	

	dst->size_var = src->last_var;
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size_var);
	done_count ++;

#endif


#else
		
#endif

	
	
	/* DISPATCH(zend_uint,T) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->T);
	done_count ++;

#endif



	


	


	if (src->brk_cont_array) {
		
	

		
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->brk_cont_array',`zend_brk_cont_element',`src->last_brk_cont' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_brk_cont_element) * src->last_brk_cont;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->brk_cont_array',`zend_brk_cont_element',`src->last_brk_cont' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->brk_cont_array = (zend_brk_cont_element *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->brk_cont_array, -1, sizeof(zend_brk_cont_element) * src->last_brk_cont);
#endif

			
		processor->p += sizeof(zend_brk_cont_element) * src->last_brk_cont;

		
	
	
	
	

			
			for (ii = 0; ii < src->last_brk_cont; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_brk_cont_element,brk_cont_array[i],) */

	assert(sizeof(zend_brk_cont_element) == sizeof(src->brk_cont_array[ii]));
	
	
	
	/* STRUCT_P_EX(zend_brk_cont_element,dst->brk_cont_array[i],src->brk_cont_array[i],brk_cont_array[i],,&) */

	
	

	
	assert(sizeof(zend_brk_cont_element) == sizeof((& src->brk_cont_array[ii])[0]));
	
	
	xc_store_zend_brk_cont_element(
		
		
		processor, & dst->brk_cont_array[ii], & src->brk_cont_array[ii]
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,last_brk_cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_brk_cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,current_brk_cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->current_brk_cont);
	done_count ++;

#endif



	

#ifndef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_globals) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_globals);
	done_count ++;

#endif



	

#endif

#ifdef ZEND_ENGINE_2
	
	if (src->try_catch_array) {
		
	

		
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->try_catch_array',`zend_try_catch_element',`src->last_try_catch' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_try_catch_element) * src->last_try_catch;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->try_catch_array',`zend_try_catch_element',`src->last_try_catch' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->try_catch_array = (zend_try_catch_element *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->try_catch_array, -1, sizeof(zend_try_catch_element) * src->last_try_catch);
#endif

			
		processor->p += sizeof(zend_try_catch_element) * src->last_try_catch;

		
	
	
	
	

			
			for (i = 0; i < src->last_try_catch; i ++) {
				
	
	

					
					
	
	/* STRUCT(zend_try_catch_element,try_catch_array[i],) */

	assert(sizeof(zend_try_catch_element) == sizeof(src->try_catch_array[i]));
	
	
	
	/* STRUCT_P_EX(zend_try_catch_element,dst->try_catch_array[i],src->try_catch_array[i],try_catch_array[i],,&) */

	
	

	
	assert(sizeof(zend_try_catch_element) == sizeof((& src->try_catch_array[i])[0]));
	
	
	xc_store_zend_try_catch_element(
		
		
		processor, & dst->try_catch_array[i], & src->try_catch_array[i]
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_try_catch) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_try_catch);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT_P(HashTable,static_variables,HashTable_zval_ptr) */

	if (src->static_variables) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->static_variables,src->static_variables,static_variables,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_variables)[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->static_variables',`HashTable' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(HashTable);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->static_variables',`HashTable' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->static_variables = (HashTable *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->static_variables, -1, sizeof(HashTable));
#endif

			
		processor->p += sizeof(HashTable);

		
	
	
	
	

	
	xc_store_HashTable_zval_ptr(
		
		
		processor,  dst->static_variables,  src->static_variables
		
		
		
		TSRMLS_CC
	);
	
	
	
	dst->static_variables = (HashTable *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->static_variables);


	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_variables);
	done_count ++;

#endif




	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start_op);
	done_count ++;

#endif


	
	
	/* DISPATCH(int,backpatch_count) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->backpatch_count);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(zend_bool,done_pass_two) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->done_pass_two);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_this) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_this);
	done_count ++;

#endif



	

#endif

	
	/* PROC_STRING(filename) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->filename == NULL) {
		
		
	}
	else {
		
		
		dst->filename = ZSTR_S(xc_store_string_n(processor, IS_STRING, ZSTR(src->filename), strlen(src->filename) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	dst->filename = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->filename);

		
	}
	
	
	
	
	
	

#ifdef IS_UNICODE
	
		
	/* PROC_STRING(script_encoding) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->script_encoding);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->script_encoding == NULL) {
		
		
	}
	else {
		
		
		dst->script_encoding = ZSTR_S(xc_store_string_n(processor, IS_STRING, ZSTR(src->script_encoding), strlen(src->script_encoding) + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	dst->script_encoding = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->script_encoding);

		
	}
	
	
	
	
	
	

	
#endif
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->doc_comment) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->doc_comment) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->doc_comment));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->doc_comment) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->doc_comment) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->doc_comment));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->doc_comment) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->doc_comment, src->doc_comment_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->doc_comment) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->doc_comment));

		
	}
	
	
	
	
	
	

#endif
	

#endif

	/* reserved */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->reserved);
	done_count ++;

#endif


#if defined(HARDENING_PATCH) && HARDENING_PATCH
	
	
	/* DISPATCH(zend_bool,created_by_eval) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->created_by_eval);
	done_count ++;

#endif



	

#endif
	} while (0);

#ifdef ZEND_ENGINE_2
		dst->prototype = (processor->active_class_entry_src && src->prototype) ? (zend_function *) -1 : NULL; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->prototype);
	done_count ++;

#endif


#endif

	
#ifdef ZEND_ENGINE_2
		
	if (src->scope) {
		dst->scope = (zend_class_entry *) xc_get_class_num(processor, src->scope);
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->scope);
	done_count ++;

#endif


#endif
	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_zend_op_array : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_zend_op_array : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_store_xc_constinfo_t */
	
 	  	 	static void inline 	xc_store_xc_constinfo_t( 		 		 		xc_processor_t *processor, xc_constinfo_t *dst, const xc_constinfo_t * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_constinfo_t*)NULL)->key) + sizeof(((xc_constinfo_t*)NULL)->key_size) + sizeof(((xc_constinfo_t*)NULL)->constant)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_constinfo_t));
			do {
		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->key) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->key) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->key));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->key) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->key) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->key));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->key) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->key) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->key));

		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	/* STRUCT(zend_constant,constant) */

	assert(sizeof(zend_constant) == sizeof(src->constant));
	
	
	
	/* STRUCT_P_EX(zend_constant,dst->constant,src->constant,constant,,&) */

	
	

	
	assert(sizeof(zend_constant) == sizeof((& src->constant)[0]));
	
	
	xc_store_zend_constant(
		
		
		processor, & dst->constant, & src->constant
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constant);
	done_count ++;

#endif




		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_store_xc_funcinfo_t */
	
		/* export:  	  	 	void 	xc_store_xc_funcinfo_t( 		 		 		xc_processor_t *processor, xc_funcinfo_t *dst, const xc_funcinfo_t * const src 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_store_xc_funcinfo_t( 		 		 		xc_processor_t *processor, xc_funcinfo_t *dst, const xc_funcinfo_t * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_funcinfo_t*)NULL)->key) + sizeof(((xc_funcinfo_t*)NULL)->key_size) + sizeof(((xc_funcinfo_t*)NULL)->func)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_funcinfo_t));
			do {
		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->key) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->key) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->key));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->key) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->key) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->key));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->key) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->key) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->key));

		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	/* STRUCT(zend_function,func) */

	assert(sizeof(zend_function) == sizeof(src->func));
	
	
	
	/* STRUCT_P_EX(zend_function,dst->func,src->func,func,,&) */

	
	

	
	assert(sizeof(zend_function) == sizeof((& src->func)[0]));
	
	
	xc_store_zend_function(
		
		
		processor, & dst->func, & src->func
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->func);
	done_count ++;

#endif




		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_store_xc_classinfo_t */
	
		/* export:  	  	 	void 	xc_store_xc_classinfo_t( 		 		 		xc_processor_t *processor, xc_classinfo_t *dst, const xc_classinfo_t * const src 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_store_xc_classinfo_t( 		 		 		xc_processor_t *processor, xc_classinfo_t *dst, const xc_classinfo_t * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_classinfo_t*)NULL)->key) + sizeof(((xc_classinfo_t*)NULL)->key_size) + sizeof(((xc_classinfo_t*)NULL)->cest) + sizeof(((xc_classinfo_t*)NULL)->oplineno)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_classinfo_t));
			do {
		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->key) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->key) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->key));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->key) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->key) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->key));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->key) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->key, src->key_size 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->key) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->key));

		
	}
	
	
	
	
	
	

#endif
	

	
#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT_P(zend_class_entry,cest) */

	if (src->cest) {
		
		
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof(( src->cest)[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->cest',`zend_class_entry' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_class_entry);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->cest',`zend_class_entry' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->cest = (zend_class_entry *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->cest, -1, sizeof(zend_class_entry));
#endif

			
		processor->p += sizeof(zend_class_entry);

		
	
	
	
	

	
	xc_store_zend_class_entry(
		
		
		processor,  dst->cest,  src->cest
		
		
		
		TSRMLS_CC
	);
	
	
	
	dst->cest = (zend_class_entry *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->cest);


	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#else
	
	
	/* STRUCT(zend_class_entry,cest) */

	assert(sizeof(zend_class_entry) == sizeof(src->cest));
	
	
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,,&) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof((& src->cest)[0]));
	
	
	xc_store_zend_class_entry(
		
		
		processor, & dst->cest, & src->cest
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(int,oplineno) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->oplineno);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2_1

	
	

/* {{{ xc_store_xc_autoglobal_t */
	
 	  	 	static void inline 	xc_store_xc_autoglobal_t( 		 		 		xc_processor_t *processor, xc_autoglobal_t *dst, const xc_autoglobal_t * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_autoglobal_t*)NULL)->key) + sizeof(((xc_autoglobal_t*)NULL)->key_len)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_autoglobal_t));
			do {
		

		
			
	
	/* DISPATCH(zend_uint,key_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_len);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	/* PROC_ZSTRING_L(type,key,key_len) */
 
	
	/* PROC_ZSTRING_N(type,key,key_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->key) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->key, src->key_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->key) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->key));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->key) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->key, src->key_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->key) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->key));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->key) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->key, src->key_len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->key) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->key));

		
	}
	
	
	
	
	
	

#endif
	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_store_xc_entry_data_php_t */
	
 	  	 	static void inline 	xc_store_xc_entry_data_php_t( 		 		 		xc_processor_t *processor, xc_entry_data_php_t *dst, const xc_entry_data_php_t * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_php_t*)NULL)->sourcesize) + sizeof(((xc_entry_data_php_t*)NULL)->device) + sizeof(((xc_entry_data_php_t*)NULL)->inode) + sizeof(((xc_entry_data_php_t*)NULL)->mtime) + sizeof(((xc_entry_data_php_t*)NULL)->op_array) + sizeof(((xc_entry_data_php_t*)NULL)->constinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->constinfos) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfos) + sizeof(((xc_entry_data_php_t*)NULL)->classinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->classinfos) + sizeof(((xc_entry_data_php_t*)NULL)->have_early_binding) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobal_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobals)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_entry_data_php_t));
			do {
		

		
			zend_uint i;

#ifdef HAVE_INODE
	
	
	/* DISPATCH(int,device) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->device);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,inode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inode);
	done_count ++;

#endif



	

#endif
	
	
	/* DISPATCH(size_t,sourcesize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->sourcesize);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(time_t,mtime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->mtime);
	done_count ++;

#endif



	


	
	
	/* STRUCT_P(zend_op_array,op_array) */

	if (src->op_array) {
		
		
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof(( src->op_array)[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->op_array',`zend_op_array' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(zend_op_array);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->op_array',`zend_op_array' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->op_array = (zend_op_array *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->op_array, -1, sizeof(zend_op_array));
#endif

			
		processor->p += sizeof(zend_op_array);

		
	
	
	
	

	
	xc_store_zend_op_array(
		
		
		processor,  dst->op_array,  src->op_array
		
		
		
		TSRMLS_CC
	);
	
	
	
	dst->op_array = (zend_op_array *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->op_array);


	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_array);
	done_count ++;

#endif




#ifdef HAVE_XCACHE_CONSTANT
	
	
	/* DISPATCH(zend_uint,constinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->constinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->constinfos',`xc_constinfo_t',`src->constinfo_cnt' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(xc_constinfo_t) * src->constinfo_cnt;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->constinfos',`xc_constinfo_t',`src->constinfo_cnt' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->constinfos = (xc_constinfo_t *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->constinfos, -1, sizeof(xc_constinfo_t) * src->constinfo_cnt);
#endif

			
		processor->p += sizeof(xc_constinfo_t) * src->constinfo_cnt;

		
	
	
	
	

			
			for (i = 0; i < src->constinfo_cnt; i ++) {
				
	
	

					
					
	
	/* STRUCT(xc_constinfo_t,constinfos[i],) */

	assert(sizeof(xc_constinfo_t) == sizeof(src->constinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_constinfo_t,dst->constinfos[i],src->constinfos[i],constinfos[i],,&) */

	
	

	
	assert(sizeof(xc_constinfo_t) == sizeof((& src->constinfos[i])[0]));
	
	
	xc_store_xc_constinfo_t(
		
		
		processor, & dst->constinfos[i], & src->constinfos[i]
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif



		
	}

#endif

	
	
	/* DISPATCH(zend_uint,funcinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->funcinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->funcinfos',`xc_funcinfo_t',`src->funcinfo_cnt' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(xc_funcinfo_t) * src->funcinfo_cnt;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->funcinfos',`xc_funcinfo_t',`src->funcinfo_cnt' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->funcinfos = (xc_funcinfo_t *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->funcinfos, -1, sizeof(xc_funcinfo_t) * src->funcinfo_cnt);
#endif

			
		processor->p += sizeof(xc_funcinfo_t) * src->funcinfo_cnt;

		
	
	
	
	

			
			for (i = 0; i < src->funcinfo_cnt; i ++) {
				
	
	

					
					
	
	/* STRUCT(xc_funcinfo_t,funcinfos[i],) */

	assert(sizeof(xc_funcinfo_t) == sizeof(src->funcinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_funcinfo_t,dst->funcinfos[i],src->funcinfos[i],funcinfos[i],,&) */

	
	

	
	assert(sizeof(xc_funcinfo_t) == sizeof((& src->funcinfos[i])[0]));
	
	
	xc_store_xc_funcinfo_t(
		
		
		processor, & dst->funcinfos[i], & src->funcinfos[i]
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif



		
	}


	
	
	/* DISPATCH(zend_uint,classinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfo_cnt);
	done_count ++;

#endif



	

	
	
	if (src->classinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->classinfos',`xc_classinfo_t',`src->classinfo_cnt' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(xc_classinfo_t) * src->classinfo_cnt;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->classinfos',`xc_classinfo_t',`src->classinfo_cnt' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->classinfos = (xc_classinfo_t *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->classinfos, -1, sizeof(xc_classinfo_t) * src->classinfo_cnt);
#endif

			
		processor->p += sizeof(xc_classinfo_t) * src->classinfo_cnt;

		
	
	
	
	

			
			for (i = 0; i < src->classinfo_cnt; i ++) {
				
	
	

					
		
			processor->active_class_num = i + 1;
		
	
					
	
	/* STRUCT(xc_classinfo_t,classinfos[i],) */

	assert(sizeof(xc_classinfo_t) == sizeof(src->classinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_classinfo_t,dst->classinfos[i],src->classinfos[i],classinfos[i],,&) */

	
	

	
	assert(sizeof(xc_classinfo_t) == sizeof((& src->classinfos[i])[0]));
	
	
	xc_store_xc_classinfo_t(
		
		
		processor, & dst->classinfos[i], & src->classinfos[i]
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif



		
	}

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,autoglobal_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobal_cnt);
	done_count ++;

#endif



	

	
	if (src->autoglobals) {
		
	

		
			
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->autoglobals',`xc_autoglobal_t',`src->autoglobal_cnt' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(xc_autoglobal_t) * src->autoglobal_cnt;
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->autoglobals',`xc_autoglobal_t',`src->autoglobal_cnt' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->autoglobals = (xc_autoglobal_t *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->autoglobals, -1, sizeof(xc_autoglobal_t) * src->autoglobal_cnt);
#endif

			
		processor->p += sizeof(xc_autoglobal_t) * src->autoglobal_cnt;

		
	
	
	
	

			
			for (i = 0; i < src->autoglobal_cnt; i ++) {
				
	
	

					
		
			processor->active_class_num = i + 1;
		
	
					
	
	/* STRUCT(xc_autoglobal_t,autoglobals[i],) */

	assert(sizeof(xc_autoglobal_t) == sizeof(src->autoglobals[i]));
	
	
	
	/* STRUCT_P_EX(xc_autoglobal_t,dst->autoglobals[i],src->autoglobals[i],autoglobals[i],,&) */

	
	

	
	assert(sizeof(xc_autoglobal_t) == sizeof((& src->autoglobals[i])[0]));
	
	
	xc_store_xc_autoglobal_t(
		
		
		processor, & dst->autoglobals[i], & src->autoglobals[i]
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif



		
	}

#endif
	
	
	/* DISPATCH(zend_bool,have_early_binding) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_early_binding);
	done_count ++;

#endif



	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_store_xc_entry_data_var_t */
	
 	  	 	static void inline 	xc_store_xc_entry_data_var_t( 		 		 		xc_processor_t *processor, xc_entry_data_var_t *dst, const xc_entry_data_var_t * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_var_t*)NULL)->value)  ), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_entry_data_var_t));
			do {
		

		
			
	
	
	/* STRUCT_P_EX(zval_ptr,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof((& src->value)[0]));
	
	
	xc_store_zval_ptr(
		
		
		processor, & dst->value, & src->value
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_store_xc_entry_t */
	
		/* export:  	  	 	void 	xc_store_xc_entry_t( 		 		 		xc_processor_t *processor, xc_entry_t *dst, const xc_entry_t * const src 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_store_xc_entry_t( 		 		 		xc_processor_t *processor, xc_entry_t *dst, const xc_entry_t * const src 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_t*)NULL)->type) + sizeof(((xc_entry_t*)NULL)->hvalue) + sizeof(((xc_entry_t*)NULL)->next) + sizeof(((xc_entry_t*)NULL)->cache) + sizeof(((xc_entry_t*)NULL)->size) + sizeof(((xc_entry_t*)NULL)->refcount) + sizeof(((xc_entry_t*)NULL)->hits) + sizeof(((xc_entry_t*)NULL)->ctime) + sizeof(((xc_entry_t*)NULL)->atime) + sizeof(((xc_entry_t*)NULL)->dtime) + sizeof(((xc_entry_t*)NULL)->ttl) + sizeof(((xc_entry_t*)NULL)->name) + sizeof(((xc_entry_t*)NULL)->data) + sizeof(((xc_entry_t*)NULL)->have_references)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			assert(!xc_is_shm(src));
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_entry_t));
			do {
		

		
		
	
		processor->xce_dst = dst;
		processor->xce_src = src;
	
	
	
	/* DISPATCH(xc_entry_type_t,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(size_t,size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(xc_hash_value_t,hvalue) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hvalue);
	done_count ++;

#endif



	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cache);
	done_count ++;

#endif


	/* skip */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->next);
	done_count ++;

#endif



	dst->refcount = 0; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	
	
	/* DISPATCH(time_t,ctime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ctime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,atime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->atime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,dtime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->dtime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(long,ttl) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ttl);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_ulong,hits) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hits);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,name_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_type);
	done_count ++;

#endif



	

#endif
		
	
	

#ifdef IS_UNICODE
		if (src->name_type == IS_UNICODE) {
			
	
	/* DISPATCH(int32_t,name.ustr.len) */

	
	
	
	
	

	

		}
		else {
			
	
	/* DISPATCH(int,name.str.len) */

	
	
	
	
	

	

		}
#else
		
	
	/* DISPATCH(int,name.str.len) */

	
	
	
	
	

	

#endif
		
#ifdef IS_UNICODE
			
	/* PROC_ZSTRING_L(name_type,name.uni.val,name.uni.len) */
 
	
	/* PROC_ZSTRING_N(name_type,name.uni.val,name.uni.len,len) */

#ifdef IS_UNICODE
	
	
	
		if (src->name_type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name.uni.val) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_U(dst->name.uni.val) = ZSTR_U(xc_store_string_n(processor, IS_UNICODE, src->name.uni.val, src->name.uni.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_U(dst->name.uni.val) = (UChar *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_U(dst->name.uni.val));

		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name.uni.val) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name.uni.val) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name.uni.val, src->name.uni.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name.uni.val) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name.uni.val));

		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name.uni.val) == NULL) {
		
		
	}
	else {
		
		
		ZSTR_S(dst->name.uni.val) = ZSTR_S(xc_store_string_n(processor, IS_STRING, src->name.uni.val, src->name.uni.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	ZSTR_S(dst->name.uni.val) = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)ZSTR_S(dst->name.uni.val));

		
	}
	
	
	
	
	
	

#endif
	

#else
			
	/* PROC_STRING_L(name.str.val,name.str.len) */
 
	/* PROC_STRING_N(name.str.val,name.str.len + 1) */
 
	
	
	
	
	
	
	
	if (src->name.str.val == NULL) {
		
		
	}
	else {
		
		
		dst->name.str.val = ZSTR_S(xc_store_string_n(processor, IS_STRING, ZSTR(src->name.str.val), src->name.str.len + 1 
#ifndef NDEBUG
		, __LINE__
#endif
));
		
		
	dst->name.str.val = (char *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->name.str.val);

		
	}
	
	
	
	
	
	

#endif
		
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		
	
	

		switch (src->type) {
		case XC_TYPE_PHP:
			
	
	/* STRUCT_P(xc_entry_data_php_t,data.php) */

	if (src->data.php) {
		
		
	
	/* STRUCT_P_EX(xc_entry_data_php_t,dst->data.php,src->data.php,data.php,) */

	
	

	
	assert(sizeof(xc_entry_data_php_t) == sizeof(( src->data.php)[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->data.php',`xc_entry_data_php_t' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(xc_entry_data_php_t);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->data.php',`xc_entry_data_php_t' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->data.php = (xc_entry_data_php_t *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->data.php, -1, sizeof(xc_entry_data_php_t));
#endif

			
		processor->p += sizeof(xc_entry_data_php_t);

		
	
	
	
	

	
	xc_store_xc_entry_data_php_t(
		
		
		processor,  dst->data.php,  src->data.php
		
		
		
		TSRMLS_CC
	);
	
	
	
	dst->data.php = (xc_entry_data_php_t *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->data.php);


	}
	else {
		
		
	
	

	}
	

			break;
		case XC_TYPE_VAR:
			
	
	/* STRUCT_P(xc_entry_data_var_t,data.var) */

	if (src->data.var) {
		
		
	
	/* STRUCT_P_EX(xc_entry_data_var_t,dst->data.var,src->data.var,data.var,) */

	
	

	
	assert(sizeof(xc_entry_data_var_t) == sizeof(( src->data.var)[0]));
	
	
	
	
	/* allocate */
	
	
		
#ifndef NDEBUG
		{
			if (!xc_stack_size(&processor->allocsizes)) {
				fprintf(stderr, "mismatch `dst->data.var',`xc_entry_data_var_t' at line %d\n", __LINE__);
			}
			else {
				unsigned long expect = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long atline = (unsigned long) xc_stack_pop(&processor->allocsizes);
				unsigned long real = sizeof(xc_entry_data_var_t);
				if (expect != real) {
					fprintf(stderr, "mismatch `dst->data.var',`xc_entry_data_var_t' at line %d(was %lu): real %lu - expect %lu = %lu\n", __LINE__, atline, real, expect, real - expect);
				}
			}
		}
#endif

		
		dst->data.var = (xc_entry_data_var_t *) (processor->p = (char *) ALIGN(processor->p));
		
				
#ifndef NDEBUG
		memset(dst->data.var, -1, sizeof(xc_entry_data_var_t));
#endif

			
		processor->p += sizeof(xc_entry_data_var_t);

		
	
	
	
	

	
	xc_store_xc_entry_data_var_t(
		
		
		processor,  dst->data.var,  src->data.var
		
		
		
		TSRMLS_CC
	);
	
	
	
	dst->data.var = (xc_entry_data_var_t *) processor->xce_src->cache->shm->handlers->to_readonly(processor->xce_src->cache->shm, (char *)dst->data.var);


	}
	else {
		
		
	
	

	}
	

			break;
		default:
			assert(0);
		}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->data);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,have_references) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_references);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_store_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_store_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


  /* {{{ Pre-declare */
 	  	 	void 	xc_restore_zval( 		 		 		 		xc_processor_t *processor, zval *dst, const zval * const src 		 		 		TSRMLS_DC 	); 	
 	  	 	static void inline 	xc_restore_zval_ptr( 		 		 		 		xc_processor_t *processor, zval_ptr *dst, const zval_ptr * const src 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_restore_zend_op_array( 		 		 		 		xc_processor_t *processor, zend_op_array *dst, const zend_op_array * const src 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_restore_zend_class_entry( 		 		 		 		xc_processor_t *processor, zend_class_entry *dst, const zend_class_entry * const src 		 		 		TSRMLS_DC 	); 	
#ifdef HAVE_XCACHE_CONSTANT
 	  	 	static void inline 	xc_restore_zend_constant( 		 		 		 		xc_processor_t *processor, zend_constant *dst, const zend_constant * const src 		 		 		TSRMLS_DC 	); 	
#endif
 	  	 	void 	xc_restore_zend_function( 		 		 		 		xc_processor_t *processor, zend_function *dst, const zend_function * const src 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_restore_xc_entry_t( 		 		 		 		xc_processor_t *processor, xc_entry_t *dst, const xc_entry_t * const src 		 		 		TSRMLS_DC 	); 	
#ifdef ZEND_ENGINE_2
 	  	 	static void inline 	xc_restore_zend_property_info( 		 		 		 		xc_processor_t *processor, zend_property_info *dst, const zend_property_info * const src 		 		 		TSRMLS_DC 	); 	
#endif
/* }}} */
#ifdef IS_CV

	
	

/* {{{ xc_restore_zend_compiled_variable */
	
 	  	 	static void inline 	xc_restore_zend_compiled_variable( 		 		 		 		xc_processor_t *processor, zend_compiled_variable *dst, const zend_compiled_variable * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_compiled_variable*)NULL)->name) + sizeof(((zend_compiled_variable*)NULL)->name_len) + sizeof(((zend_compiled_variable*)NULL)->hash_value)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_compiled_variable));
			do {
		

		
		
	
	
	/* DISPATCH(int,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_U(dst->name) = (zstr_uchar *) emalloc(sizeof(zstr_uchar) * (src->name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_U(dst->name), -1, sizeof(zstr_uchar) * (src->name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_U(dst->name), ZSTR_U(src->name), sizeof(zstr_uchar) * (src->name_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->name) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->name), -1, sizeof(zstr_char) * (src->name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->name), ZSTR_S(src->name), sizeof(zstr_char) * (src->name_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->name) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->name), -1, sizeof(zstr_char) * (src->name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->name), ZSTR_S(src->name), sizeof(zstr_char) * (src->name_len + 1));
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,hash_value) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hash_value);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_restore_zend_uint */
	
 	  	 	static void inline 	xc_restore_zend_uint( 		 		 		 		xc_processor_t *processor, zend_uint *dst, const zend_uint * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_uint), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_uint));
			do {
		

		
		
	dst[0] = src[0];
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifndef ZEND_ENGINE_2

	
	

/* {{{ xc_restore_int */
	
 	  	 	static void inline 	xc_restore_int( 		 		 		 		xc_processor_t *processor, int *dst, const int * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(int), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(int));
			do {
		

		
		
	*dst = *src;
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_restore_zend_try_catch_element */
	
 	  	 	static void inline 	xc_restore_zend_try_catch_element( 		 		 		 		xc_processor_t *processor, zend_try_catch_element *dst, const zend_try_catch_element * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_try_catch_element*)NULL)->try_op) + sizeof(((zend_try_catch_element*)NULL)->catch_op)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_try_catch_element));
			do {
		

		
		
	
	
	/* DISPATCH(zend_uint,try_op) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_op);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,catch_op) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->catch_op);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif /* ifdef ZEND_ENGINE_2 */

	
	

/* {{{ xc_restore_zend_brk_cont_element */
	
 	  	 	static void inline 	xc_restore_zend_brk_cont_element( 		 		 		 		xc_processor_t *processor, zend_brk_cont_element *dst, const zend_brk_cont_element * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_brk_cont_element*)NULL)->start) + sizeof(((zend_brk_cont_element*)NULL)->cont) + sizeof(((zend_brk_cont_element*)NULL)->brk) + sizeof(((zend_brk_cont_element*)NULL)->parent)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_brk_cont_element));
			do {
		

		
		
#ifdef ZEND_ENGINE_2_2
#ifndef IS_UNICODE
	
	
	/* DISPATCH(int,start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start);
	done_count ++;

#endif



	

#endif
#endif
	
	
	/* DISPATCH(int,cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,brk) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,parent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	
	

/* {{{ xc_restore_HashTable_zval_ptr */
	
 	  	 	static void inline 	xc_restore_HashTable_zval_ptr( 		 		 		 		xc_processor_t *processor, HashTable *dst, const HashTable * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(HashTable));
			do {
		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				uint n;
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					dst->canary = zend_hash_canary; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		dst->pInternalPointer = NULL;	/* Used for element traversal */ 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		dst->pListHead = NULL; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
			dst->arBuckets = (Bucket* *) ecalloc(src->nTableSize, sizeof(Bucket*));
		
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
	
	
			pnew = (Bucket *) emalloc(sizeof(char) * bucketsize);
			
#ifndef NDEBUG
		memset(pnew, -1, sizeof(char) * bucketsize);
#endif

		
	
	
	

			memcpy(pnew, b, bucketsize);
			
				n = b->h & src->nTableMask;
				/* pnew into hash node chain */
				pnew->pLast = NULL;
				if (dst->arBuckets[n]) {
					pnew->pNext = dst->arBuckets[n];
					pnew->pNext->pLast = pnew;
				}
				else {
					pnew->pNext = NULL;
				}
				dst->arBuckets[n] = pnew;
			
			
			if (sizeof(void *) == sizeof(zval_ptr)) {
				pnew->pData = &pnew->pDataPtr;
								
	
	/* STRUCT_P_EX(zval_ptr,pnew->pData,((zval_ptr*)b->pData),,, ) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof((  ((zval_ptr*)b->pData))[0]));
	
	
	xc_restore_zval_ptr(
		
		
		
		processor,   pnew->pData,   ((zval_ptr*)b->pData)
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zval_ptr,pnew->pData,((zval_ptr*)b->pData),,) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof(( ((zval_ptr*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
			pnew->pData = (zval_ptr *) emalloc(sizeof(zval_ptr));
			
#ifndef NDEBUG
		memset(pnew->pData, -1, sizeof(zval_ptr));
#endif

		
	
	
	

	
	xc_restore_zval_ptr(
		
		
		
		processor,  pnew->pData,  ((zval_ptr*)b->pData)
		
		
		TSRMLS_CC
	);
	
	
	

				pnew->pDataPtr = NULL;
			}

			if (first) {
				dst->pListHead = pnew;
				first = 0;
			}

			
				/* flat link */
				pnew->pListLast = prev;
				pnew->pListNext = NULL;
				if (prev) {
					prev->pListNext = pnew;
				}
			
			prev = pnew;
		}
		
	
	

		dst->pListTail = pnew; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		dst->pDestructor = src->pDestructor; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	
	

/* {{{ xc_restore_HashTable_zend_constant */
	
 	  	 	static void inline 	xc_restore_HashTable_zend_constant( 		 		 		 		xc_processor_t *processor, HashTable *dst, const HashTable * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(HashTable));
			do {
		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				uint n;
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					dst->canary = zend_hash_canary; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		dst->pInternalPointer = NULL;	/* Used for element traversal */ 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		dst->pListHead = NULL; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
			dst->arBuckets = (Bucket* *) ecalloc(src->nTableSize, sizeof(Bucket*));
		
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
	
	
			pnew = (Bucket *) emalloc(sizeof(char) * bucketsize);
			
#ifndef NDEBUG
		memset(pnew, -1, sizeof(char) * bucketsize);
#endif

		
	
	
	

			memcpy(pnew, b, bucketsize);
			
				n = b->h & src->nTableMask;
				/* pnew into hash node chain */
				pnew->pLast = NULL;
				if (dst->arBuckets[n]) {
					pnew->pNext = dst->arBuckets[n];
					pnew->pNext->pLast = pnew;
				}
				else {
					pnew->pNext = NULL;
				}
				dst->arBuckets[n] = pnew;
			
			
			if (sizeof(void *) == sizeof(zend_constant)) {
				pnew->pData = &pnew->pDataPtr;
								
	
	/* STRUCT_P_EX(zend_constant,pnew->pData,((zend_constant*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_constant) == sizeof((  ((zend_constant*)b->pData))[0]));
	
	
	xc_restore_zend_constant(
		
		
		
		processor,   pnew->pData,   ((zend_constant*)b->pData)
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_constant,pnew->pData,((zend_constant*)b->pData),,) */

	
	

	
	assert(sizeof(zend_constant) == sizeof(( ((zend_constant*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
			pnew->pData = (zend_constant *) emalloc(sizeof(zend_constant));
			
#ifndef NDEBUG
		memset(pnew->pData, -1, sizeof(zend_constant));
#endif

		
	
	
	

	
	xc_restore_zend_constant(
		
		
		
		processor,  pnew->pData,  ((zend_constant*)b->pData)
		
		
		TSRMLS_CC
	);
	
	
	

				pnew->pDataPtr = NULL;
			}

			if (first) {
				dst->pListHead = pnew;
				first = 0;
			}

			
				/* flat link */
				pnew->pListLast = prev;
				pnew->pListNext = NULL;
				if (prev) {
					prev->pListNext = pnew;
				}
			
			prev = pnew;
		}
		
	
	

		dst->pListTail = pnew; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		dst->pDestructor = src->pDestructor; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	
	

/* {{{ xc_restore_HashTable_zend_function */
	
		/* export:  	  	 	void 	xc_restore_HashTable_zend_function( 		 		 		 		xc_processor_t *processor, HashTable *dst, const HashTable * const src 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_restore_HashTable_zend_function( 		 		 		 		xc_processor_t *processor, HashTable *dst, const HashTable * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(HashTable));
			do {
		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				uint n;
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					dst->canary = zend_hash_canary; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		dst->pInternalPointer = NULL;	/* Used for element traversal */ 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		dst->pListHead = NULL; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
			dst->arBuckets = (Bucket* *) ecalloc(src->nTableSize, sizeof(Bucket*));
		
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
	
	
			pnew = (Bucket *) emalloc(sizeof(char) * bucketsize);
			
#ifndef NDEBUG
		memset(pnew, -1, sizeof(char) * bucketsize);
#endif

		
	
	
	

			memcpy(pnew, b, bucketsize);
			
				n = b->h & src->nTableMask;
				/* pnew into hash node chain */
				pnew->pLast = NULL;
				if (dst->arBuckets[n]) {
					pnew->pNext = dst->arBuckets[n];
					pnew->pNext->pLast = pnew;
				}
				else {
					pnew->pNext = NULL;
				}
				dst->arBuckets[n] = pnew;
			
			
			if (sizeof(void *) == sizeof(zend_function)) {
				pnew->pData = &pnew->pDataPtr;
								
	
	/* STRUCT_P_EX(zend_function,pnew->pData,((zend_function*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_function) == sizeof((  ((zend_function*)b->pData))[0]));
	
	
	xc_restore_zend_function(
		
		
		
		processor,   pnew->pData,   ((zend_function*)b->pData)
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_function,pnew->pData,((zend_function*)b->pData),,) */

	
	

	
	assert(sizeof(zend_function) == sizeof(( ((zend_function*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
			pnew->pData = (zend_function *) emalloc(sizeof(zend_function));
			
#ifndef NDEBUG
		memset(pnew->pData, -1, sizeof(zend_function));
#endif

		
	
	
	

	
	xc_restore_zend_function(
		
		
		
		processor,  pnew->pData,  ((zend_function*)b->pData)
		
		
		TSRMLS_CC
	);
	
	
	

				pnew->pDataPtr = NULL;
			}

			if (first) {
				dst->pListHead = pnew;
				first = 0;
			}

			
				/* flat link */
				pnew->pListLast = prev;
				pnew->pListNext = NULL;
				if (prev) {
					prev->pListNext = pnew;
				}
			
			prev = pnew;
		}
		
	
	

		dst->pListTail = pnew; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		dst->pDestructor = src->pDestructor; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef ZEND_ENGINE_2

	
	
	

/* {{{ xc_restore_HashTable_zend_property_info */
	
 	  	 	static void inline 	xc_restore_HashTable_zend_property_info( 		 		 		 		xc_processor_t *processor, HashTable *dst, const HashTable * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(HashTable));
			do {
		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				uint n;
		int bucketsize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					dst->canary = zend_hash_canary; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		dst->pInternalPointer = NULL;	/* Used for element traversal */ 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		dst->pListHead = NULL; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
			dst->arBuckets = (Bucket* *) ecalloc(src->nTableSize, sizeof(Bucket*));
		
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			bucketsize = BUCKET_SIZE(b);
			
	
	
	
	/* allocate */
	
	
	
			pnew = (Bucket *) emalloc(sizeof(char) * bucketsize);
			
#ifndef NDEBUG
		memset(pnew, -1, sizeof(char) * bucketsize);
#endif

		
	
	
	

			memcpy(pnew, b, bucketsize);
			
				n = b->h & src->nTableMask;
				/* pnew into hash node chain */
				pnew->pLast = NULL;
				if (dst->arBuckets[n]) {
					pnew->pNext = dst->arBuckets[n];
					pnew->pNext->pLast = pnew;
				}
				else {
					pnew->pNext = NULL;
				}
				dst->arBuckets[n] = pnew;
			
			
			if (sizeof(void *) == sizeof(zend_property_info)) {
				pnew->pData = &pnew->pDataPtr;
								
	
	/* STRUCT_P_EX(zend_property_info,pnew->pData,((zend_property_info*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_property_info) == sizeof((  ((zend_property_info*)b->pData))[0]));
	
	
	xc_restore_zend_property_info(
		
		
		
		processor,   pnew->pData,   ((zend_property_info*)b->pData)
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_property_info,pnew->pData,((zend_property_info*)b->pData),,) */

	
	

	
	assert(sizeof(zend_property_info) == sizeof(( ((zend_property_info*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
			pnew->pData = (zend_property_info *) emalloc(sizeof(zend_property_info));
			
#ifndef NDEBUG
		memset(pnew->pData, -1, sizeof(zend_property_info));
#endif

		
	
	
	

	
	xc_restore_zend_property_info(
		
		
		
		processor,  pnew->pData,  ((zend_property_info*)b->pData)
		
		
		TSRMLS_CC
	);
	
	
	

				pnew->pDataPtr = NULL;
			}

			if (first) {
				dst->pListHead = pnew;
				first = 0;
			}

			
				/* flat link */
				pnew->pListLast = prev;
				pnew->pListNext = NULL;
				if (prev) {
					prev->pListNext = pnew;
				}
			
			prev = pnew;
		}
		
	
	

		dst->pListTail = pnew; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		dst->pDestructor = src->pDestructor; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	

/* {{{ xc_restore_zval */
	
		/* export:  	  	 	void 	xc_restore_zval( 		 		 		 		xc_processor_t *processor, zval *dst, const zval * const src 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_restore_zval( 		 		 		 		xc_processor_t *processor, zval *dst, const zval * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zval*)NULL)->value) + sizeof(((zval*)NULL)->refcount) + sizeof(((zval*)NULL)->type) + sizeof(((zval*)NULL)->is_ref)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zval));
			do {
		

		
			
				/* Variable information */
		
	
	

		switch (src->type & ~IS_CONSTANT_INDEX) {
			case IS_LONG:
			case IS_RESOURCE:
			case IS_BOOL:
				
	
	/* DISPATCH(long,value.lval) */

	
	
	
	
	

	

				break;
			case IS_DOUBLE:
				
	
	/* DISPATCH(double,value.dval) */

	
	
	
	
	

	

				break;
			case IS_NULL:
				
				break;

			case IS_CONSTANT:
#ifdef IS_UNICODE
				if (UG(unicode)) {
					goto proc_unicode;
				}
#endif
			case IS_STRING:
#ifdef FLAG_IS_BC
			case FLAG_IS_BC:
#endif
				
	
	/* DISPATCH(int,value.str.len) */

	
	
	
	
	

	

				
	/* PROC_STRING_L(value.str.val,value.str.len) */
 
	/* PROC_STRING_N(value.str.val,value.str.len + 1) */
 
	
	
	
	
	
	
	
	if (src->value.str.val == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			dst->value.str.val = (char *) emalloc(sizeof(char) * (src->value.str.len + 1));
			
#ifndef NDEBUG
		memset(dst->value.str.val, -1, sizeof(char) * (src->value.str.len + 1));
#endif

		
	
	
	

			memcpy(dst->value.str.val, src->value.str.val, sizeof(char) * (src->value.str.len + 1));
		
		
		
	}
	
	
	
	
	
	

				break;
#ifdef IS_UNICODE
			case IS_UNICODE:
proc_unicode:
				
	
	/* DISPATCH(int32_t,value.uni.len) */

	
	
	
	
	

	

				
	/* PROC_ZSTRING_L(1,value.uni.val,value.uni.len) */
 
	
	/* PROC_ZSTRING_N(1,value.uni.val,value.uni.len,len) */

#ifdef IS_UNICODE
	
	
	
	
	
	
	
	
	
	
	if (ZSTR_U(src->value.uni.val) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_U(dst->value.uni.val) = (zstr_uchar *) emalloc(sizeof(zstr_uchar) * (src->value.uni.len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_U(dst->value.uni.val), -1, sizeof(zstr_uchar) * (src->value.uni.len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_U(dst->value.uni.val), ZSTR_U(src->value.uni.val), sizeof(zstr_uchar) * (src->value.uni.len + 1));
		
		
		
	}
	
	
	
	
	
	

	
#else
	
	
	
	
	
	
	
	
	
	if (ZSTR_S(src->value.uni.val) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->value.uni.val) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->value.uni.len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->value.uni.val), -1, sizeof(zstr_char) * (src->value.uni.len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->value.uni.val), ZSTR_S(src->value.uni.val), sizeof(zstr_char) * (src->value.uni.len + 1));
		
		
		
	}
	
	
	
	
	
	

#endif
	

				break;
#endif

			case IS_ARRAY:
			case IS_CONSTANT_ARRAY:
				
	
	/* STRUCT_P(HashTable,value.ht,HashTable_zval_ptr) */

	if (src->value.ht) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->value.ht,src->value.ht,value.ht,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->value.ht)[0]));
	
	
	
	
	/* allocate */
	
	
	
			ALLOC_HASHTABLE(dst->value.ht);
			
#ifndef NDEBUG
		memset(dst->value.ht, -1, sizeof(HashTable));
#endif

		
	
	
	

	
	xc_restore_HashTable_zval_ptr(
		
		
		
		processor,  dst->value.ht,  src->value.ht
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

				break;

			case IS_OBJECT:
				
				#ifndef ZEND_ENGINE_2
				
	
	/* STRUCT_P(zend_class_entry,value.obj.ce) */

	if (src->value.obj.ce) {
		
		
	
	/* STRUCT_P_EX(zend_class_entry,dst->value.obj.ce,src->value.obj.ce,value.obj.ce,) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof(( src->value.obj.ce)[0]));
	
	
	
	
	/* allocate */
	
	
	
			dst->value.obj.ce = (zend_class_entry *) emalloc(sizeof(zend_class_entry));
			
#ifndef NDEBUG
		memset(dst->value.obj.ce, -1, sizeof(zend_class_entry));
#endif

		
	
	
	

	
	xc_restore_zend_class_entry(
		
		
		
		processor,  dst->value.obj.ce,  src->value.obj.ce
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

				
	
	/* STRUCT_P(HashTable,value.obj.properties,HashTable_zval_ptr) */

	if (src->value.obj.properties) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->value.obj.properties,src->value.obj.properties,value.obj.properties,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->value.obj.properties)[0]));
	
	
	
	
	/* allocate */
	
	
	
			ALLOC_HASHTABLE(dst->value.obj.properties);
			
#ifndef NDEBUG
		memset(dst->value.obj.properties, -1, sizeof(HashTable));
#endif

		
	
	
	

	
	xc_restore_HashTable_zval_ptr(
		
		
		
		processor,  dst->value.obj.properties,  src->value.obj.properties
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

#endif
				break;

			default:
				assert(0);
		}
		
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif


		
	
	/* DISPATCH(zval_data_type,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_uchar,is_ref) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->is_ref);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_ushort,refcount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_3
		
	
	/* DISPATCH(zend_uchar,idx_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->idx_type);
	done_count ++;

#endif



	

#endif
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_restore_zval_ptr */
	
 	  	 	static void inline 	xc_restore_zval_ptr( 		 		 		 		xc_processor_t *processor, zval_ptr *dst, const zval_ptr * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zval_ptr), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zval_ptr));
			do {
		

		
			
		do {
			
				if (processor->reference) {
					zval_ptr *ppzv;
					if (zend_hash_find(&processor->zvalptrs, (char *) &src[0], sizeof(src[0]), (void **) &ppzv) == SUCCESS) {
						
							dst[0] = *ppzv;
							/* *dst is updated */
													
						
						
						assert(!xc_is_shm(dst[0]));
						break;
					}
				}
			
			
			
	
	
	
	/* allocate */
	
	
	
			ALLOC_ZVAL(dst[0]);
			
#ifndef NDEBUG
		memset(dst[0], -1, sizeof(zval));
#endif

		
	
	
	

			
				if (processor->reference) {
					
						zval_ptr pzv = dst[0];
						
					
					if (zend_hash_add(&processor->zvalptrs, (char *) &src[0], sizeof(src[0]), (void *) &pzv, sizeof(pzv), NULL) == SUCCESS) {
						/* first add, go on */
											}
					else {
						assert(0);
					}
				}
			
			
							
			
			
	
	/* STRUCT_P_EX(zval,dst[0],src[0],[0],, ) */

	
	

	
	assert(sizeof(zval) == sizeof((  src[0])[0]));
	
	
	xc_restore_zval(
		
		
		
		processor,   dst[0],   src[0]
		
		
		TSRMLS_CC
	);
	
	
	

			
		} while (0);
	
	
#ifndef NDEBUG
		
	done_size += sizeof(zval_ptr);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_restore_zend_arg_info */
	
 	  	 	static void inline 	xc_restore_zend_arg_info( 		 		 		 		xc_processor_t *processor, zend_arg_info *dst, const zend_arg_info * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_arg_info*)NULL)->name) + sizeof(((zend_arg_info*)NULL)->name_len) + sizeof(((zend_arg_info*)NULL)->class_name) + sizeof(((zend_arg_info*)NULL)->class_name_len) + sizeof(((zend_arg_info*)NULL)->array_type_hint) + sizeof(((zend_arg_info*)NULL)->allow_null) + sizeof(((zend_arg_info*)NULL)->pass_by_reference) + sizeof(((zend_arg_info*)NULL)->return_reference) + sizeof(((zend_arg_info*)NULL)->required_num_args)  ), assert_count = 9;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_arg_info));
			do {
		

		
		
	
	
	/* DISPATCH(zend_uint,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_U(dst->name) = (zstr_uchar *) emalloc(sizeof(zstr_uchar) * (src->name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_U(dst->name), -1, sizeof(zstr_uchar) * (src->name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_U(dst->name), ZSTR_U(src->name), sizeof(zstr_uchar) * (src->name_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->name) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->name), -1, sizeof(zstr_char) * (src->name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->name), ZSTR_S(src->name), sizeof(zstr_char) * (src->name_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->name) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->name), -1, sizeof(zstr_char) * (src->name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->name), ZSTR_S(src->name), sizeof(zstr_char) * (src->name_len + 1));
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_uint,class_name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,class_name,class_name_len) */
 
	
	/* PROC_ZSTRING_N(,class_name,class_name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->class_name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_U(dst->class_name) = (zstr_uchar *) emalloc(sizeof(zstr_uchar) * (src->class_name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_U(dst->class_name), -1, sizeof(zstr_uchar) * (src->class_name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_U(dst->class_name), ZSTR_U(src->class_name), sizeof(zstr_uchar) * (src->class_name_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->class_name) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->class_name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->class_name), -1, sizeof(zstr_char) * (src->class_name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->class_name), ZSTR_S(src->class_name), sizeof(zstr_char) * (src->class_name_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->class_name) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->class_name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->class_name), -1, sizeof(zstr_char) * (src->class_name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->class_name), ZSTR_S(src->class_name), sizeof(zstr_char) * (src->class_name_len + 1));
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_bool,array_type_hint) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->array_type_hint);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,allow_null) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->allow_null);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_by_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_by_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,return_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,required_num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_restore_zend_constant */
	
 	  	 	static void inline 	xc_restore_zend_constant( 		 		 		 		xc_processor_t *processor, zend_constant *dst, const zend_constant * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_constant*)NULL)->value) + sizeof(((zend_constant*)NULL)->flags) + sizeof(((zend_constant*)NULL)->name) + sizeof(((zend_constant*)NULL)->name_len) + sizeof(((zend_constant*)NULL)->module_number)  ), assert_count = 5;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_constant));
			do {
		

		
			
	
	/* STRUCT(zval,value) */

	assert(sizeof(zval) == sizeof(src->value));
	
	
	
	/* STRUCT_P_EX(zval,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->value)[0]));
	
	
	xc_restore_zval(
		
		
		
		processor, & dst->value, & src->value
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



	
	
	/* DISPATCH(int,flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	
	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_U(dst->name) = (zstr_uchar *) malloc(sizeof(zstr_uchar) * (src->name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_U(dst->name), -1, sizeof(zstr_uchar) * (src->name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_U(dst->name), ZSTR_U(src->name), sizeof(zstr_uchar) * (src->name_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->name) = (zstr_char *) malloc(sizeof(zstr_char) * (src->name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->name), -1, sizeof(zstr_char) * (src->name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->name), ZSTR_S(src->name), sizeof(zstr_char) * (src->name_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->name) = (zstr_char *) malloc(sizeof(zstr_char) * (src->name_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->name), -1, sizeof(zstr_char) * (src->name_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->name), ZSTR_S(src->name), sizeof(zstr_char) * (src->name_len + 1));
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	
	/* DISPATCH(int,module_number) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module_number);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_restore_zend_function */
	
		/* export:  	  	 	void 	xc_restore_zend_function( 		 		 		 		xc_processor_t *processor, zend_function *dst, const zend_function * const src 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_restore_zend_function( 		 		 		 		xc_processor_t *processor, zend_function *dst, const zend_function * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_function), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_function));
			do {
		

		
			
	
	

	switch (src->type) {
	case ZEND_INTERNAL_FUNCTION:
	case ZEND_OVERLOADED_FUNCTION:
		
		break;

	case ZEND_USER_FUNCTION:
	case ZEND_EVAL_CODE:
		
		
	
	/* STRUCT(zend_op_array,op_array) */

	assert(sizeof(zend_op_array) == sizeof(src->op_array));
	
	
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,,&) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof((& src->op_array)[0]));
	
	
	xc_restore_zend_op_array(
		
		
		
		processor, & dst->op_array, & src->op_array
		
		
		TSRMLS_CC
	);
	
	
	

	

		break;

	default:
		assert(0);
	}
	
	
	

	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_restore_zend_property_info */
	
 	  	 	static void inline 	xc_restore_zend_property_info( 		 		 		 		xc_processor_t *processor, zend_property_info *dst, const zend_property_info * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_property_info*)NULL)->flags) + sizeof(((zend_property_info*)NULL)->name) + sizeof(((zend_property_info*)NULL)->name_length) + sizeof(((zend_property_info*)NULL)->h) + sizeof(((zend_property_info*)NULL)->doc_comment) + sizeof(((zend_property_info*)NULL)->doc_comment_len) + sizeof(((zend_property_info*)NULL)->ce)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_property_info));
			do {
		

		
		
	
	
	/* DISPATCH(zend_uint,flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,name_length) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_U(dst->name) = (zstr_uchar *) emalloc(sizeof(zstr_uchar) * (src->name_length + 1));
			
#ifndef NDEBUG
		memset(ZSTR_U(dst->name), -1, sizeof(zstr_uchar) * (src->name_length + 1));
#endif

		
	
	
	

			memcpy(ZSTR_U(dst->name), ZSTR_U(src->name), sizeof(zstr_uchar) * (src->name_length + 1));
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->name) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->name_length + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->name), -1, sizeof(zstr_char) * (src->name_length + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->name), ZSTR_S(src->name), sizeof(zstr_char) * (src->name_length + 1));
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->name) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->name_length + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->name), -1, sizeof(zstr_char) * (src->name_length + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->name), ZSTR_S(src->name), sizeof(zstr_char) * (src->name_length + 1));
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,h) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->h);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_U(dst->doc_comment) = (zstr_uchar *) emalloc(sizeof(zstr_uchar) * (src->doc_comment_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_U(dst->doc_comment), -1, sizeof(zstr_uchar) * (src->doc_comment_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_U(dst->doc_comment), ZSTR_U(src->doc_comment), sizeof(zstr_uchar) * (src->doc_comment_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->doc_comment) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->doc_comment_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->doc_comment), -1, sizeof(zstr_char) * (src->doc_comment_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->doc_comment), ZSTR_S(src->doc_comment), sizeof(zstr_char) * (src->doc_comment_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->doc_comment) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->doc_comment_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->doc_comment), -1, sizeof(zstr_char) * (src->doc_comment_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->doc_comment), ZSTR_S(src->doc_comment), sizeof(zstr_char) * (src->doc_comment_len + 1));
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif
	#if defined(ZEND_ENGINE_2_2) && !defined(IS_UNICODE)
	
	if (src->ce) {
		
		dst->ce = xc_get_class(processor, (zend_ulong) src->ce);
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce);
	done_count ++;

#endif


#endif

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_restore_zend_class_entry */
	
		/* export:  	  	 	void 	xc_restore_zend_class_entry( 		 		 		 		xc_processor_t *processor, zend_class_entry *dst, const zend_class_entry * const src 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_restore_zend_class_entry( 		 		 		 		xc_processor_t *processor, zend_class_entry *dst, const zend_class_entry * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_class_entry*)NULL)->type) + sizeof(((zend_class_entry*)NULL)->name) + sizeof(((zend_class_entry*)NULL)->name_length) + sizeof(((zend_class_entry*)NULL)->parent) + sizeof(((zend_class_entry*)NULL)->refcount) + sizeof(((zend_class_entry*)NULL)->constants_updated) + sizeof(((zend_class_entry*)NULL)->ce_flags) + sizeof(((zend_class_entry*)NULL)->function_table) + sizeof(((zend_class_entry*)NULL)->default_properties) + sizeof(((zend_class_entry*)NULL)->properties_info) + sizeof(((zend_class_entry*)NULL)->default_static_members) + sizeof(((zend_class_entry*)NULL)->static_members) + sizeof(((zend_class_entry*)NULL)->constants_table) + sizeof(((zend_class_entry*)NULL)->builtin_functions) + sizeof(((zend_class_entry*)NULL)->constructor) + sizeof(((zend_class_entry*)NULL)->destructor) + sizeof(((zend_class_entry*)NULL)->clone) + sizeof(((zend_class_entry*)NULL)->__get) + sizeof(((zend_class_entry*)NULL)->__set) + sizeof(((zend_class_entry*)NULL)->__unset) + sizeof(((zend_class_entry*)NULL)->__isset) + sizeof(((zend_class_entry*)NULL)->__call) + sizeof(((zend_class_entry*)NULL)->__tostring) + sizeof(((zend_class_entry*)NULL)->serialize_func) + sizeof(((zend_class_entry*)NULL)->unserialize_func) + sizeof(((zend_class_entry*)NULL)->iterator_funcs) + sizeof(((zend_class_entry*)NULL)->create_object) + sizeof(((zend_class_entry*)NULL)->get_iterator) + sizeof(((zend_class_entry*)NULL)->interface_gets_implemented) + sizeof(((zend_class_entry*)NULL)->serialize) + sizeof(((zend_class_entry*)NULL)->unserialize) + sizeof(((zend_class_entry*)NULL)->interfaces) + sizeof(((zend_class_entry*)NULL)->num_interfaces) + sizeof(((zend_class_entry*)NULL)->filename) + sizeof(((zend_class_entry*)NULL)->line_start) + sizeof(((zend_class_entry*)NULL)->line_end) + sizeof(((zend_class_entry*)NULL)->doc_comment) + sizeof(((zend_class_entry*)NULL)->doc_comment_len) + sizeof(((zend_class_entry*)NULL)->module)  ), assert_count = 39;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_class_entry));
			do {
		

		
			
		processor->active_class_entry_src = src;
		processor->active_class_entry_dst = dst;
	
	
	
	/* DISPATCH(char,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,name_length) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_U(dst->name) = (zstr_uchar *) emalloc(sizeof(zstr_uchar) * (src->name_length + 1));
			
#ifndef NDEBUG
		memset(ZSTR_U(dst->name), -1, sizeof(zstr_uchar) * (src->name_length + 1));
#endif

		
	
	
	

			memcpy(ZSTR_U(dst->name), ZSTR_U(src->name), sizeof(zstr_uchar) * (src->name_length + 1));
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->name) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->name_length + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->name), -1, sizeof(zstr_char) * (src->name_length + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->name), ZSTR_S(src->name), sizeof(zstr_char) * (src->name_length + 1));
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->name) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->name_length + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->name), -1, sizeof(zstr_char) * (src->name_length + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->name), ZSTR_S(src->name), sizeof(zstr_char) * (src->name_length + 1));
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
#ifndef ZEND_ENGINE_2
		/* just copy parent and resolve on install_class */
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif


#else
		
	if (src->parent) {
		
		dst->parent = xc_get_class(processor, (zend_ulong) src->parent);
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif


#endif
	
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(int,refcount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	

#else
	
	
	/* STRUCT_P(int,refcount) */

	if (src->refcount) {
		
		
	
	/* STRUCT_P_EX(int,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(int) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
	
	
			dst->refcount = (int *) emalloc(sizeof(int));
			
#ifndef NDEBUG
		memset(dst->refcount, -1, sizeof(int));
#endif

		
	
	
	

	
	xc_restore_int(
		
		
		
		processor,  dst->refcount,  src->refcount
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(zend_bool,constants_updated) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_updated);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,ce_flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce_flags);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT(HashTable,default_properties,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_properties));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->default_properties,src->default_properties,default_properties,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_properties)[0]));
	
	
	xc_restore_HashTable_zval_ptr(
		
		
		
		processor, & dst->default_properties, & src->default_properties
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_properties);
	done_count ++;

#endif



	dst->builtin_functions = src->builtin_functions;
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->builtin_functions);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT(HashTable,properties_info,HashTable_zend_property_info) */

	assert(sizeof(HashTable) == sizeof(src->properties_info));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->properties_info,src->properties_info,properties_info,HashTable_zend_property_info,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->properties_info)[0]));
	
	
	xc_restore_HashTable_zend_property_info(
		
		
		
		processor, & dst->properties_info, & src->properties_info
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->properties_info);
	done_count ++;

#endif



#	ifdef ZEND_ENGINE_2_1
	
	
	/* STRUCT(HashTable,default_static_members,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_static_members));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->default_static_members,src->default_static_members,default_static_members,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_static_members)[0]));
	
	
	xc_restore_HashTable_zval_ptr(
		
		
		
		processor, & dst->default_static_members, & src->default_static_members
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_static_members);
	done_count ++;

#endif



	dst->static_members = &dst->default_static_members;
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif


#	else
	
	
	/* STRUCT_P(HashTable,static_members,HashTable_zval_ptr) */

	if (src->static_members) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->static_members,src->static_members,static_members,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_members)[0]));
	
	
	
	
	/* allocate */
	
	
	
			ALLOC_HASHTABLE(dst->static_members);
			
#ifndef NDEBUG
		memset(dst->static_members, -1, sizeof(HashTable));
#endif

		
	
	
	

	
	xc_restore_HashTable_zval_ptr(
		
		
		
		processor,  dst->static_members,  src->static_members
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif



#	endif
	
	
	/* STRUCT(HashTable,constants_table,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->constants_table));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->constants_table,src->constants_table,constants_table,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->constants_table)[0]));
	
	
	xc_restore_HashTable_zval_ptr(
		
		
		
		processor, & dst->constants_table, & src->constants_table
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_table);
	done_count ++;

#endif




		
		if (src->num_interfaces) {
			
	
	
	
	/* allocate */
	
	
	
			dst->interfaces = (zend_class_entry* *) ecalloc(src->num_interfaces, sizeof(zend_class_entry*));
		
	
	
	

			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interfaces);
	done_count ++;

#endif


		}
		else {
			
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interfaces);
	done_count ++;

#endif



		}
	
	
	
	
	
	/* DISPATCH(zend_uint,num_interfaces) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_interfaces);
	done_count ++;

#endif



	


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_U(dst->doc_comment) = (zstr_uchar *) emalloc(sizeof(zstr_uchar) * (src->doc_comment_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_U(dst->doc_comment), -1, sizeof(zstr_uchar) * (src->doc_comment_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_U(dst->doc_comment), ZSTR_U(src->doc_comment), sizeof(zstr_uchar) * (src->doc_comment_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->doc_comment) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->doc_comment_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->doc_comment), -1, sizeof(zstr_char) * (src->doc_comment_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->doc_comment), ZSTR_S(src->doc_comment), sizeof(zstr_char) * (src->doc_comment_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->doc_comment) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->doc_comment_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->doc_comment), -1, sizeof(zstr_char) * (src->doc_comment_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->doc_comment), ZSTR_S(src->doc_comment), sizeof(zstr_char) * (src->doc_comment_len + 1));
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->iterator_funcs);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->create_object);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_iterator);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interface_gets_implemented);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2_3
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_static_method);
	done_count ++;

#endif


#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize);
	done_count ++;

#endif


	/* deal with it inside xc_fix_method */
	dst->constructor = NULL;
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->destructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->clone);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__set);
	done_count ++;

#endif


/* should be >5.1 */
#ifdef ZEND_ENGINE_2_1
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__unset);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__isset);
	done_count ++;

#endif


# if defined(ZEND_ENGINE_2_2) || PHP_MAJOR_VERSION >= 6
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__tostring);
	done_count ++;

#endif


# endif
#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__call);
	done_count ++;

#endif


#ifdef ZEND_CALLSTATIC_FUNC_NAME
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__callstatic);
	done_count ++;

#endif


#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module);
	done_count ++;

#endif


#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_function_call);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_set);
	done_count ++;

#endif


#endif
		
	
	/* STRUCT(HashTable,function_table,HashTable_zend_function) */

	assert(sizeof(HashTable) == sizeof(src->function_table));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->function_table,src->function_table,function_table,HashTable_zend_function,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->function_table)[0]));
	
	
	xc_restore_HashTable_zend_function(
		
		
		
		processor, & dst->function_table, & src->function_table
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_table);
	done_count ++;

#endif



	dst->function_table.pDestructor = ZEND_FUNCTION_DTOR;
	
		processor->active_class_entry_src = NULL;
		processor->active_class_entry_dst = NULL;
	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_restore_znode */
	
 	  	 	static void inline 	xc_restore_znode( 		 		 		 		xc_processor_t *processor, znode *dst, const znode * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((znode*)NULL)->op_type) + sizeof(((znode*)NULL)->u)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(znode));
			do {
		

		
			
	
	/* DISPATCH(int,op_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_type);
	done_count ++;

#endif



	


#ifdef IS_CV
#	define XCACHE_IS_CV IS_CV
#else
/* compatible with zend optimizer */
#	define XCACHE_IS_CV 16
#endif
	assert(src->op_type == IS_CONST ||
		src->op_type == IS_VAR ||
		src->op_type == XCACHE_IS_CV ||
		src->op_type == IS_TMP_VAR ||
		src->op_type == IS_UNUSED);
#undef XCACHE_IS_CV
		
	
	

	switch (src->op_type) {
		case IS_CONST:
			
	
	/* STRUCT(zval,u.constant) */

	assert(sizeof(zval) == sizeof(src->u.constant));
	
	
	
	/* STRUCT_P_EX(zval,dst->u.constant,src->u.constant,u.constant,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->u.constant)[0]));
	
	
	xc_restore_zval(
		
		
		
		processor, & dst->u.constant, & src->u.constant
		
		
		TSRMLS_CC
	);
	
	
	

	

			break;
		
			
		
	}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->u);
	done_count ++;

#endif



		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_restore_zend_op */
	
		/* export:  	  	 	void 	xc_restore_zend_op( 		 		 		 		xc_processor_t *processor, zend_op *dst, const zend_op * const src 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_restore_zend_op( 		 		 		 		xc_processor_t *processor, zend_op *dst, const zend_op * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op*)NULL)->handler) + sizeof(((zend_op*)NULL)->result) + sizeof(((zend_op*)NULL)->op1) + sizeof(((zend_op*)NULL)->op2) + sizeof(((zend_op*)NULL)->extended_value) + sizeof(((zend_op*)NULL)->lineno) + sizeof(((zend_op*)NULL)->opcode)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_op));
			do {
		

		
			
	
	/* DISPATCH(zend_uchar,opcode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcode);
	done_count ++;

#endif



	

	
	
	/* STRUCT(znode,result) */

	assert(sizeof(znode) == sizeof(src->result));
	
	
	
	/* STRUCT_P_EX(znode,dst->result,src->result,result,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->result)[0]));
	
	
	xc_restore_znode(
		
		
		
		processor, & dst->result, & src->result
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->result);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op1) */

	assert(sizeof(znode) == sizeof(src->op1));
	
	
	
	/* STRUCT_P_EX(znode,dst->op1,src->op1,op1,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op1)[0]));
	
	
	xc_restore_znode(
		
		
		
		processor, & dst->op1, & src->op1
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op1);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op2) */

	assert(sizeof(znode) == sizeof(src->op2));
	
	
	
	/* STRUCT_P_EX(znode,dst->op2,src->op2,op2,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op2)[0]));
	
	
	xc_restore_znode(
		
		
		
		processor, & dst->op2, & src->op2
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op2);
	done_count ++;

#endif



	
	
	/* DISPATCH(ulong,extended_value) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->extended_value);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,lineno) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->lineno);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
		switch (src->opcode) {
			case ZEND_JMP:
				dst->op1.u.jmp_addr = processor->active_opcodes_dst + (src->op1.u.jmp_addr - processor->active_opcodes_src);
				break;

			case ZEND_JMPZ:
			case ZEND_JMPNZ:
			case ZEND_JMPZ_EX:
			case ZEND_JMPNZ_EX:
				dst->op2.u.jmp_addr = processor->active_opcodes_dst + (src->op2.u.jmp_addr - processor->active_opcodes_src);
				break;

			default:
				break;
		}
	
	
	
	/* DISPATCH(opcode_handler_t,handler) */

	/* is copying enough? */
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handler);
	done_count ++;

#endif


	

#endif

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_restore_zend_op_array */
	
		/* export:  	  	 	void 	xc_restore_zend_op_array( 		 		 		 		xc_processor_t *processor, zend_op_array *dst, const zend_op_array * const src 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_restore_zend_op_array( 		 		 		 		xc_processor_t *processor, zend_op_array *dst, const zend_op_array * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op_array*)NULL)->type) + sizeof(((zend_op_array*)NULL)->function_name) + sizeof(((zend_op_array*)NULL)->scope) + sizeof(((zend_op_array*)NULL)->fn_flags) + sizeof(((zend_op_array*)NULL)->prototype) + sizeof(((zend_op_array*)NULL)->num_args) + sizeof(((zend_op_array*)NULL)->required_num_args) + sizeof(((zend_op_array*)NULL)->arg_info) + sizeof(((zend_op_array*)NULL)->pass_rest_by_reference) + sizeof(((zend_op_array*)NULL)->return_reference) + sizeof(((zend_op_array*)NULL)->refcount) + sizeof(((zend_op_array*)NULL)->opcodes) + sizeof(((zend_op_array*)NULL)->last) + sizeof(((zend_op_array*)NULL)->size) + sizeof(((zend_op_array*)NULL)->vars) + sizeof(((zend_op_array*)NULL)->last_var) + sizeof(((zend_op_array*)NULL)->size_var) + sizeof(((zend_op_array*)NULL)->T) + sizeof(((zend_op_array*)NULL)->brk_cont_array) + sizeof(((zend_op_array*)NULL)->last_brk_cont) + sizeof(((zend_op_array*)NULL)->current_brk_cont) + sizeof(((zend_op_array*)NULL)->try_catch_array) + sizeof(((zend_op_array*)NULL)->last_try_catch) + sizeof(((zend_op_array*)NULL)->static_variables) + sizeof(((zend_op_array*)NULL)->start_op) + sizeof(((zend_op_array*)NULL)->backpatch_count) + sizeof(((zend_op_array*)NULL)->done_pass_two) + sizeof(((zend_op_array*)NULL)->uses_this) + sizeof(((zend_op_array*)NULL)->filename) + sizeof(((zend_op_array*)NULL)->line_start) + sizeof(((zend_op_array*)NULL)->line_end) + sizeof(((zend_op_array*)NULL)->doc_comment) + sizeof(((zend_op_array*)NULL)->doc_comment_len) + sizeof(((zend_op_array*)NULL)->reserved)  ), assert_count = 34;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(zend_op_array));
			do {
		

		
			
				if (!processor->readonly_protection && !(src == processor->xce_src->data.php->op_array && processor->xce_src->data.php->have_early_binding)) {
		/* really fast shallow copy */
		memcpy(dst, src, sizeof(src[0]));
		dst->refcount[0] = 1000;
		/* deep */
		
	
	/* STRUCT_P(HashTable,static_variables,HashTable_zval_ptr) */

	if (src->static_variables) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->static_variables,src->static_variables,static_variables,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_variables)[0]));
	
	
	
	
	/* allocate */
	
	
	
			ALLOC_HASHTABLE(dst->static_variables);
			
#ifndef NDEBUG
		memset(dst->static_variables, -1, sizeof(HashTable));
#endif

		
	
	
	

	
	xc_restore_HashTable_zval_ptr(
		
		
		
		processor,  dst->static_variables,  src->static_variables
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_variables);
	done_count ++;

#endif



		
	}
	else
	
	do {
		zend_uint ii;
	int i;

	/* Common elements */
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING(,function_name) */
 
	
	/* PROC_ZSTRING_N(,function_name,,strlen) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->function_name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_U(dst->function_name) = (zstr_uchar *) emalloc(sizeof(zstr_uchar) * (xc_zstrlen_uchar (src->function_name) + 1));
			
#ifndef NDEBUG
		memset(ZSTR_U(dst->function_name), -1, sizeof(zstr_uchar) * (xc_zstrlen_uchar (src->function_name) + 1));
#endif

		
	
	
	

			memcpy(ZSTR_U(dst->function_name), ZSTR_U(src->function_name), sizeof(zstr_uchar) * (xc_zstrlen_uchar (src->function_name) + 1));
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->function_name) = (zstr_char *) emalloc(sizeof(zstr_char) * (xc_zstrlen_char (src->function_name) + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->function_name), -1, sizeof(zstr_char) * (xc_zstrlen_char (src->function_name) + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->function_name), ZSTR_S(src->function_name), sizeof(zstr_char) * (xc_zstrlen_char (src->function_name) + 1));
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->function_name) = (zstr_char *) emalloc(sizeof(zstr_char) * (xc_zstrlen_char (src->function_name) + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->function_name), -1, sizeof(zstr_char) * (xc_zstrlen_char (src->function_name) + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->function_name), ZSTR_S(src->function_name), sizeof(zstr_char) * (xc_zstrlen_char (src->function_name) + 1));
		
		
		
	}
	
	
	
	
	
	

#endif
	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,fn_flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->fn_flags);
	done_count ++;

#endif



	

	


	if (src->arg_info) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
			dst->arg_info = (zend_arg_info *) emalloc(sizeof(zend_arg_info) * src->num_args);
			
#ifndef NDEBUG
		memset(dst->arg_info, -1, sizeof(zend_arg_info) * src->num_args);
#endif

		
	
	
	

			
			for (ii = 0; ii < src->num_args; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_arg_info,arg_info[i],) */

	assert(sizeof(zend_arg_info) == sizeof(src->arg_info[ii]));
	
	
	
	/* STRUCT_P_EX(zend_arg_info,dst->arg_info[i],src->arg_info[i],arg_info[i],,&) */

	
	

	
	assert(sizeof(zend_arg_info) == sizeof((& src->arg_info[ii])[0]));
	
	
	xc_restore_zend_arg_info(
		
		
		
		processor, & dst->arg_info[ii], & src->arg_info[ii]
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,required_num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_rest_by_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_rest_by_reference);
	done_count ++;

#endif



	

#else
	if (src->arg_types) {
		
	
	
	
	/* allocate */
	
	
	
			dst->arg_types = (zend_uchar *) emalloc(sizeof(zend_uchar) * src->arg_types[0] + 1);
			
#ifndef NDEBUG
		memset(dst->arg_types, -1, sizeof(zend_uchar) * src->arg_types[0] + 1);
#endif

		
	
	
	

		memcpy(dst->arg_types, src->arg_types, sizeof(src->arg_types[0]) * (src->arg_types[0]+1));
		
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif


	}
	else {
		
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif



		
	}
#endif
	
	
	/* DISPATCH(unsigned char,return_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	/* END of common elements */
#ifdef IS_UNICODE
	#endif

	
	
	/* STRUCT_P(zend_uint,refcount) */

	if (src->refcount) {
		
		
	
	/* STRUCT_P_EX(zend_uint,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(zend_uint) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
	
	
			dst->refcount = (zend_uint *) emalloc(sizeof(zend_uint));
			
#ifndef NDEBUG
		memset(dst->refcount, -1, sizeof(zend_uint));
#endif

		
	
	
	

	
	xc_restore_zend_uint(
		
		
		
		processor,  dst->refcount,  src->refcount
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	
	

	
	


	if (src->opcodes) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
			dst->opcodes = (zend_op *) emalloc(sizeof(zend_op) * src->last);
			
#ifndef NDEBUG
		memset(dst->opcodes, -1, sizeof(zend_op) * src->last);
#endif

		
	
	
	

			
		processor->active_opcodes_dst = dst->opcodes;
		processor->active_opcodes_src = src->opcodes;
	
			for (ii = 0; ii < src->last; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_op,opcodes[i],) */

	assert(sizeof(zend_op) == sizeof(src->opcodes[ii]));
	
	
	
	/* STRUCT_P_EX(zend_op,dst->opcodes[i],src->opcodes[i],opcodes[i],,&) */

	
	

	
	assert(sizeof(zend_op) == sizeof((& src->opcodes[ii])[0]));
	
	
	xc_restore_zend_op(
		
		
		
		processor, & dst->opcodes[ii], & src->opcodes[ii]
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif



		
		processor->active_opcodes_dst = dst->opcodes;
		processor->active_opcodes_src = src->opcodes;
	
	}



	
	
	
	/* DISPATCH(zend_uint,last) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last);
	done_count ++;

#endif



	

	dst->size = src->last;
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



#ifdef IS_CV
	
	if (src->vars) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
			dst->vars = (zend_compiled_variable *) emalloc(sizeof(zend_compiled_variable) * src->last_var);
			
#ifndef NDEBUG
		memset(dst->vars, -1, sizeof(zend_compiled_variable) * src->last_var);
#endif

		
	
	
	

			
			for (i = 0; i < src->last_var; i ++) {
				
	
	

					
					
	
	/* STRUCT(zend_compiled_variable,vars[i],) */

	assert(sizeof(zend_compiled_variable) == sizeof(src->vars[i]));
	
	
	
	/* STRUCT_P_EX(zend_compiled_variable,dst->vars[i],src->vars[i],vars[i],,&) */

	
	

	
	assert(sizeof(zend_compiled_variable) == sizeof((& src->vars[i])[0]));
	
	
	xc_restore_zend_compiled_variable(
		
		
		
		processor, & dst->vars[i], & src->vars[i]
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_var) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_var);
	done_count ++;

#endif



	

	dst->size_var = src->last_var;
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size_var);
	done_count ++;

#endif


#else
		
#endif

	
	
	/* DISPATCH(zend_uint,T) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->T);
	done_count ++;

#endif



	


	


	if (src->brk_cont_array) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
			dst->brk_cont_array = (zend_brk_cont_element *) emalloc(sizeof(zend_brk_cont_element) * src->last_brk_cont);
			
#ifndef NDEBUG
		memset(dst->brk_cont_array, -1, sizeof(zend_brk_cont_element) * src->last_brk_cont);
#endif

		
	
	
	

			
			for (ii = 0; ii < src->last_brk_cont; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_brk_cont_element,brk_cont_array[i],) */

	assert(sizeof(zend_brk_cont_element) == sizeof(src->brk_cont_array[ii]));
	
	
	
	/* STRUCT_P_EX(zend_brk_cont_element,dst->brk_cont_array[i],src->brk_cont_array[i],brk_cont_array[i],,&) */

	
	

	
	assert(sizeof(zend_brk_cont_element) == sizeof((& src->brk_cont_array[ii])[0]));
	
	
	xc_restore_zend_brk_cont_element(
		
		
		
		processor, & dst->brk_cont_array[ii], & src->brk_cont_array[ii]
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,last_brk_cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_brk_cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,current_brk_cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->current_brk_cont);
	done_count ++;

#endif



	

#ifndef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_globals) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_globals);
	done_count ++;

#endif



	

#endif

#ifdef ZEND_ENGINE_2
	
	if (src->try_catch_array) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
			dst->try_catch_array = (zend_try_catch_element *) emalloc(sizeof(zend_try_catch_element) * src->last_try_catch);
			
#ifndef NDEBUG
		memset(dst->try_catch_array, -1, sizeof(zend_try_catch_element) * src->last_try_catch);
#endif

		
	
	
	

			
			for (i = 0; i < src->last_try_catch; i ++) {
				
	
	

					
					
	
	/* STRUCT(zend_try_catch_element,try_catch_array[i],) */

	assert(sizeof(zend_try_catch_element) == sizeof(src->try_catch_array[i]));
	
	
	
	/* STRUCT_P_EX(zend_try_catch_element,dst->try_catch_array[i],src->try_catch_array[i],try_catch_array[i],,&) */

	
	

	
	assert(sizeof(zend_try_catch_element) == sizeof((& src->try_catch_array[i])[0]));
	
	
	xc_restore_zend_try_catch_element(
		
		
		
		processor, & dst->try_catch_array[i], & src->try_catch_array[i]
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_try_catch) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_try_catch);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT_P(HashTable,static_variables,HashTable_zval_ptr) */

	if (src->static_variables) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->static_variables,src->static_variables,static_variables,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_variables)[0]));
	
	
	
	
	/* allocate */
	
	
	
			ALLOC_HASHTABLE(dst->static_variables);
			
#ifndef NDEBUG
		memset(dst->static_variables, -1, sizeof(HashTable));
#endif

		
	
	
	

	
	xc_restore_HashTable_zval_ptr(
		
		
		
		processor,  dst->static_variables,  src->static_variables
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_variables);
	done_count ++;

#endif




	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start_op);
	done_count ++;

#endif


	
	
	/* DISPATCH(int,backpatch_count) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->backpatch_count);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(zend_bool,done_pass_two) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->done_pass_two);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_this) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_this);
	done_count ++;

#endif



	

#endif

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


#ifdef IS_UNICODE
	
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->script_encoding);
	done_count ++;

#endif


	
#endif
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_U(dst->doc_comment) = (zstr_uchar *) emalloc(sizeof(zstr_uchar) * (src->doc_comment_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_U(dst->doc_comment), -1, sizeof(zstr_uchar) * (src->doc_comment_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_U(dst->doc_comment), ZSTR_U(src->doc_comment), sizeof(zstr_uchar) * (src->doc_comment_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->doc_comment) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->doc_comment_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->doc_comment), -1, sizeof(zstr_char) * (src->doc_comment_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->doc_comment), ZSTR_S(src->doc_comment), sizeof(zstr_char) * (src->doc_comment_len + 1));
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
			
	
	
	
	/* allocate */
	
	
	
			ZSTR_S(dst->doc_comment) = (zstr_char *) emalloc(sizeof(zstr_char) * (src->doc_comment_len + 1));
			
#ifndef NDEBUG
		memset(ZSTR_S(dst->doc_comment), -1, sizeof(zstr_char) * (src->doc_comment_len + 1));
#endif

		
	
	
	

			memcpy(ZSTR_S(dst->doc_comment), ZSTR_S(src->doc_comment), sizeof(zstr_char) * (src->doc_comment_len + 1));
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif

	/* reserved */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->reserved);
	done_count ++;

#endif


#if defined(HARDENING_PATCH) && HARDENING_PATCH
	
	
	/* DISPATCH(zend_bool,created_by_eval) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->created_by_eval);
	done_count ++;

#endif



	

#endif
	} while (0);

#ifdef ZEND_ENGINE_2
		
			do {
				zend_function *parent;
				if (src->prototype != NULL
				 && zend_u_hash_find(&(processor->active_class_entry_dst->parent->function_table),
						UG(unicode) ? IS_UNICODE : IS_STRING,
						src->function_name, xc_zstrlen(UG(unicode), src->function_name) + 1,
						(void **) &parent) == SUCCESS) {
					/* see do_inherit_method_check() */
					if ((parent->common.fn_flags & ZEND_ACC_ABSTRACT)) {
					  dst->prototype = parent;
					} else if (!(parent->common.fn_flags & ZEND_ACC_CTOR) || (parent->common.prototype && (parent->common.prototype->common.scope->ce_flags & ZEND_ACC_INTERFACE))) {
						/* ctors only have a prototype if it comes from an interface */
						dst->prototype = parent->common.prototype ? parent->common.prototype : parent;
					}
					else {
						dst->prototype = NULL;
					}
				}
				else {
					dst->prototype = NULL;
				}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->prototype);
	done_count ++;

#endif


			} while (0);
			
	
#endif

	
#ifdef ZEND_ENGINE_2
		if (src->scope) {
			dst->scope = xc_get_class(processor, (zend_ulong) src->scope);
			xc_fix_method(processor, dst);
		}
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->scope);
	done_count ++;

#endif


#endif
	

	
		if (xc_have_op_array_ctor) {
			zend_llist_apply_with_argument(&zend_extensions, (llist_apply_with_arg_func_t) xc_zend_extension_op_array_ctor_handler, dst TSRMLS_CC);
		}
	

		
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_restore_xc_constinfo_t */
	
 	  	 	static void inline 	xc_restore_xc_constinfo_t( 		 		 		 		xc_processor_t *processor, xc_constinfo_t *dst, const xc_constinfo_t * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_constinfo_t*)NULL)->key) + sizeof(((xc_constinfo_t*)NULL)->key_size) + sizeof(((xc_constinfo_t*)NULL)->constant)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_constinfo_t));
			do {
		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	/* STRUCT(zend_constant,constant) */

	assert(sizeof(zend_constant) == sizeof(src->constant));
	
	
	
	/* STRUCT_P_EX(zend_constant,dst->constant,src->constant,constant,,&) */

	
	

	
	assert(sizeof(zend_constant) == sizeof((& src->constant)[0]));
	
	
	xc_restore_zend_constant(
		
		
		
		processor, & dst->constant, & src->constant
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constant);
	done_count ++;

#endif




		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_restore_xc_funcinfo_t */
	
		/* export:  	  	 	void 	xc_restore_xc_funcinfo_t( 		 		 		 		xc_processor_t *processor, xc_funcinfo_t *dst, const xc_funcinfo_t * const src 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_restore_xc_funcinfo_t( 		 		 		 		xc_processor_t *processor, xc_funcinfo_t *dst, const xc_funcinfo_t * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_funcinfo_t*)NULL)->key) + sizeof(((xc_funcinfo_t*)NULL)->key_size) + sizeof(((xc_funcinfo_t*)NULL)->func)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_funcinfo_t));
			do {
		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	/* STRUCT(zend_function,func) */

	assert(sizeof(zend_function) == sizeof(src->func));
	
	
	
	/* STRUCT_P_EX(zend_function,dst->func,src->func,func,,&) */

	
	

	
	assert(sizeof(zend_function) == sizeof((& src->func)[0]));
	
	
	xc_restore_zend_function(
		
		
		
		processor, & dst->func, & src->func
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->func);
	done_count ++;

#endif




		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_restore_xc_classinfo_t */
	
		/* export:  	  	 	void 	xc_restore_xc_classinfo_t( 		 		 		 		xc_processor_t *processor, xc_classinfo_t *dst, const xc_classinfo_t * const src 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_restore_xc_classinfo_t( 		 		 		 		xc_processor_t *processor, xc_classinfo_t *dst, const xc_classinfo_t * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_classinfo_t*)NULL)->key) + sizeof(((xc_classinfo_t*)NULL)->key_size) + sizeof(((xc_classinfo_t*)NULL)->cest) + sizeof(((xc_classinfo_t*)NULL)->oplineno)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_classinfo_t));
			do {
		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT_P(zend_class_entry,cest) */

	if (src->cest) {
		
		
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof(( src->cest)[0]));
	
	
	
	
	/* allocate */
	
	
	
			dst->cest = (zend_class_entry *) emalloc(sizeof(zend_class_entry));
			
#ifndef NDEBUG
		memset(dst->cest, -1, sizeof(zend_class_entry));
#endif

		
	
	
	

	
	xc_restore_zend_class_entry(
		
		
		
		processor,  dst->cest,  src->cest
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#else
	
	
	/* STRUCT(zend_class_entry,cest) */

	assert(sizeof(zend_class_entry) == sizeof(src->cest));
	
	
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,,&) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof((& src->cest)[0]));
	
	
	xc_restore_zend_class_entry(
		
		
		
		processor, & dst->cest, & src->cest
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(int,oplineno) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->oplineno);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2_1

	
	

/* {{{ xc_restore_xc_autoglobal_t */
	
 	  	 	static void inline 	xc_restore_xc_autoglobal_t( 		 		 		 		xc_processor_t *processor, xc_autoglobal_t *dst, const xc_autoglobal_t * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_autoglobal_t*)NULL)->key) + sizeof(((xc_autoglobal_t*)NULL)->key_len)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_autoglobal_t));
			do {
		

		
			
	
	/* DISPATCH(zend_uint,key_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_len);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif



		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_restore_xc_entry_data_php_t */
	
 	  	 	static void inline 	xc_restore_xc_entry_data_php_t( 		 		 		 		xc_processor_t *processor, xc_entry_data_php_t *dst, const xc_entry_data_php_t * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_php_t*)NULL)->sourcesize) + sizeof(((xc_entry_data_php_t*)NULL)->device) + sizeof(((xc_entry_data_php_t*)NULL)->inode) + sizeof(((xc_entry_data_php_t*)NULL)->mtime) + sizeof(((xc_entry_data_php_t*)NULL)->op_array) + sizeof(((xc_entry_data_php_t*)NULL)->constinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->constinfos) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfos) + sizeof(((xc_entry_data_php_t*)NULL)->classinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->classinfos) + sizeof(((xc_entry_data_php_t*)NULL)->have_early_binding) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobal_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobals)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_entry_data_php_t));
			do {
		

		
			zend_uint i;

#ifdef HAVE_INODE
	
	
	/* DISPATCH(int,device) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->device);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,inode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inode);
	done_count ++;

#endif



	

#endif
	
	
	/* DISPATCH(size_t,sourcesize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->sourcesize);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(time_t,mtime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->mtime);
	done_count ++;

#endif



	


	
	
	/* STRUCT_P(zend_op_array,op_array) */

	if (src->op_array) {
		
		
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof(( src->op_array)[0]));
	
	
	
	
	/* allocate */
	
	
	
			dst->op_array = (zend_op_array *) emalloc(sizeof(zend_op_array));
			
#ifndef NDEBUG
		memset(dst->op_array, -1, sizeof(zend_op_array));
#endif

		
	
	
	

	
	xc_restore_zend_op_array(
		
		
		
		processor,  dst->op_array,  src->op_array
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_array);
	done_count ++;

#endif




#ifdef HAVE_XCACHE_CONSTANT
	
	
	/* DISPATCH(zend_uint,constinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->constinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
			dst->constinfos = (xc_constinfo_t *) emalloc(sizeof(xc_constinfo_t) * src->constinfo_cnt);
			
#ifndef NDEBUG
		memset(dst->constinfos, -1, sizeof(xc_constinfo_t) * src->constinfo_cnt);
#endif

		
	
	
	

			
			for (i = 0; i < src->constinfo_cnt; i ++) {
				
	
	

					
					
	
	/* STRUCT(xc_constinfo_t,constinfos[i],) */

	assert(sizeof(xc_constinfo_t) == sizeof(src->constinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_constinfo_t,dst->constinfos[i],src->constinfos[i],constinfos[i],,&) */

	
	

	
	assert(sizeof(xc_constinfo_t) == sizeof((& src->constinfos[i])[0]));
	
	
	xc_restore_xc_constinfo_t(
		
		
		
		processor, & dst->constinfos[i], & src->constinfos[i]
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif



		
	}

#endif

	
	
	/* DISPATCH(zend_uint,funcinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->funcinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
			dst->funcinfos = (xc_funcinfo_t *) emalloc(sizeof(xc_funcinfo_t) * src->funcinfo_cnt);
			
#ifndef NDEBUG
		memset(dst->funcinfos, -1, sizeof(xc_funcinfo_t) * src->funcinfo_cnt);
#endif

		
	
	
	

			
			for (i = 0; i < src->funcinfo_cnt; i ++) {
				
	
	

					
					
	
	/* STRUCT(xc_funcinfo_t,funcinfos[i],) */

	assert(sizeof(xc_funcinfo_t) == sizeof(src->funcinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_funcinfo_t,dst->funcinfos[i],src->funcinfos[i],funcinfos[i],,&) */

	
	

	
	assert(sizeof(xc_funcinfo_t) == sizeof((& src->funcinfos[i])[0]));
	
	
	xc_restore_xc_funcinfo_t(
		
		
		
		processor, & dst->funcinfos[i], & src->funcinfos[i]
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif



		
	}


	
	
	/* DISPATCH(zend_uint,classinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfo_cnt);
	done_count ++;

#endif



	

	
	
	if (src->classinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
			dst->classinfos = (xc_classinfo_t *) emalloc(sizeof(xc_classinfo_t) * src->classinfo_cnt);
			
#ifndef NDEBUG
		memset(dst->classinfos, -1, sizeof(xc_classinfo_t) * src->classinfo_cnt);
#endif

		
	
	
	

			
			for (i = 0; i < src->classinfo_cnt; i ++) {
				
	
	

					
		
			processor->active_class_num = i + 1;
		
	
					
	
	/* STRUCT(xc_classinfo_t,classinfos[i],) */

	assert(sizeof(xc_classinfo_t) == sizeof(src->classinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_classinfo_t,dst->classinfos[i],src->classinfos[i],classinfos[i],,&) */

	
	

	
	assert(sizeof(xc_classinfo_t) == sizeof((& src->classinfos[i])[0]));
	
	
	xc_restore_xc_classinfo_t(
		
		
		
		processor, & dst->classinfos[i], & src->classinfos[i]
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif



		
	}

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,autoglobal_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobal_cnt);
	done_count ++;

#endif



	

	
	if (src->autoglobals) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
			dst->autoglobals = (xc_autoglobal_t *) emalloc(sizeof(xc_autoglobal_t) * src->autoglobal_cnt);
			
#ifndef NDEBUG
		memset(dst->autoglobals, -1, sizeof(xc_autoglobal_t) * src->autoglobal_cnt);
#endif

		
	
	
	

			
			for (i = 0; i < src->autoglobal_cnt; i ++) {
				
	
	

					
		
			processor->active_class_num = i + 1;
		
	
					
	
	/* STRUCT(xc_autoglobal_t,autoglobals[i],) */

	assert(sizeof(xc_autoglobal_t) == sizeof(src->autoglobals[i]));
	
	
	
	/* STRUCT_P_EX(xc_autoglobal_t,dst->autoglobals[i],src->autoglobals[i],autoglobals[i],,&) */

	
	

	
	assert(sizeof(xc_autoglobal_t) == sizeof((& src->autoglobals[i])[0]));
	
	
	xc_restore_xc_autoglobal_t(
		
		
		
		processor, & dst->autoglobals[i], & src->autoglobals[i]
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif



		
	}

#endif
	
	
	/* DISPATCH(zend_bool,have_early_binding) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_early_binding);
	done_count ++;

#endif



	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_restore_xc_entry_data_var_t */
	
 	  	 	static void inline 	xc_restore_xc_entry_data_var_t( 		 		 		 		xc_processor_t *processor, xc_entry_data_var_t *dst, const xc_entry_data_var_t * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_var_t*)NULL)->value)  ), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_entry_data_var_t));
			do {
		

		
			
	
	
	/* STRUCT_P_EX(zval_ptr,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof((& src->value)[0]));
	
	
	xc_restore_zval_ptr(
		
		
		
		processor, & dst->value, & src->value
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_restore_xc_entry_t */
	
		/* export:  	  	 	void 	xc_restore_xc_entry_t( 		 		 		 		xc_processor_t *processor, xc_entry_t *dst, const xc_entry_t * const src 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_restore_xc_entry_t( 		 		 		 		xc_processor_t *processor, xc_entry_t *dst, const xc_entry_t * const src 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_t*)NULL)->type) + sizeof(((xc_entry_t*)NULL)->hvalue) + sizeof(((xc_entry_t*)NULL)->next) + sizeof(((xc_entry_t*)NULL)->cache) + sizeof(((xc_entry_t*)NULL)->size) + sizeof(((xc_entry_t*)NULL)->refcount) + sizeof(((xc_entry_t*)NULL)->hits) + sizeof(((xc_entry_t*)NULL)->ctime) + sizeof(((xc_entry_t*)NULL)->atime) + sizeof(((xc_entry_t*)NULL)->dtime) + sizeof(((xc_entry_t*)NULL)->ttl) + sizeof(((xc_entry_t*)NULL)->name) + sizeof(((xc_entry_t*)NULL)->data) + sizeof(((xc_entry_t*)NULL)->have_references)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			assert(xc_is_shm(src));
			
			do {
		
#endif


		
			memcpy(dst, src, sizeof(xc_entry_t));
			do {
		

		
		
	
		processor->xce_dst = dst;
		processor->xce_src = src;
	
	
	
	/* DISPATCH(xc_entry_type_t,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(size_t,size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(xc_hash_value_t,hvalue) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hvalue);
	done_count ++;

#endif



	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cache);
	done_count ++;

#endif


	/* skip */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->next);
	done_count ++;

#endif



	
	
	/* DISPATCH(long,refcount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(time_t,ctime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ctime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,atime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->atime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,dtime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->dtime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(long,ttl) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ttl);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_ulong,hits) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hits);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,name_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_type);
	done_count ++;

#endif



	

#endif
		
	
	

#ifdef IS_UNICODE
		if (src->name_type == IS_UNICODE) {
			
	
	/* DISPATCH(int32_t,name.ustr.len) */

	
	
	
	
	

	

		}
		else {
			
	
	/* DISPATCH(int,name.str.len) */

	
	
	
	
	

	

		}
#else
		
	
	/* DISPATCH(int,name.str.len) */

	
	
	
	
	

	

#endif
		
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		
	
	

		switch (src->type) {
		case XC_TYPE_PHP:
			
	
	/* STRUCT_P(xc_entry_data_php_t,data.php) */

	if (src->data.php) {
		
		
	
	/* STRUCT_P_EX(xc_entry_data_php_t,dst->data.php,src->data.php,data.php,) */

	
	

	
	assert(sizeof(xc_entry_data_php_t) == sizeof(( src->data.php)[0]));
	
	
	
	
	/* allocate */
	
	
	
			dst->data.php = (xc_entry_data_php_t *) emalloc(sizeof(xc_entry_data_php_t));
			
#ifndef NDEBUG
		memset(dst->data.php, -1, sizeof(xc_entry_data_php_t));
#endif

		
	
	
	

	
	xc_restore_xc_entry_data_php_t(
		
		
		
		processor,  dst->data.php,  src->data.php
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

			break;
		case XC_TYPE_VAR:
			
	
	/* STRUCT_P(xc_entry_data_var_t,data.var) */

	if (src->data.var) {
		
		
	
	/* STRUCT_P_EX(xc_entry_data_var_t,dst->data.var,src->data.var,data.var,) */

	
	

	
	assert(sizeof(xc_entry_data_var_t) == sizeof(( src->data.var)[0]));
	
	
	
	
	/* allocate */
	
	
	
			dst->data.var = (xc_entry_data_var_t *) emalloc(sizeof(xc_entry_data_var_t));
			
#ifndef NDEBUG
		memset(dst->data.var, -1, sizeof(xc_entry_data_var_t));
#endif

		
	
	
	

	
	xc_restore_xc_entry_data_var_t(
		
		
		
		processor,  dst->data.var,  src->data.var
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

			break;
		default:
			assert(0);
		}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->data);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,have_references) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_references);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_restore_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_restore_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
			} while (0);
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	



 
#ifdef HAVE_XCACHE_DPRINT
  /* {{{ Pre-declare */
 	  	 	void 	xc_dprint_zval( 		const zval * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	
 	  	 	static void inline 	xc_dprint_zval_ptr( 		const zval_ptr * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_dprint_zend_op_array( 		const zend_op_array * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_dprint_zend_class_entry( 		const zend_class_entry * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	
#ifdef HAVE_XCACHE_CONSTANT
 	  	 	static void inline 	xc_dprint_zend_constant( 		const zend_constant * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	
#endif
 	  	 	void 	xc_dprint_zend_function( 		const zend_function * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_dprint_xc_entry_t( 		const xc_entry_t * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	
#ifdef ZEND_ENGINE_2
 	  	 	static void inline 	xc_dprint_zend_property_info( 		const zend_property_info * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	
#endif
/* }}} */
#ifdef IS_CV

	
	

/* {{{ xc_dprint_zend_compiled_variable */
	
 	  	 	static void inline 	xc_dprint_zend_compiled_variable( 		const zend_compiled_variable * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_compiled_variable*)NULL)->name) + sizeof(((zend_compiled_variable*)NULL)->name_len) + sizeof(((zend_compiled_variable*)NULL)->hash_value)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
	
	
	/* DISPATCH(int,name_len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:name_len:\t%d\n", src->name_len);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->name), src->name_len + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->name");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->name_len + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name");
			xc_dprint_str_len(ZSTR_S(src->name), src->name_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name");
			xc_dprint_str_len(ZSTR_S(src->name), src->name_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,hash_value) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "ulong:hash_value:\t%lu\n", src->hash_value);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hash_value);
	done_count ++;

#endif



	


		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_dprint_zend_uint */
	
 	  	 	static void inline 	xc_dprint_zend_uint( 		const zend_uint * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_uint), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "%u\n", src[0]);
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifndef ZEND_ENGINE_2

	
	

/* {{{ xc_dprint_int */
	
 	  	 	static void inline 	xc_dprint_int( 		const int * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(int), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "%d\n", src[0]);
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_dprint_zend_try_catch_element */
	
 	  	 	static void inline 	xc_dprint_zend_try_catch_element( 		const zend_try_catch_element * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_try_catch_element*)NULL)->try_op) + sizeof(((zend_try_catch_element*)NULL)->catch_op)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
	
	
	/* DISPATCH(zend_uint,try_op) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:try_op:\t%u\n", src->try_op);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_op);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,catch_op) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:catch_op:\t%u\n", src->catch_op);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->catch_op);
	done_count ++;

#endif



	


		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif /* ifdef ZEND_ENGINE_2 */

	
	

/* {{{ xc_dprint_zend_brk_cont_element */
	
 	  	 	static void inline 	xc_dprint_zend_brk_cont_element( 		const zend_brk_cont_element * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_brk_cont_element*)NULL)->start) + sizeof(((zend_brk_cont_element*)NULL)->cont) + sizeof(((zend_brk_cont_element*)NULL)->brk) + sizeof(((zend_brk_cont_element*)NULL)->parent)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
#ifdef ZEND_ENGINE_2_2
#ifndef IS_UNICODE
	
	
	/* DISPATCH(int,start) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:start:\t%d\n", src->start);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start);
	done_count ++;

#endif



	

#endif
#endif
	
	
	/* DISPATCH(int,cont) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:cont:\t%d\n", src->cont);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,brk) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:brk:\t%d\n", src->brk);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,parent) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:parent:\t%d\n", src->parent);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif



	


		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	
	

/* {{{ xc_dprint_HashTable_zval_ptr */
	
 	  	 	static void inline 	xc_dprint_HashTable_zval_ptr( 		const HashTable * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "unsigned int:canary:\t%u\n", src->canary);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nTableSize:\t%u\n", src->nTableSize);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nTableMask:\t%u\n", src->nTableMask);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nNumOfElements:\t%u\n", src->nNumOfElements);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "ulong:nNextFreeElement:\t%lu\n", src->nNextFreeElement);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			
			
				xc_dprint_indent(indent);
				fprintf(stderr, "zval_ptr:\"");
				xc_dprint_str_len(BUCKET_KEY_S(b), BUCKET_KEY_SIZE(b));
				fprintf(stderr, "\" %d:h=%lu", BUCKET_KEY_SIZE(b), b->h);
			
			if (sizeof(void *) == sizeof(zval_ptr)) {
				
								
	
	/* STRUCT_P_EX(zval_ptr,pnew->pData,((zval_ptr*)b->pData),,, ) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof((  ((zval_ptr*)b->pData))[0]));
	
	
	xc_dprint_zval_ptr(
		             ((zval_ptr*)b->pData), indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zval_ptr,pnew->pData,((zval_ptr*)b->pData),,) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof(( ((zval_ptr*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_zval_ptr(
		            ((zval_ptr*)b->pData), indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:persistent:\t%u\n", src->persistent);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:unicode:\t%u\n", src->unicode);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "unsigned char:nApplyCount:\t%u\n", src->nApplyCount);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:bApplyProtection:\t%u\n", src->bApplyProtection);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:inconsistent:\t%d\n", src->inconsistent);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	
	

/* {{{ xc_dprint_HashTable_zend_constant */
	
 	  	 	static void inline 	xc_dprint_HashTable_zend_constant( 		const HashTable * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "unsigned int:canary:\t%u\n", src->canary);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nTableSize:\t%u\n", src->nTableSize);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nTableMask:\t%u\n", src->nTableMask);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nNumOfElements:\t%u\n", src->nNumOfElements);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "ulong:nNextFreeElement:\t%lu\n", src->nNextFreeElement);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			
			
				xc_dprint_indent(indent);
				fprintf(stderr, "zend_constant:\"");
				xc_dprint_str_len(BUCKET_KEY_S(b), BUCKET_KEY_SIZE(b));
				fprintf(stderr, "\" %d:h=%lu", BUCKET_KEY_SIZE(b), b->h);
			
			if (sizeof(void *) == sizeof(zend_constant)) {
				
								
	
	/* STRUCT_P_EX(zend_constant,pnew->pData,((zend_constant*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_constant) == sizeof((  ((zend_constant*)b->pData))[0]));
	
	
	xc_dprint_zend_constant(
		             ((zend_constant*)b->pData), indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_constant,pnew->pData,((zend_constant*)b->pData),,) */

	
	

	
	assert(sizeof(zend_constant) == sizeof(( ((zend_constant*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_zend_constant(
		            ((zend_constant*)b->pData), indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:persistent:\t%u\n", src->persistent);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:unicode:\t%u\n", src->unicode);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "unsigned char:nApplyCount:\t%u\n", src->nApplyCount);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:bApplyProtection:\t%u\n", src->bApplyProtection);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:inconsistent:\t%d\n", src->inconsistent);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	
	

/* {{{ xc_dprint_HashTable_zend_function */
	
		/* export:  	  	 	void 	xc_dprint_HashTable_zend_function( 		const HashTable * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dprint_HashTable_zend_function( 		const HashTable * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "unsigned int:canary:\t%u\n", src->canary);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nTableSize:\t%u\n", src->nTableSize);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nTableMask:\t%u\n", src->nTableMask);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nNumOfElements:\t%u\n", src->nNumOfElements);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "ulong:nNextFreeElement:\t%lu\n", src->nNextFreeElement);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			
			
				xc_dprint_indent(indent);
				fprintf(stderr, "zend_function:\"");
				xc_dprint_str_len(BUCKET_KEY_S(b), BUCKET_KEY_SIZE(b));
				fprintf(stderr, "\" %d:h=%lu", BUCKET_KEY_SIZE(b), b->h);
			
			if (sizeof(void *) == sizeof(zend_function)) {
				
								
	
	/* STRUCT_P_EX(zend_function,pnew->pData,((zend_function*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_function) == sizeof((  ((zend_function*)b->pData))[0]));
	
	
	xc_dprint_zend_function(
		             ((zend_function*)b->pData), indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_function,pnew->pData,((zend_function*)b->pData),,) */

	
	

	
	assert(sizeof(zend_function) == sizeof(( ((zend_function*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_zend_function(
		            ((zend_function*)b->pData), indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:persistent:\t%u\n", src->persistent);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:unicode:\t%u\n", src->unicode);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "unsigned char:nApplyCount:\t%u\n", src->nApplyCount);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:bApplyProtection:\t%u\n", src->bApplyProtection);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:inconsistent:\t%d\n", src->inconsistent);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef ZEND_ENGINE_2

	
	
	

/* {{{ xc_dprint_HashTable_zend_property_info */
	
 	  	 	static void inline 	xc_dprint_HashTable_zend_property_info( 		const HashTable * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		
					
								
	
	/* DISPATCH(unsigned int,canary) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "unsigned int:canary:\t%u\n", src->canary);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif



	

			
		
#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nTableSize:\t%u\n", src->nTableSize);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nTableMask:\t%u\n", src->nTableMask);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:nNumOfElements:\t%u\n", src->nNumOfElements);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "ulong:nNextFreeElement:\t%lu\n", src->nNextFreeElement);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			
			
				xc_dprint_indent(indent);
				fprintf(stderr, "zend_property_info:\"");
				xc_dprint_str_len(BUCKET_KEY_S(b), BUCKET_KEY_SIZE(b));
				fprintf(stderr, "\" %d:h=%lu", BUCKET_KEY_SIZE(b), b->h);
			
			if (sizeof(void *) == sizeof(zend_property_info)) {
				
								
	
	/* STRUCT_P_EX(zend_property_info,pnew->pData,((zend_property_info*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_property_info) == sizeof((  ((zend_property_info*)b->pData))[0]));
	
	
	xc_dprint_zend_property_info(
		             ((zend_property_info*)b->pData), indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_property_info,pnew->pData,((zend_property_info*)b->pData),,) */

	
	

	
	assert(sizeof(zend_property_info) == sizeof(( ((zend_property_info*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_zend_property_info(
		            ((zend_property_info*)b->pData), indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:persistent:\t%u\n", src->persistent);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:unicode:\t%u\n", src->unicode);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "unsigned char:nApplyCount:\t%u\n", src->nApplyCount);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:bApplyProtection:\t%u\n", src->bApplyProtection);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:inconsistent:\t%d\n", src->inconsistent);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	

/* {{{ xc_dprint_zval */
	
		/* export:  	  	 	void 	xc_dprint_zval( 		const zval * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dprint_zval( 		const zval * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zval*)NULL)->value) + sizeof(((zval*)NULL)->refcount) + sizeof(((zval*)NULL)->type) + sizeof(((zval*)NULL)->is_ref)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
				/* Variable information */
		
	
	

		switch (src->type & ~IS_CONSTANT_INDEX) {
			case IS_LONG:
			case IS_RESOURCE:
			case IS_BOOL:
				
	
	/* DISPATCH(long,value.lval) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "long:value.lval:\t%ld\n", src->value.lval);
	
	
	

	

				break;
			case IS_DOUBLE:
				
	
	/* DISPATCH(double,value.dval) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "double:value.dval:\t%f\n", src->value.dval);
	
	
	

	

				break;
			case IS_NULL:
				xc_dprint_indent(indent);fprintf(stderr, "\tNULL\n");
				break;

			case IS_CONSTANT:
#ifdef IS_UNICODE
				if (UG(unicode)) {
					goto proc_unicode;
				}
#endif
			case IS_STRING:
#ifdef FLAG_IS_BC
			case FLAG_IS_BC:
#endif
				
	
	/* DISPATCH(int,value.str.len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:value.str.len:\t%d\n", src->value.str.len);
	
	
	

	

				
	/* PROC_STRING_L(value.str.val,value.str.len) */
 
	/* PROC_STRING_N(value.str.val,value.str.len + 1) */
 
	
	
	
	
	
	
	
	if (src->value.str.val == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->value.str.val");
			xc_dprint_str_len(src->value.str.val, src->value.str.len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->value.str.len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

				break;
#ifdef IS_UNICODE
			case IS_UNICODE:
proc_unicode:
				
	
	/* DISPATCH(int32_t,value.uni.len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int32_t:value.uni.len:\t%d\n", src->value.uni.len);
	
	
	

	

				
	/* PROC_ZSTRING_L(1,value.uni.val,value.uni.len) */
 
	
	/* PROC_ZSTRING_N(1,value.uni.val,value.uni.len,len) */

#ifdef IS_UNICODE
	
	
	
	
	
	
	
	
	
	
	if (ZSTR_U(src->value.uni.val) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->value.uni.val), src->value.uni.len + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->value.uni.val");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->value.uni.len + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

	
#else
	
	
	
	
	
	
	
	
	
	if (ZSTR_S(src->value.uni.val) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->value.uni.val");
			xc_dprint_str_len(ZSTR_S(src->value.uni.val), src->value.uni.len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->value.uni.len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

				break;
#endif

			case IS_ARRAY:
			case IS_CONSTANT_ARRAY:
				
	
	/* STRUCT_P(HashTable,value.ht,HashTable_zval_ptr) */

	if (src->value.ht) {
		xc_dprint_indent(indent);fprintf(stderr, "HashTable:value.ht");
		
	
	/* STRUCT_P_EX(HashTable,dst->value.ht,src->value.ht,value.ht,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->value.ht)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_HashTable_zval_ptr(
		            src->value.ht, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		xc_dprint_indent(indent);fprintf(stderr, "HashTable:value.ht:\tNULL\n");
		
	
	

	}
	

				break;

			case IS_OBJECT:
				
				#ifndef ZEND_ENGINE_2
				
	
	/* STRUCT_P(zend_class_entry,value.obj.ce) */

	if (src->value.obj.ce) {
		xc_dprint_indent(indent);fprintf(stderr, "zend_class_entry:value.obj.ce");
		
	
	/* STRUCT_P_EX(zend_class_entry,dst->value.obj.ce,src->value.obj.ce,value.obj.ce,) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof(( src->value.obj.ce)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_zend_class_entry(
		            src->value.obj.ce, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		xc_dprint_indent(indent);fprintf(stderr, "zend_class_entry:value.obj.ce:\tNULL\n");
		
	
	

	}
	

				
	
	/* STRUCT_P(HashTable,value.obj.properties,HashTable_zval_ptr) */

	if (src->value.obj.properties) {
		xc_dprint_indent(indent);fprintf(stderr, "HashTable:value.obj.properties");
		
	
	/* STRUCT_P_EX(HashTable,dst->value.obj.properties,src->value.obj.properties,value.obj.properties,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->value.obj.properties)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_HashTable_zval_ptr(
		            src->value.obj.properties, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		xc_dprint_indent(indent);fprintf(stderr, "HashTable:value.obj.properties:\tNULL\n");
		
	
	

	}
	

#endif
				break;

			default:
				assert(0);
		}
		
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif


		
	
	/* DISPATCH(zval_data_type,type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zval_data_type:type:\t%d %s\n", src->type, xc_get_data_type(src->type));
		
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_uchar,is_ref) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uchar:is_ref:\t%u\n", src->is_ref);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->is_ref);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_ushort,refcount) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_ushort:refcount:\t%hu\n", src->refcount);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_3
		
	
	/* DISPATCH(zend_uchar,idx_type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uchar:idx_type:\t%u\n", src->idx_type);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->idx_type);
	done_count ++;

#endif



	

#endif
	
		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dprint_zval_ptr */
	
 	  	 	static void inline 	xc_dprint_zval_ptr( 		const zval_ptr * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zval_ptr), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
		do {
			
			
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			
			xc_dprint_indent(indent);fprintf(stderr, "[%p]", src[0]);
			
	
	/* STRUCT_P_EX(zval,dst[0],src[0],[0],, ) */

	
	

	
	assert(sizeof(zval) == sizeof((  src[0])[0]));
	
	
	xc_dprint_zval(
		             src[0], indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

			
		} while (0);
	
	
#ifndef NDEBUG
		
	done_size += sizeof(zval_ptr);
	done_count ++;

#endif


		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_dprint_zend_arg_info */
	
 	  	 	static void inline 	xc_dprint_zend_arg_info( 		const zend_arg_info * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_arg_info*)NULL)->name) + sizeof(((zend_arg_info*)NULL)->name_len) + sizeof(((zend_arg_info*)NULL)->class_name) + sizeof(((zend_arg_info*)NULL)->class_name_len) + sizeof(((zend_arg_info*)NULL)->array_type_hint) + sizeof(((zend_arg_info*)NULL)->allow_null) + sizeof(((zend_arg_info*)NULL)->pass_by_reference) + sizeof(((zend_arg_info*)NULL)->return_reference) + sizeof(((zend_arg_info*)NULL)->required_num_args)  ), assert_count = 9;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
	
	
	/* DISPATCH(zend_uint,name_len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:name_len:\t%u\n", src->name_len);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->name), src->name_len + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->name");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->name_len + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name");
			xc_dprint_str_len(ZSTR_S(src->name), src->name_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name");
			xc_dprint_str_len(ZSTR_S(src->name), src->name_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_uint,class_name_len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:class_name_len:\t%u\n", src->class_name_len);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,class_name,class_name_len) */
 
	
	/* PROC_ZSTRING_N(,class_name,class_name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->class_name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->class_name), src->class_name_len + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->class_name");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->class_name_len + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->class_name");
			xc_dprint_str_len(ZSTR_S(src->class_name), src->class_name_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->class_name_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->class_name");
			xc_dprint_str_len(ZSTR_S(src->class_name), src->class_name_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->class_name_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_bool,array_type_hint) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:array_type_hint:\t%u\n", src->array_type_hint);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->array_type_hint);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,allow_null) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:allow_null:\t%u\n", src->allow_null);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->allow_null);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_by_reference) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:pass_by_reference:\t%u\n", src->pass_by_reference);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_by_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,return_reference) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:return_reference:\t%u\n", src->return_reference);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,required_num_args) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:required_num_args:\t%d\n", src->required_num_args);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	


		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_dprint_zend_constant */
	
 	  	 	static void inline 	xc_dprint_zend_constant( 		const zend_constant * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_constant*)NULL)->value) + sizeof(((zend_constant*)NULL)->flags) + sizeof(((zend_constant*)NULL)->name) + sizeof(((zend_constant*)NULL)->name_len) + sizeof(((zend_constant*)NULL)->module_number)  ), assert_count = 5;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
	
	/* STRUCT(zval,value) */

	assert(sizeof(zval) == sizeof(src->value));
	xc_dprint_indent(indent);fprintf(stderr, "zval:value");
	
	
	/* STRUCT_P_EX(zval,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->value)[0]));
	
	
	xc_dprint_zval(
		           & src->value, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



	
	
	/* DISPATCH(int,flags) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:flags:\t%d\n", src->flags);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,name_len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:name_len:\t%u\n", src->name_len);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	
	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->name), src->name_len + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->name");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->name_len + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name");
			xc_dprint_str_len(ZSTR_S(src->name), src->name_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name");
			xc_dprint_str_len(ZSTR_S(src->name), src->name_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	
	/* DISPATCH(int,module_number) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:module_number:\t%d\n", src->module_number);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module_number);
	done_count ++;

#endif



	


		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_dprint_zend_function */
	
		/* export:  	  	 	void 	xc_dprint_zend_function( 		const zend_function * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dprint_zend_function( 		const zend_function * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_function), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
	
	

	switch (src->type) {
	case ZEND_INTERNAL_FUNCTION:
	case ZEND_OVERLOADED_FUNCTION:
		
		break;

	case ZEND_USER_FUNCTION:
	case ZEND_EVAL_CODE:
		
		
	
	/* STRUCT(zend_op_array,op_array) */

	assert(sizeof(zend_op_array) == sizeof(src->op_array));
	xc_dprint_indent(indent);fprintf(stderr, "zend_op_array:op_array");
	
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,,&) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof((& src->op_array)[0]));
	
	
	xc_dprint_zend_op_array(
		           & src->op_array, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

		break;

	default:
		assert(0);
	}
	
	
	

	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_dprint_zend_property_info */
	
 	  	 	static void inline 	xc_dprint_zend_property_info( 		const zend_property_info * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_property_info*)NULL)->flags) + sizeof(((zend_property_info*)NULL)->name) + sizeof(((zend_property_info*)NULL)->name_length) + sizeof(((zend_property_info*)NULL)->h) + sizeof(((zend_property_info*)NULL)->doc_comment) + sizeof(((zend_property_info*)NULL)->doc_comment_len) + sizeof(((zend_property_info*)NULL)->ce)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
	
	
	/* DISPATCH(zend_uint,flags) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:flags:\t%u\n", src->flags);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,name_length) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:name_length:\t%d\n", src->name_length);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->name), src->name_length + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->name");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->name_length + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name");
			xc_dprint_str_len(ZSTR_S(src->name), src->name_length + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name_length + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name");
			xc_dprint_str_len(ZSTR_S(src->name), src->name_length + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name_length + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,h) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "ulong:h:\t%lu\n", src->h);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->h);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:doc_comment_len:\t%d\n", src->doc_comment_len);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->doc_comment), src->doc_comment_len + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->doc_comment");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->doc_comment_len + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->doc_comment");
			xc_dprint_str_len(ZSTR_S(src->doc_comment), src->doc_comment_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->doc_comment_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->doc_comment");
			xc_dprint_str_len(ZSTR_S(src->doc_comment), src->doc_comment_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->doc_comment_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif
	#if defined(ZEND_ENGINE_2_2) && !defined(IS_UNICODE)
	
	if (src->ce) {
		
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce);
	done_count ++;

#endif


#endif

		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_dprint_zend_class_entry */
	
		/* export:  	  	 	void 	xc_dprint_zend_class_entry( 		const zend_class_entry * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dprint_zend_class_entry( 		const zend_class_entry * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_class_entry*)NULL)->type) + sizeof(((zend_class_entry*)NULL)->name) + sizeof(((zend_class_entry*)NULL)->name_length) + sizeof(((zend_class_entry*)NULL)->parent) + sizeof(((zend_class_entry*)NULL)->refcount) + sizeof(((zend_class_entry*)NULL)->constants_updated) + sizeof(((zend_class_entry*)NULL)->ce_flags) + sizeof(((zend_class_entry*)NULL)->function_table) + sizeof(((zend_class_entry*)NULL)->default_properties) + sizeof(((zend_class_entry*)NULL)->properties_info) + sizeof(((zend_class_entry*)NULL)->default_static_members) + sizeof(((zend_class_entry*)NULL)->static_members) + sizeof(((zend_class_entry*)NULL)->constants_table) + sizeof(((zend_class_entry*)NULL)->builtin_functions) + sizeof(((zend_class_entry*)NULL)->constructor) + sizeof(((zend_class_entry*)NULL)->destructor) + sizeof(((zend_class_entry*)NULL)->clone) + sizeof(((zend_class_entry*)NULL)->__get) + sizeof(((zend_class_entry*)NULL)->__set) + sizeof(((zend_class_entry*)NULL)->__unset) + sizeof(((zend_class_entry*)NULL)->__isset) + sizeof(((zend_class_entry*)NULL)->__call) + sizeof(((zend_class_entry*)NULL)->__tostring) + sizeof(((zend_class_entry*)NULL)->serialize_func) + sizeof(((zend_class_entry*)NULL)->unserialize_func) + sizeof(((zend_class_entry*)NULL)->iterator_funcs) + sizeof(((zend_class_entry*)NULL)->create_object) + sizeof(((zend_class_entry*)NULL)->get_iterator) + sizeof(((zend_class_entry*)NULL)->interface_gets_implemented) + sizeof(((zend_class_entry*)NULL)->serialize) + sizeof(((zend_class_entry*)NULL)->unserialize) + sizeof(((zend_class_entry*)NULL)->interfaces) + sizeof(((zend_class_entry*)NULL)->num_interfaces) + sizeof(((zend_class_entry*)NULL)->filename) + sizeof(((zend_class_entry*)NULL)->line_start) + sizeof(((zend_class_entry*)NULL)->line_end) + sizeof(((zend_class_entry*)NULL)->doc_comment) + sizeof(((zend_class_entry*)NULL)->doc_comment_len) + sizeof(((zend_class_entry*)NULL)->module)  ), assert_count = 39;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
	
	
	/* DISPATCH(char,type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "char:type:\t%d\n", src->type);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,name_length) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:name_length:\t%u\n", src->name_length);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->name), src->name_length + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->name");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->name_length + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name");
			xc_dprint_str_len(ZSTR_S(src->name), src->name_length + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name_length + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name");
			xc_dprint_str_len(ZSTR_S(src->name), src->name_length + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name_length + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
		
	if (src->parent) {
		
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif


	
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(int,refcount) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:refcount:\t%d\n", src->refcount);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	

#else
	
	
	/* STRUCT_P(int,refcount) */

	if (src->refcount) {
		xc_dprint_indent(indent);fprintf(stderr, "int:refcount");
		
	
	/* STRUCT_P_EX(int,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(int) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_int(
		            src->refcount, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		xc_dprint_indent(indent);fprintf(stderr, "int:refcount:\tNULL\n");
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(zend_bool,constants_updated) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:constants_updated:\t%u\n", src->constants_updated);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_updated);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,ce_flags) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:ce_flags:\t%u\n", src->ce_flags);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce_flags);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT(HashTable,default_properties,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_properties));
	xc_dprint_indent(indent);fprintf(stderr, "HashTable:default_properties");
	
	
	/* STRUCT_P_EX(HashTable,dst->default_properties,src->default_properties,default_properties,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_properties)[0]));
	
	
	xc_dprint_HashTable_zval_ptr(
		           & src->default_properties, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_properties);
	done_count ++;

#endif



	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->builtin_functions);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT(HashTable,properties_info,HashTable_zend_property_info) */

	assert(sizeof(HashTable) == sizeof(src->properties_info));
	xc_dprint_indent(indent);fprintf(stderr, "HashTable:properties_info");
	
	
	/* STRUCT_P_EX(HashTable,dst->properties_info,src->properties_info,properties_info,HashTable_zend_property_info,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->properties_info)[0]));
	
	
	xc_dprint_HashTable_zend_property_info(
		           & src->properties_info, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->properties_info);
	done_count ++;

#endif



#	ifdef ZEND_ENGINE_2_1
	
	
	/* STRUCT(HashTable,default_static_members,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_static_members));
	xc_dprint_indent(indent);fprintf(stderr, "HashTable:default_static_members");
	
	
	/* STRUCT_P_EX(HashTable,dst->default_static_members,src->default_static_members,default_static_members,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_static_members)[0]));
	
	
	xc_dprint_HashTable_zval_ptr(
		           & src->default_static_members, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_static_members);
	done_count ++;

#endif



	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif


#	else
	
	
	/* STRUCT_P(HashTable,static_members,HashTable_zval_ptr) */

	if (src->static_members) {
		xc_dprint_indent(indent);fprintf(stderr, "HashTable:static_members");
		
	
	/* STRUCT_P_EX(HashTable,dst->static_members,src->static_members,static_members,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_members)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_HashTable_zval_ptr(
		            src->static_members, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		xc_dprint_indent(indent);fprintf(stderr, "HashTable:static_members:\tNULL\n");
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif



#	endif
	
	
	/* STRUCT(HashTable,constants_table,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->constants_table));
	xc_dprint_indent(indent);fprintf(stderr, "HashTable:constants_table");
	
	
	/* STRUCT_P_EX(HashTable,dst->constants_table,src->constants_table,constants_table,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->constants_table)[0]));
	
	
	xc_dprint_HashTable_zval_ptr(
		           & src->constants_table, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_table);
	done_count ++;

#endif




		
	
	
		
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interfaces);
	done_count ++;

#endif


		
	
	
	
	/* DISPATCH(zend_uint,num_interfaces) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:num_interfaces:\t%u\n", src->num_interfaces);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_interfaces);
	done_count ++;

#endif



	


	
	/* PROC_STRING(filename) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->filename == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->filename");
			xc_dprint_str_len(src->filename, strlen(src->filename) + 1 - 1);
			fprintf(stderr, "\" len=%d\n", strlen(src->filename) + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:line_start:\t%u\n", src->line_start);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:line_end:\t%u\n", src->line_end);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,doc_comment_len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:doc_comment_len:\t%u\n", src->doc_comment_len);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->doc_comment), src->doc_comment_len + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->doc_comment");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->doc_comment_len + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->doc_comment");
			xc_dprint_str_len(ZSTR_S(src->doc_comment), src->doc_comment_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->doc_comment_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->doc_comment");
			xc_dprint_str_len(ZSTR_S(src->doc_comment), src->doc_comment_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->doc_comment_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->iterator_funcs);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->create_object);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_iterator);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interface_gets_implemented);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2_3
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_static_method);
	done_count ++;

#endif


#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize);
	done_count ++;

#endif


	/* deal with it inside xc_fix_method */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->destructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->clone);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__set);
	done_count ++;

#endif


/* should be >5.1 */
#ifdef ZEND_ENGINE_2_1
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__unset);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__isset);
	done_count ++;

#endif


# if defined(ZEND_ENGINE_2_2) || PHP_MAJOR_VERSION >= 6
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__tostring);
	done_count ++;

#endif


# endif
#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__call);
	done_count ++;

#endif


#ifdef ZEND_CALLSTATIC_FUNC_NAME
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__callstatic);
	done_count ++;

#endif


#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module);
	done_count ++;

#endif


#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_function_call);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_set);
	done_count ++;

#endif


#endif
		
	
	/* STRUCT(HashTable,function_table,HashTable_zend_function) */

	assert(sizeof(HashTable) == sizeof(src->function_table));
	xc_dprint_indent(indent);fprintf(stderr, "HashTable:function_table");
	
	
	/* STRUCT_P_EX(HashTable,dst->function_table,src->function_table,function_table,HashTable_zend_function,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->function_table)[0]));
	
	
	xc_dprint_HashTable_zend_function(
		           & src->function_table, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_table);
	done_count ++;

#endif



	
	

		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dprint_znode */
	
 	  	 	static void inline 	xc_dprint_znode( 		const znode * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((znode*)NULL)->op_type) + sizeof(((znode*)NULL)->u)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
	
	/* DISPATCH(int,op_type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:op_type:\t%d %s\n", src->op_type, xc_get_op_type(src->op_type));
		
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_type);
	done_count ++;

#endif



	


#ifdef IS_CV
#	define XCACHE_IS_CV IS_CV
#else
/* compatible with zend optimizer */
#	define XCACHE_IS_CV 16
#endif
	assert(src->op_type == IS_CONST ||
		src->op_type == IS_VAR ||
		src->op_type == XCACHE_IS_CV ||
		src->op_type == IS_TMP_VAR ||
		src->op_type == IS_UNUSED);
#undef XCACHE_IS_CV
		
	
	

	switch (src->op_type) {
		case IS_CONST:
			
	
	/* STRUCT(zval,u.constant) */

	assert(sizeof(zval) == sizeof(src->u.constant));
	xc_dprint_indent(indent);fprintf(stderr, "zval:u.constant");
	
	
	/* STRUCT_P_EX(zval,dst->u.constant,src->u.constant,u.constant,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->u.constant)[0]));
	
	
	xc_dprint_zval(
		           & src->u.constant, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

			break;
		
		case IS_VAR:
		case IS_TMP_VAR:
#ifdef IS_CV
		case IS_CV:
#else
		case 16:
#endif
			
	
	/* DISPATCH(zend_uint,u.var) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:u.var:\t%u\n", src->u.var);
	
	
	

	

			
	
	/* DISPATCH(zend_uint,u.EA.type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:u.EA.type:\t%u\n", src->u.EA.type);
	
	
	

	

			break;
		case IS_UNUSED:
			
			
	
	/* DISPATCH(zend_uint,u.opline_num) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:u.opline_num:\t%u\n", src->u.opline_num);
	
	
	

	

#ifndef ZEND_ENGINE_2
			
	
	/* DISPATCH(zend_uint,u.fetch_type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:u.fetch_type:\t%u\n", src->u.fetch_type);
	
	
	

	

#endif
			
	
	/* DISPATCH(zend_uint,u.EA.type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:u.EA.type:\t%u\n", src->u.EA.type);
	
	
	

	

			break;
		
	}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->u);
	done_count ++;

#endif



		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dprint_zend_op */
	
		/* export:  	  	 	void 	xc_dprint_zend_op( 		const zend_op * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dprint_zend_op( 		const zend_op * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op*)NULL)->handler) + sizeof(((zend_op*)NULL)->result) + sizeof(((zend_op*)NULL)->op1) + sizeof(((zend_op*)NULL)->op2) + sizeof(((zend_op*)NULL)->extended_value) + sizeof(((zend_op*)NULL)->lineno) + sizeof(((zend_op*)NULL)->opcode)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
	
	/* DISPATCH(zend_uchar,opcode) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uchar:opcode:\t%d %s\n", src->opcode, xc_get_opcode(src->opcode));
		
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcode);
	done_count ++;

#endif



	

	
	
	/* STRUCT(znode,result) */

	assert(sizeof(znode) == sizeof(src->result));
	xc_dprint_indent(indent);fprintf(stderr, "znode:result");
	
	
	/* STRUCT_P_EX(znode,dst->result,src->result,result,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->result)[0]));
	
	
	xc_dprint_znode(
		           & src->result, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->result);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op1) */

	assert(sizeof(znode) == sizeof(src->op1));
	xc_dprint_indent(indent);fprintf(stderr, "znode:op1");
	
	
	/* STRUCT_P_EX(znode,dst->op1,src->op1,op1,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op1)[0]));
	
	
	xc_dprint_znode(
		           & src->op1, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op1);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op2) */

	assert(sizeof(znode) == sizeof(src->op2));
	xc_dprint_indent(indent);fprintf(stderr, "znode:op2");
	
	
	/* STRUCT_P_EX(znode,dst->op2,src->op2,op2,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op2)[0]));
	
	
	xc_dprint_znode(
		           & src->op2, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op2);
	done_count ++;

#endif



	
	
	/* DISPATCH(ulong,extended_value) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "ulong:extended_value:\t%lu\n", src->extended_value);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->extended_value);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,lineno) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "uint:lineno:\t%u\n", src->lineno);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->lineno);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	
	/* DISPATCH(opcode_handler_t,handler) */

	/* is copying enough? */
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handler);
	done_count ++;

#endif


	

#endif

		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dprint_zend_op_array */
	
		/* export:  	  	 	void 	xc_dprint_zend_op_array( 		const zend_op_array * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dprint_zend_op_array( 		const zend_op_array * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op_array*)NULL)->type) + sizeof(((zend_op_array*)NULL)->function_name) + sizeof(((zend_op_array*)NULL)->scope) + sizeof(((zend_op_array*)NULL)->fn_flags) + sizeof(((zend_op_array*)NULL)->prototype) + sizeof(((zend_op_array*)NULL)->num_args) + sizeof(((zend_op_array*)NULL)->required_num_args) + sizeof(((zend_op_array*)NULL)->arg_info) + sizeof(((zend_op_array*)NULL)->pass_rest_by_reference) + sizeof(((zend_op_array*)NULL)->return_reference) + sizeof(((zend_op_array*)NULL)->refcount) + sizeof(((zend_op_array*)NULL)->opcodes) + sizeof(((zend_op_array*)NULL)->last) + sizeof(((zend_op_array*)NULL)->size) + sizeof(((zend_op_array*)NULL)->vars) + sizeof(((zend_op_array*)NULL)->last_var) + sizeof(((zend_op_array*)NULL)->size_var) + sizeof(((zend_op_array*)NULL)->T) + sizeof(((zend_op_array*)NULL)->brk_cont_array) + sizeof(((zend_op_array*)NULL)->last_brk_cont) + sizeof(((zend_op_array*)NULL)->current_brk_cont) + sizeof(((zend_op_array*)NULL)->try_catch_array) + sizeof(((zend_op_array*)NULL)->last_try_catch) + sizeof(((zend_op_array*)NULL)->static_variables) + sizeof(((zend_op_array*)NULL)->start_op) + sizeof(((zend_op_array*)NULL)->backpatch_count) + sizeof(((zend_op_array*)NULL)->done_pass_two) + sizeof(((zend_op_array*)NULL)->uses_this) + sizeof(((zend_op_array*)NULL)->filename) + sizeof(((zend_op_array*)NULL)->line_start) + sizeof(((zend_op_array*)NULL)->line_end) + sizeof(((zend_op_array*)NULL)->doc_comment) + sizeof(((zend_op_array*)NULL)->doc_comment_len) + sizeof(((zend_op_array*)NULL)->reserved)  ), assert_count = 34;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
	do {
		zend_uint ii;
	int i;

	/* Common elements */
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uchar:type:\t%u\n", src->type);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING(,function_name) */
 
	
	/* PROC_ZSTRING_N(,function_name,,strlen) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->function_name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->function_name), xc_zstrlen_uchar (src->function_name) + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->function_name");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", xc_zstrlen_uchar (src->function_name) + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->function_name");
			xc_dprint_str_len(ZSTR_S(src->function_name), xc_zstrlen_char (src->function_name) + 1 - 1);
			fprintf(stderr, "\" len=%d\n", xc_zstrlen_char (src->function_name) + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->function_name");
			xc_dprint_str_len(ZSTR_S(src->function_name), xc_zstrlen_char (src->function_name) + 1 - 1);
			fprintf(stderr, "\" len=%d\n", xc_zstrlen_char (src->function_name) + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,fn_flags) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:fn_flags:\t%u\n", src->fn_flags);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->fn_flags);
	done_count ++;

#endif



	

	


	if (src->arg_info) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (ii = 0; ii < src->num_args; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_arg_info,arg_info[i],) */

	assert(sizeof(zend_arg_info) == sizeof(src->arg_info[ii]));
	xc_dprint_indent(indent);fprintf(stderr, "zend_arg_info:arg_info[ii]");
	
	
	/* STRUCT_P_EX(zend_arg_info,dst->arg_info[i],src->arg_info[i],arg_info[i],,&) */

	
	

	
	assert(sizeof(zend_arg_info) == sizeof((& src->arg_info[ii])[0]));
	
	
	xc_dprint_zend_arg_info(
		           & src->arg_info[ii], indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,num_args) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:num_args:\t%u\n", src->num_args);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,required_num_args) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:required_num_args:\t%u\n", src->required_num_args);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_rest_by_reference) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:pass_rest_by_reference:\t%u\n", src->pass_rest_by_reference);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_rest_by_reference);
	done_count ++;

#endif



	

#else
	if (src->arg_types) {
		
	
	
	
	/* allocate */
	
	
	
	
	

		
		
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif


	}
	else {
		
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif



		
	}
#endif
	
	
	/* DISPATCH(unsigned char,return_reference) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "unsigned char:return_reference:\t%u\n", src->return_reference);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	/* END of common elements */
#ifdef IS_UNICODE
	#endif

	
	
	/* STRUCT_P(zend_uint,refcount) */

	if (src->refcount) {
		xc_dprint_indent(indent);fprintf(stderr, "zend_uint:refcount");
		
	
	/* STRUCT_P_EX(zend_uint,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(zend_uint) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_zend_uint(
		            src->refcount, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		xc_dprint_indent(indent);fprintf(stderr, "zend_uint:refcount:\tNULL\n");
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	
	

	
	


	if (src->opcodes) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (ii = 0; ii < src->last; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_op,opcodes[i],) */

	assert(sizeof(zend_op) == sizeof(src->opcodes[ii]));
	xc_dprint_indent(indent);fprintf(stderr, "zend_op:opcodes[ii]");
	
	
	/* STRUCT_P_EX(zend_op,dst->opcodes[i],src->opcodes[i],opcodes[i],,&) */

	
	

	
	assert(sizeof(zend_op) == sizeof((& src->opcodes[ii])[0]));
	
	
	xc_dprint_zend_op(
		           & src->opcodes[ii], indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif



		
	}



	
	
	
	/* DISPATCH(zend_uint,last) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:last:\t%u\n", src->last);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,size) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:size:\t%u\n", src->size);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



	


#ifdef IS_CV
	
	if (src->vars) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->last_var; i ++) {
				
	
	

					
					
	
	/* STRUCT(zend_compiled_variable,vars[i],) */

	assert(sizeof(zend_compiled_variable) == sizeof(src->vars[i]));
	xc_dprint_indent(indent);fprintf(stderr, "zend_compiled_variable:vars[i]");
	
	
	/* STRUCT_P_EX(zend_compiled_variable,dst->vars[i],src->vars[i],vars[i],,&) */

	
	

	
	assert(sizeof(zend_compiled_variable) == sizeof((& src->vars[i])[0]));
	
	
	xc_dprint_zend_compiled_variable(
		           & src->vars[i], indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_var) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:last_var:\t%d\n", src->last_var);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_var);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,size_var) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:size_var:\t%u\n", src->size_var);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size_var);
	done_count ++;

#endif



	

#else
		
#endif

	
	
	/* DISPATCH(zend_uint,T) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:T:\t%u\n", src->T);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->T);
	done_count ++;

#endif



	


	


	if (src->brk_cont_array) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (ii = 0; ii < src->last_brk_cont; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_brk_cont_element,brk_cont_array[i],) */

	assert(sizeof(zend_brk_cont_element) == sizeof(src->brk_cont_array[ii]));
	xc_dprint_indent(indent);fprintf(stderr, "zend_brk_cont_element:brk_cont_array[ii]");
	
	
	/* STRUCT_P_EX(zend_brk_cont_element,dst->brk_cont_array[i],src->brk_cont_array[i],brk_cont_array[i],,&) */

	
	

	
	assert(sizeof(zend_brk_cont_element) == sizeof((& src->brk_cont_array[ii])[0]));
	
	
	xc_dprint_zend_brk_cont_element(
		           & src->brk_cont_array[ii], indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,last_brk_cont) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:last_brk_cont:\t%u\n", src->last_brk_cont);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_brk_cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,current_brk_cont) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:current_brk_cont:\t%u\n", src->current_brk_cont);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->current_brk_cont);
	done_count ++;

#endif



	

#ifndef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_globals) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:uses_globals:\t%u\n", src->uses_globals);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_globals);
	done_count ++;

#endif



	

#endif

#ifdef ZEND_ENGINE_2
	
	if (src->try_catch_array) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->last_try_catch; i ++) {
				
	
	

					
					
	
	/* STRUCT(zend_try_catch_element,try_catch_array[i],) */

	assert(sizeof(zend_try_catch_element) == sizeof(src->try_catch_array[i]));
	xc_dprint_indent(indent);fprintf(stderr, "zend_try_catch_element:try_catch_array[i]");
	
	
	/* STRUCT_P_EX(zend_try_catch_element,dst->try_catch_array[i],src->try_catch_array[i],try_catch_array[i],,&) */

	
	

	
	assert(sizeof(zend_try_catch_element) == sizeof((& src->try_catch_array[i])[0]));
	
	
	xc_dprint_zend_try_catch_element(
		           & src->try_catch_array[i], indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_try_catch) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:last_try_catch:\t%d\n", src->last_try_catch);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_try_catch);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT_P(HashTable,static_variables,HashTable_zval_ptr) */

	if (src->static_variables) {
		xc_dprint_indent(indent);fprintf(stderr, "HashTable:static_variables");
		
	
	/* STRUCT_P_EX(HashTable,dst->static_variables,src->static_variables,static_variables,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_variables)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_HashTable_zval_ptr(
		            src->static_variables, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		xc_dprint_indent(indent);fprintf(stderr, "HashTable:static_variables:\tNULL\n");
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_variables);
	done_count ++;

#endif




	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start_op);
	done_count ++;

#endif


	
	
	/* DISPATCH(int,backpatch_count) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:backpatch_count:\t%d\n", src->backpatch_count);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->backpatch_count);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(zend_bool,done_pass_two) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:done_pass_two:\t%u\n", src->done_pass_two);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->done_pass_two);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_this) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:uses_this:\t%u\n", src->uses_this);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_this);
	done_count ++;

#endif



	

#endif

	
	/* PROC_STRING(filename) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->filename == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->filename");
			xc_dprint_str_len(src->filename, strlen(src->filename) + 1 - 1);
			fprintf(stderr, "\" len=%d\n", strlen(src->filename) + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#ifdef IS_UNICODE
	
		
	/* PROC_STRING(script_encoding) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->script_encoding);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->script_encoding == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->script_encoding");
			xc_dprint_str_len(src->script_encoding, strlen(src->script_encoding) + 1 - 1);
			fprintf(stderr, "\" len=%d\n", strlen(src->script_encoding) + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

	
#endif
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:line_start:\t%u\n", src->line_start);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:line_end:\t%u\n", src->line_end);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:doc_comment_len:\t%d\n", src->doc_comment_len);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->doc_comment), src->doc_comment_len + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->doc_comment");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->doc_comment_len + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->doc_comment");
			xc_dprint_str_len(ZSTR_S(src->doc_comment), src->doc_comment_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->doc_comment_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->doc_comment");
			xc_dprint_str_len(ZSTR_S(src->doc_comment), src->doc_comment_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->doc_comment_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif

	/* reserved */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->reserved);
	done_count ++;

#endif


#if defined(HARDENING_PATCH) && HARDENING_PATCH
	
	
	/* DISPATCH(zend_bool,created_by_eval) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:created_by_eval:\t%u\n", src->created_by_eval);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->created_by_eval);
	done_count ++;

#endif



	

#endif
	} while (0);

#ifdef ZEND_ENGINE_2
		
			
				
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->prototype);
	done_count ++;

#endif



			
	
#endif

	
#ifdef ZEND_ENGINE_2
		
	if (src->scope) {
		
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->scope);
	done_count ++;

#endif


#endif
	

	

		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_zend_op_array : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_zend_op_array : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_dprint_xc_constinfo_t */
	
 	  	 	static void inline 	xc_dprint_xc_constinfo_t( 		const xc_constinfo_t * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_constinfo_t*)NULL)->key) + sizeof(((xc_constinfo_t*)NULL)->key_size) + sizeof(((xc_constinfo_t*)NULL)->constant)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:key_size:\t%u\n", src->key_size);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uchar:type:\t%u\n", src->type);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->key), src->key_size - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->key");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->key_size - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->key");
			xc_dprint_str_len(ZSTR_S(src->key), src->key_size - 1);
			fprintf(stderr, "\" len=%d\n", src->key_size - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->key");
			xc_dprint_str_len(ZSTR_S(src->key), src->key_size - 1);
			fprintf(stderr, "\" len=%d\n", src->key_size - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	/* STRUCT(zend_constant,constant) */

	assert(sizeof(zend_constant) == sizeof(src->constant));
	xc_dprint_indent(indent);fprintf(stderr, "zend_constant:constant");
	
	
	/* STRUCT_P_EX(zend_constant,dst->constant,src->constant,constant,,&) */

	
	

	
	assert(sizeof(zend_constant) == sizeof((& src->constant)[0]));
	
	
	xc_dprint_zend_constant(
		           & src->constant, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constant);
	done_count ++;

#endif




		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_dprint_xc_funcinfo_t */
	
		/* export:  	  	 	void 	xc_dprint_xc_funcinfo_t( 		const xc_funcinfo_t * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dprint_xc_funcinfo_t( 		const xc_funcinfo_t * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_funcinfo_t*)NULL)->key) + sizeof(((xc_funcinfo_t*)NULL)->key_size) + sizeof(((xc_funcinfo_t*)NULL)->func)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:key_size:\t%u\n", src->key_size);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uchar:type:\t%u\n", src->type);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->key), src->key_size - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->key");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->key_size - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->key");
			xc_dprint_str_len(ZSTR_S(src->key), src->key_size - 1);
			fprintf(stderr, "\" len=%d\n", src->key_size - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->key");
			xc_dprint_str_len(ZSTR_S(src->key), src->key_size - 1);
			fprintf(stderr, "\" len=%d\n", src->key_size - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	/* STRUCT(zend_function,func) */

	assert(sizeof(zend_function) == sizeof(src->func));
	xc_dprint_indent(indent);fprintf(stderr, "zend_function:func");
	
	
	/* STRUCT_P_EX(zend_function,dst->func,src->func,func,,&) */

	
	

	
	assert(sizeof(zend_function) == sizeof((& src->func)[0]));
	
	
	xc_dprint_zend_function(
		           & src->func, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->func);
	done_count ++;

#endif




		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dprint_xc_classinfo_t */
	
		/* export:  	  	 	void 	xc_dprint_xc_classinfo_t( 		const xc_classinfo_t * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dprint_xc_classinfo_t( 		const xc_classinfo_t * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_classinfo_t*)NULL)->key) + sizeof(((xc_classinfo_t*)NULL)->key_size) + sizeof(((xc_classinfo_t*)NULL)->cest) + sizeof(((xc_classinfo_t*)NULL)->oplineno)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:key_size:\t%u\n", src->key_size);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uchar:type:\t%u\n", src->type);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->key), src->key_size - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->key");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->key_size - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->key");
			xc_dprint_str_len(ZSTR_S(src->key), src->key_size - 1);
			fprintf(stderr, "\" len=%d\n", src->key_size - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->key");
			xc_dprint_str_len(ZSTR_S(src->key), src->key_size - 1);
			fprintf(stderr, "\" len=%d\n", src->key_size - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT_P(zend_class_entry,cest) */

	if (src->cest) {
		xc_dprint_indent(indent);fprintf(stderr, "zend_class_entry:cest");
		
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof(( src->cest)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_zend_class_entry(
		            src->cest, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		xc_dprint_indent(indent);fprintf(stderr, "zend_class_entry:cest:\tNULL\n");
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#else
	
	
	/* STRUCT(zend_class_entry,cest) */

	assert(sizeof(zend_class_entry) == sizeof(src->cest));
	xc_dprint_indent(indent);fprintf(stderr, "zend_class_entry:cest");
	
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,,&) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof((& src->cest)[0]));
	
	
	xc_dprint_zend_class_entry(
		           & src->cest, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(int,oplineno) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:oplineno:\t%d\n", src->oplineno);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->oplineno);
	done_count ++;

#endif



	


		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2_1

	
	

/* {{{ xc_dprint_xc_autoglobal_t */
	
 	  	 	static void inline 	xc_dprint_xc_autoglobal_t( 		const xc_autoglobal_t * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_autoglobal_t*)NULL)->key) + sizeof(((xc_autoglobal_t*)NULL)->key_len)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			
	
	/* DISPATCH(zend_uint,key_len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:key_len:\t%u\n", src->key_len);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_len);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uchar:type:\t%u\n", src->type);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	/* PROC_ZSTRING_L(type,key,key_len) */
 
	
	/* PROC_ZSTRING_N(type,key,key_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->key), src->key_len + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->key");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->key_len + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->key");
			xc_dprint_str_len(ZSTR_S(src->key), src->key_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->key_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->key");
			xc_dprint_str_len(ZSTR_S(src->key), src->key_len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->key_len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	

		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_dprint_xc_entry_data_php_t */
	
 	  	 	static void inline 	xc_dprint_xc_entry_data_php_t( 		const xc_entry_data_php_t * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_php_t*)NULL)->sourcesize) + sizeof(((xc_entry_data_php_t*)NULL)->device) + sizeof(((xc_entry_data_php_t*)NULL)->inode) + sizeof(((xc_entry_data_php_t*)NULL)->mtime) + sizeof(((xc_entry_data_php_t*)NULL)->op_array) + sizeof(((xc_entry_data_php_t*)NULL)->constinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->constinfos) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfos) + sizeof(((xc_entry_data_php_t*)NULL)->classinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->classinfos) + sizeof(((xc_entry_data_php_t*)NULL)->have_early_binding) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobal_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobals)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			zend_uint i;

#ifdef HAVE_INODE
	
	
	/* DISPATCH(int,device) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:device:\t%d\n", src->device);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->device);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,inode) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:inode:\t%d\n", src->inode);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inode);
	done_count ++;

#endif



	

#endif
	
	
	/* DISPATCH(size_t,sourcesize) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "size_t:sourcesize:\t%u\n", src->sourcesize);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->sourcesize);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(time_t,mtime) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "time_t:mtime:\t%ld\n", src->mtime);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->mtime);
	done_count ++;

#endif



	


	
	
	/* STRUCT_P(zend_op_array,op_array) */

	if (src->op_array) {
		xc_dprint_indent(indent);fprintf(stderr, "zend_op_array:op_array");
		
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof(( src->op_array)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_zend_op_array(
		            src->op_array, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		xc_dprint_indent(indent);fprintf(stderr, "zend_op_array:op_array:\tNULL\n");
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_array);
	done_count ++;

#endif




#ifdef HAVE_XCACHE_CONSTANT
	
	
	/* DISPATCH(zend_uint,constinfo_cnt) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:constinfo_cnt:\t%u\n", src->constinfo_cnt);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->constinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->constinfo_cnt; i ++) {
				
	
	

					
					
	
	/* STRUCT(xc_constinfo_t,constinfos[i],) */

	assert(sizeof(xc_constinfo_t) == sizeof(src->constinfos[i]));
	xc_dprint_indent(indent);fprintf(stderr, "xc_constinfo_t:constinfos[i]");
	
	
	/* STRUCT_P_EX(xc_constinfo_t,dst->constinfos[i],src->constinfos[i],constinfos[i],,&) */

	
	

	
	assert(sizeof(xc_constinfo_t) == sizeof((& src->constinfos[i])[0]));
	
	
	xc_dprint_xc_constinfo_t(
		           & src->constinfos[i], indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif



		
	}

#endif

	
	
	/* DISPATCH(zend_uint,funcinfo_cnt) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:funcinfo_cnt:\t%u\n", src->funcinfo_cnt);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->funcinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->funcinfo_cnt; i ++) {
				
	
	

					
					
	
	/* STRUCT(xc_funcinfo_t,funcinfos[i],) */

	assert(sizeof(xc_funcinfo_t) == sizeof(src->funcinfos[i]));
	xc_dprint_indent(indent);fprintf(stderr, "xc_funcinfo_t:funcinfos[i]");
	
	
	/* STRUCT_P_EX(xc_funcinfo_t,dst->funcinfos[i],src->funcinfos[i],funcinfos[i],,&) */

	
	

	
	assert(sizeof(xc_funcinfo_t) == sizeof((& src->funcinfos[i])[0]));
	
	
	xc_dprint_xc_funcinfo_t(
		           & src->funcinfos[i], indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif



		
	}


	
	
	/* DISPATCH(zend_uint,classinfo_cnt) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:classinfo_cnt:\t%u\n", src->classinfo_cnt);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfo_cnt);
	done_count ++;

#endif



	

	
	
	if (src->classinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->classinfo_cnt; i ++) {
				
	
	

					
		
	
					
	
	/* STRUCT(xc_classinfo_t,classinfos[i],) */

	assert(sizeof(xc_classinfo_t) == sizeof(src->classinfos[i]));
	xc_dprint_indent(indent);fprintf(stderr, "xc_classinfo_t:classinfos[i]");
	
	
	/* STRUCT_P_EX(xc_classinfo_t,dst->classinfos[i],src->classinfos[i],classinfos[i],,&) */

	
	

	
	assert(sizeof(xc_classinfo_t) == sizeof((& src->classinfos[i])[0]));
	
	
	xc_dprint_xc_classinfo_t(
		           & src->classinfos[i], indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif



		
	}

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,autoglobal_cnt) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uint:autoglobal_cnt:\t%u\n", src->autoglobal_cnt);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobal_cnt);
	done_count ++;

#endif



	

	
	if (src->autoglobals) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->autoglobal_cnt; i ++) {
				
	
	

					
		
	
					
	
	/* STRUCT(xc_autoglobal_t,autoglobals[i],) */

	assert(sizeof(xc_autoglobal_t) == sizeof(src->autoglobals[i]));
	xc_dprint_indent(indent);fprintf(stderr, "xc_autoglobal_t:autoglobals[i]");
	
	
	/* STRUCT_P_EX(xc_autoglobal_t,dst->autoglobals[i],src->autoglobals[i],autoglobals[i],,&) */

	
	

	
	assert(sizeof(xc_autoglobal_t) == sizeof((& src->autoglobals[i])[0]));
	
	
	xc_dprint_xc_autoglobal_t(
		           & src->autoglobals[i], indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif



		
	}

#endif
	
	
	/* DISPATCH(zend_bool,have_early_binding) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:have_early_binding:\t%u\n", src->have_early_binding);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_early_binding);
	done_count ++;

#endif



	

	

		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dprint_xc_entry_data_var_t */
	
 	  	 	static void inline 	xc_dprint_xc_entry_data_var_t( 		const xc_entry_data_var_t * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_var_t*)NULL)->value)  ), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
			xc_dprint_indent(indent);fprintf(stderr, "zval:value");
	
	
	/* STRUCT_P_EX(zval_ptr,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof((& src->value)[0]));
	
	
	xc_dprint_zval_ptr(
		           & src->value, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dprint_xc_entry_t */
	
		/* export:  	  	 	void 	xc_dprint_xc_entry_t( 		const xc_entry_t * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dprint_xc_entry_t( 		const xc_entry_t * const src, int indent 		 		 		 		 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_t*)NULL)->type) + sizeof(((xc_entry_t*)NULL)->hvalue) + sizeof(((xc_entry_t*)NULL)->next) + sizeof(((xc_entry_t*)NULL)->cache) + sizeof(((xc_entry_t*)NULL)->size) + sizeof(((xc_entry_t*)NULL)->refcount) + sizeof(((xc_entry_t*)NULL)->hits) + sizeof(((xc_entry_t*)NULL)->ctime) + sizeof(((xc_entry_t*)NULL)->atime) + sizeof(((xc_entry_t*)NULL)->dtime) + sizeof(((xc_entry_t*)NULL)->ttl) + sizeof(((xc_entry_t*)NULL)->name) + sizeof(((xc_entry_t*)NULL)->data) + sizeof(((xc_entry_t*)NULL)->have_references)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			fprintf(stderr, "%s", " {\n");
			indent ++;
		
		
	
	
	
	/* DISPATCH(xc_entry_type_t,type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "xc_entry_type_t:type:\t%d\n", src->type);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(size_t,size) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "size_t:size:\t%u\n", src->size);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(xc_hash_value_t,hvalue) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "xc_hash_value_t:hvalue:\t%lu\n", src->hvalue);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hvalue);
	done_count ++;

#endif



	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cache);
	done_count ++;

#endif


	/* skip */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->next);
	done_count ++;

#endif



	
	
	/* DISPATCH(long,refcount) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "long:refcount:\t%ld\n", src->refcount);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(time_t,ctime) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "time_t:ctime:\t%ld\n", src->ctime);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ctime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,atime) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "time_t:atime:\t%ld\n", src->atime);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->atime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,dtime) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "time_t:dtime:\t%ld\n", src->dtime);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->dtime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(long,ttl) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "long:ttl:\t%ld\n", src->ttl);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ttl);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_ulong,hits) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_ulong:hits:\t%lu\n", src->hits);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hits);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,name_type) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_uchar:name_type:\t%u\n", src->name_type);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_type);
	done_count ++;

#endif



	

#endif
		
	
	

#ifdef IS_UNICODE
		if (src->name_type == IS_UNICODE) {
			
	
	/* DISPATCH(int32_t,name.ustr.len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int32_t:name.ustr.len:\t%d\n", src->name.ustr.len);
	
	
	

	

		}
		else {
			
	
	/* DISPATCH(int,name.str.len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:name.str.len:\t%d\n", src->name.str.len);
	
	
	

	

		}
#else
		
	
	/* DISPATCH(int,name.str.len) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "int:name.str.len:\t%d\n", src->name.str.len);
	
	
	

	

#endif
		
#ifdef IS_UNICODE
			
	/* PROC_ZSTRING_L(name_type,name.uni.val,name.uni.len) */
 
	
	/* PROC_ZSTRING_N(name_type,name.uni.val,name.uni.len,len) */

#ifdef IS_UNICODE
	
	
	
		if (src->name_type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name.uni.val) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
#ifdef IS_UNICODE
			do {
				zval zv;
				zval reszv;
				int usecopy;

				INIT_ZVAL(zv);
				ZVAL_UNICODEL(&zv, (UChar *) (src->name.uni.val), src->name.uni.len + 1 - 1, 1);
				zend_make_printable_zval(&zv, &reszv, &usecopy);
				fprintf(stderr, "string:%s:\t\"", "dst->name.uni.val");
				xc_dprint_str_len(Z_STRVAL(reszv), Z_STRLEN(reszv));
				fprintf(stderr, "\" len=%d\n", src->name.uni.len + 1 - 1);
				if (usecopy) {
					zval_dtor(&reszv);
				}
				zval_dtor(&zv);
			} while (0);
#endif
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name.uni.val) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name.uni.val");
			xc_dprint_str_len(ZSTR_S(src->name.uni.val), src->name.uni.len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name.uni.len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name.uni.val) == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name.uni.val");
			xc_dprint_str_len(ZSTR_S(src->name.uni.val), src->name.uni.len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name.uni.len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#else
			
	/* PROC_STRING_L(name.str.val,name.str.len) */
 
	/* PROC_STRING_N(name.str.val,name.str.len + 1) */
 
	
	
	
	
	
	
	
	if (src->name.str.val == NULL) {
		
		
	}
	else {
		xc_dprint_indent(indent);
			
			fprintf(stderr, "string:%s:\t\"", "dst->name.str.val");
			xc_dprint_str_len(src->name.str.val, src->name.str.len + 1 - 1);
			fprintf(stderr, "\" len=%d\n", src->name.str.len + 1 - 1);
			
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
		
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		
	
	

		switch (src->type) {
		case XC_TYPE_PHP:
			
	
	/* STRUCT_P(xc_entry_data_php_t,data.php) */

	if (src->data.php) {
		xc_dprint_indent(indent);fprintf(stderr, "xc_entry_data_php_t:data.php");
		
	
	/* STRUCT_P_EX(xc_entry_data_php_t,dst->data.php,src->data.php,data.php,) */

	
	

	
	assert(sizeof(xc_entry_data_php_t) == sizeof(( src->data.php)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_xc_entry_data_php_t(
		            src->data.php, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		xc_dprint_indent(indent);fprintf(stderr, "xc_entry_data_php_t:data.php:\tNULL\n");
		
	
	

	}
	

			break;
		case XC_TYPE_VAR:
			
	
	/* STRUCT_P(xc_entry_data_var_t,data.var) */

	if (src->data.var) {
		xc_dprint_indent(indent);fprintf(stderr, "xc_entry_data_var_t:data.var");
		
	
	/* STRUCT_P_EX(xc_entry_data_var_t,dst->data.var,src->data.var,data.var,) */

	
	

	
	assert(sizeof(xc_entry_data_var_t) == sizeof(( src->data.var)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_dprint_xc_entry_data_var_t(
		            src->data.var, indent
		
		
		
		
		
		TSRMLS_CC
	);
	
	
	

	}
	else {
		xc_dprint_indent(indent);fprintf(stderr, "xc_entry_data_var_t:data.var:\tNULL\n");
		
	
	

	}
	

			break;
		default:
			assert(0);
		}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->data);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,have_references) */

	
	
	
		xc_dprint_indent(indent);
		fprintf(stderr, "zend_bool:have_references:\t%u\n", src->have_references);
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_references);
	done_count ++;

#endif



	


		
			indent --;
			xc_dprint_indent(indent);fprintf(stderr, "}\n");
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dprint_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dprint_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif /* HAVE_XCACHE_DPRINT */
#ifdef HAVE_XCACHE_DISASSEMBLER
  /* {{{ Pre-declare */
 	  	 	void 	xc_dasm_zval( 		 		 		 		 		zval *dst, const zval * const src 		 		TSRMLS_DC 	); 	
 	  	 	static void inline 	xc_dasm_zval_ptr( 		 		 		 		 		zval *dst, const zval_ptr * const src 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_dasm_zend_op_array( 		 		 		 		 		zval *dst, const zend_op_array * const src 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_dasm_zend_class_entry( 		 		 		 		 		zval *dst, const zend_class_entry * const src 		 		TSRMLS_DC 	); 	
#ifdef HAVE_XCACHE_CONSTANT
 	  	 	static void inline 	xc_dasm_zend_constant( 		 		 		 		 		zval *dst, const zend_constant * const src 		 		TSRMLS_DC 	); 	
#endif
 	  	 	void 	xc_dasm_zend_function( 		 		 		 		 		zval *dst, const zend_function * const src 		 		TSRMLS_DC 	); 	
 	  	 	void 	xc_dasm_xc_entry_t( 		 		 		 		 		zval *dst, const xc_entry_t * const src 		 		TSRMLS_DC 	); 	
#ifdef ZEND_ENGINE_2
 	  	 	static void inline 	xc_dasm_zend_property_info( 		 		 		 		 		zval *dst, const zend_property_info * const src 		 		TSRMLS_DC 	); 	
#endif
/* }}} */
#ifdef IS_CV

	
	

/* {{{ xc_dasm_zend_compiled_variable */
	
 	  	 	static void inline 	xc_dasm_zend_compiled_variable( 		 		 		 		 		zval *dst, const zend_compiled_variable * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_compiled_variable*)NULL)->name) + sizeof(((zend_compiled_variable*)NULL)->name_len) + sizeof(((zend_compiled_variable*)NULL)->hash_value)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(int,name_len) */

	
	
	
	
		add_assoc_long_ex(dst, ("name_len"), (sizeof("name_len")), src->name_len);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("name"), (sizeof("name")), src->name, src->name_len + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name"), (sizeof("name")), src->name, src->name_len + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name"), (sizeof("name")), src->name, src->name_len + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,hash_value) */

	
	
	
	
		add_assoc_long_ex(dst, ("hash_value"), (sizeof("hash_value")), src->hash_value);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hash_value);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_dasm_zend_uint */
	
 	  	 	static void inline 	xc_dasm_zend_uint( 		 		 		 		 		zval *dst, const zend_uint * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_uint), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifndef ZEND_ENGINE_2

	
	

/* {{{ xc_dasm_int */
	
 	  	 	static void inline 	xc_dasm_int( 		 		 		 		 		zval *dst, const int * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(int), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_dasm_zend_try_catch_element */
	
 	  	 	static void inline 	xc_dasm_zend_try_catch_element( 		 		 		 		 		zval *dst, const zend_try_catch_element * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_try_catch_element*)NULL)->try_op) + sizeof(((zend_try_catch_element*)NULL)->catch_op)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(zend_uint,try_op) */

	
	
	
	
		add_assoc_long_ex(dst, ("try_op"), (sizeof("try_op")), src->try_op);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_op);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,catch_op) */

	
	
	
	
		add_assoc_long_ex(dst, ("catch_op"), (sizeof("catch_op")), src->catch_op);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->catch_op);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif /* ifdef ZEND_ENGINE_2 */

	
	

/* {{{ xc_dasm_zend_brk_cont_element */
	
 	  	 	static void inline 	xc_dasm_zend_brk_cont_element( 		 		 		 		 		zval *dst, const zend_brk_cont_element * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_brk_cont_element*)NULL)->start) + sizeof(((zend_brk_cont_element*)NULL)->cont) + sizeof(((zend_brk_cont_element*)NULL)->brk) + sizeof(((zend_brk_cont_element*)NULL)->parent)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
#ifdef ZEND_ENGINE_2_2
#ifndef IS_UNICODE
	
	
	/* DISPATCH(int,start) */

	
	
	
	
		add_assoc_long_ex(dst, ("start"), (sizeof("start")), src->start);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start);
	done_count ++;

#endif



	

#endif
#endif
	
	
	/* DISPATCH(int,cont) */

	
	
	
	
		add_assoc_long_ex(dst, ("cont"), (sizeof("cont")), src->cont);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,brk) */

	
	
	
	
		add_assoc_long_ex(dst, ("brk"), (sizeof("brk")), src->brk);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,parent) */

	
	
	
	
		add_assoc_long_ex(dst, ("parent"), (sizeof("parent")), src->parent);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	
	

/* {{{ xc_dasm_HashTable_zval_ptr */
	
 	  	 	static void inline 	xc_dasm_HashTable_zval_ptr( 		 		 		 		 		zval *dst, const HashTable * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
		
	

				
			Bucket *b;
			zval *zv;
			int bufsize = 2;
			char *buf = emalloc(bufsize);
			int keysize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


#endif
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif


#if ZEND_DEBUG
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif


#endif
#ifdef IS_UNICODE
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif


#endif

			
	
	

			for (b = src->pListHead; b != NULL; b = b->pListNext) {
				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_zval_ptr (zv, ((zval_ptr*)b->pData) TSRMLS_CC);
				keysize = BUCKET_KEY_SIZE(b) + 2;
				if (keysize > bufsize) {
					do {
						bufsize *= 2;
					} while (keysize > bufsize);
					buf = erealloc(buf, bufsize);
				}
				memcpy(buf, BUCKET_KEY_S(b), keysize);
				buf[keysize - 2] = buf[keysize - 1] = ""[0];
				keysize = b->nKeyLength;
#ifdef IS_UNICODE
				if (BUCKET_KEY_TYPE(b) == IS_UNICODE) {
					if (buf[0] == ""[0] && buf[1] == ""[0]) {
						keysize ++;
					}
				} else
#endif
				{
					if (buf[0] == ""[0]) {
						keysize ++;
					}
				}
				add_u_assoc_zval_ex(dst, BUCKET_KEY_TYPE(b), buf, keysize, zv);
			}
			
	
	


			efree(buf);
			return; /* no check size */
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	
	

/* {{{ xc_dasm_HashTable_zend_constant */
	
 	  	 	static void inline 	xc_dasm_HashTable_zend_constant( 		 		 		 		 		zval *dst, const HashTable * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
		
	

				
			Bucket *b;
			zval *zv;
			int bufsize = 2;
			char *buf = emalloc(bufsize);
			int keysize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


#endif
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif


#if ZEND_DEBUG
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif


#endif
#ifdef IS_UNICODE
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif


#endif

			
	
	

			for (b = src->pListHead; b != NULL; b = b->pListNext) {
				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_zend_constant (zv, ((zend_constant*)b->pData) TSRMLS_CC);
				keysize = BUCKET_KEY_SIZE(b) + 2;
				if (keysize > bufsize) {
					do {
						bufsize *= 2;
					} while (keysize > bufsize);
					buf = erealloc(buf, bufsize);
				}
				memcpy(buf, BUCKET_KEY_S(b), keysize);
				buf[keysize - 2] = buf[keysize - 1] = ""[0];
				keysize = b->nKeyLength;
#ifdef IS_UNICODE
				if (BUCKET_KEY_TYPE(b) == IS_UNICODE) {
					if (buf[0] == ""[0] && buf[1] == ""[0]) {
						keysize ++;
					}
				} else
#endif
				{
					if (buf[0] == ""[0]) {
						keysize ++;
					}
				}
				add_u_assoc_zval_ex(dst, BUCKET_KEY_TYPE(b), buf, keysize, zv);
			}
			
	
	


			efree(buf);
			return; /* no check size */
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	
	

/* {{{ xc_dasm_HashTable_zend_function */
	
		/* export:  	  	 	void 	xc_dasm_HashTable_zend_function( 		 		 		 		 		zval *dst, const HashTable * const src 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dasm_HashTable_zend_function( 		 		 		 		 		zval *dst, const HashTable * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
		
	

				
			Bucket *b;
			zval *zv;
			int bufsize = 2;
			char *buf = emalloc(bufsize);
			int keysize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


#endif
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif


#if ZEND_DEBUG
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif


#endif
#ifdef IS_UNICODE
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif


#endif

			
	
	

			for (b = src->pListHead; b != NULL; b = b->pListNext) {
				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_zend_function (zv, ((zend_function*)b->pData) TSRMLS_CC);
				keysize = BUCKET_KEY_SIZE(b) + 2;
				if (keysize > bufsize) {
					do {
						bufsize *= 2;
					} while (keysize > bufsize);
					buf = erealloc(buf, bufsize);
				}
				memcpy(buf, BUCKET_KEY_S(b), keysize);
				buf[keysize - 2] = buf[keysize - 1] = ""[0];
				keysize = b->nKeyLength;
#ifdef IS_UNICODE
				if (BUCKET_KEY_TYPE(b) == IS_UNICODE) {
					if (buf[0] == ""[0] && buf[1] == ""[0]) {
						keysize ++;
					}
				} else
#endif
				{
					if (buf[0] == ""[0]) {
						keysize ++;
					}
				}
				add_u_assoc_zval_ex(dst, BUCKET_KEY_TYPE(b), buf, keysize, zv);
			}
			
	
	


			efree(buf);
			return; /* no check size */
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef ZEND_ENGINE_2

	
	
	

/* {{{ xc_dasm_HashTable_zend_property_info */
	
 	  	 	static void inline 	xc_dasm_HashTable_zend_property_info( 		 		 		 		 		zval *dst, const HashTable * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
		
	

				
			Bucket *b;
			zval *zv;
			int bufsize = 2;
			char *buf = emalloc(bufsize);
			int keysize;

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


#endif
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif


			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif


#if ZEND_DEBUG
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif


#endif
#ifdef IS_UNICODE
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif


#endif

			
	
	

			for (b = src->pListHead; b != NULL; b = b->pListNext) {
				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_zend_property_info (zv, ((zend_property_info*)b->pData) TSRMLS_CC);
				keysize = BUCKET_KEY_SIZE(b) + 2;
				if (keysize > bufsize) {
					do {
						bufsize *= 2;
					} while (keysize > bufsize);
					buf = erealloc(buf, bufsize);
				}
				memcpy(buf, BUCKET_KEY_S(b), keysize);
				buf[keysize - 2] = buf[keysize - 1] = ""[0];
				keysize = b->nKeyLength;
#ifdef IS_UNICODE
				if (BUCKET_KEY_TYPE(b) == IS_UNICODE) {
					if (buf[0] == ""[0] && buf[1] == ""[0]) {
						keysize ++;
					}
				} else
#endif
				{
					if (buf[0] == ""[0]) {
						keysize ++;
					}
				}
				add_u_assoc_zval_ex(dst, BUCKET_KEY_TYPE(b), buf, keysize, zv);
			}
			
	
	


			efree(buf);
			return; /* no check size */
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	

/* {{{ xc_dasm_zval */
	
		/* export:  	  	 	void 	xc_dasm_zval( 		 		 		 		 		zval *dst, const zval * const src 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dasm_zval( 		 		 		 		 		zval *dst, const zval * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zval*)NULL)->value) + sizeof(((zval*)NULL)->refcount) + sizeof(((zval*)NULL)->type) + sizeof(((zval*)NULL)->is_ref)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			do {
		zval_dtor(dst);
		*dst = *src;
		zval_copy_ctor(dst);
		ZVAL_REFCOUNT(dst) = 1;
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif


		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif


		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif


		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->is_ref);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2_3
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->idx_type);
	done_count ++;

#endif


#endif
	} while(0);
	return;
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dasm_zval_ptr */
	
 	  	 	static void inline 	xc_dasm_zval_ptr( 		 		 		 		 		zval *dst, const zval_ptr * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zval_ptr), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
		
	

		xc_dasm_zval (dst, src[0] TSRMLS_CC);
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(zval_ptr);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_dasm_zend_arg_info */
	
 	  	 	static void inline 	xc_dasm_zend_arg_info( 		 		 		 		 		zval *dst, const zend_arg_info * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_arg_info*)NULL)->name) + sizeof(((zend_arg_info*)NULL)->name_len) + sizeof(((zend_arg_info*)NULL)->class_name) + sizeof(((zend_arg_info*)NULL)->class_name_len) + sizeof(((zend_arg_info*)NULL)->array_type_hint) + sizeof(((zend_arg_info*)NULL)->allow_null) + sizeof(((zend_arg_info*)NULL)->pass_by_reference) + sizeof(((zend_arg_info*)NULL)->return_reference) + sizeof(((zend_arg_info*)NULL)->required_num_args)  ), assert_count = 9;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(zend_uint,name_len) */

	
	
	
	
		add_assoc_long_ex(dst, ("name_len"), (sizeof("name_len")), src->name_len);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("name"), (sizeof("name")), src->name, src->name_len + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name"), (sizeof("name")), src->name, src->name_len + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name"), (sizeof("name")), src->name, src->name_len + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_uint,class_name_len) */

	
	
	
	
		add_assoc_long_ex(dst, ("class_name_len"), (sizeof("class_name_len")), src->class_name_len);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,class_name,class_name_len) */
 
	
	/* PROC_ZSTRING_N(,class_name,class_name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->class_name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("class_name"), (sizeof("class_name")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("class_name"), (sizeof("class_name")), src->class_name, src->class_name_len + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("class_name"), (sizeof("class_name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("class_name"), (sizeof("class_name")), src->class_name, src->class_name_len + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("class_name"), (sizeof("class_name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("class_name"), (sizeof("class_name")), src->class_name, src->class_name_len + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_bool,array_type_hint) */

	
	
	
	
		add_assoc_bool_ex(dst, ("array_type_hint"), (sizeof("array_type_hint")), src->array_type_hint ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->array_type_hint);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,allow_null) */

	
	
	
	
		add_assoc_bool_ex(dst, ("allow_null"), (sizeof("allow_null")), src->allow_null ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->allow_null);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_by_reference) */

	
	
	
	
		add_assoc_bool_ex(dst, ("pass_by_reference"), (sizeof("pass_by_reference")), src->pass_by_reference ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_by_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,return_reference) */

	
	
	
	
		add_assoc_bool_ex(dst, ("return_reference"), (sizeof("return_reference")), src->return_reference ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,required_num_args) */

	
	
	
	
		add_assoc_long_ex(dst, ("required_num_args"), (sizeof("required_num_args")), src->required_num_args);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_dasm_zend_constant */
	
 	  	 	static void inline 	xc_dasm_zend_constant( 		 		 		 		 		zval *dst, const zend_constant * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_constant*)NULL)->value) + sizeof(((zend_constant*)NULL)->flags) + sizeof(((zend_constant*)NULL)->name) + sizeof(((zend_constant*)NULL)->name_len) + sizeof(((zend_constant*)NULL)->module_number)  ), assert_count = 5;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* STRUCT(zval,value) */

	assert(sizeof(zval) == sizeof(src->value));
	
	
	
	/* STRUCT_P_EX(zval,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->value)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_zval(
		
		
		
		
		zv, & src->value
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("value"), (sizeof("value")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



	
	
	/* DISPATCH(int,flags) */

	
	
	
	
		add_assoc_long_ex(dst, ("flags"), (sizeof("flags")), src->flags);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,name_len) */

	
	
	
	
		add_assoc_long_ex(dst, ("name_len"), (sizeof("name_len")), src->name_len);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	
	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("name"), (sizeof("name")), src->name, src->name_len + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name"), (sizeof("name")), src->name, src->name_len + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name"), (sizeof("name")), src->name, src->name_len + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

	
	
	
	
	/* DISPATCH(int,module_number) */

	
	
	
	
		add_assoc_long_ex(dst, ("module_number"), (sizeof("module_number")), src->module_number);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module_number);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_dasm_zend_function */
	
		/* export:  	  	 	void 	xc_dasm_zend_function( 		 		 		 		 		zval *dst, const zend_function * const src 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dasm_zend_function( 		 		 		 		 		zval *dst, const zend_function * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_function), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	

	switch (src->type) {
	case ZEND_INTERNAL_FUNCTION:
	case ZEND_OVERLOADED_FUNCTION:
		
		break;

	case ZEND_USER_FUNCTION:
	case ZEND_EVAL_CODE:
		
		
	
	/* STRUCT(zend_op_array,op_array) */

	assert(sizeof(zend_op_array) == sizeof(src->op_array));
	
	
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,,&) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof((& src->op_array)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_zend_op_array(
		
		
		
		
		zv, & src->op_array
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("op_array"), (sizeof("op_array")), zv);
	} while (0);
	
	
	

	

		break;

	default:
		assert(0);
	}
	
	
	

	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_dasm_zend_property_info */
	
 	  	 	static void inline 	xc_dasm_zend_property_info( 		 		 		 		 		zval *dst, const zend_property_info * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_property_info*)NULL)->flags) + sizeof(((zend_property_info*)NULL)->name) + sizeof(((zend_property_info*)NULL)->name_length) + sizeof(((zend_property_info*)NULL)->h) + sizeof(((zend_property_info*)NULL)->doc_comment) + sizeof(((zend_property_info*)NULL)->doc_comment_len) + sizeof(((zend_property_info*)NULL)->ce)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(zend_uint,flags) */

	
	
	
	
		add_assoc_long_ex(dst, ("flags"), (sizeof("flags")), src->flags);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,name_length) */

	
	
	
	
		add_assoc_long_ex(dst, ("name_length"), (sizeof("name_length")), src->name_length);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("name"), (sizeof("name")), src->name, src->name_length + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name"), (sizeof("name")), src->name, src->name_length + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name"), (sizeof("name")), src->name, src->name_length + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,h) */

	
	
	
	
		add_assoc_long_ex(dst, ("h"), (sizeof("h")), src->h);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->h);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
	
		add_assoc_long_ex(dst, ("doc_comment_len"), (sizeof("doc_comment_len")), src->doc_comment_len);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
			add_assoc_null_ex(dst, ("doc_comment"), (sizeof("doc_comment")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("doc_comment"), (sizeof("doc_comment")), src->doc_comment, src->doc_comment_len + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
			add_assoc_null_ex(dst, ("doc_comment"), (sizeof("doc_comment")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("doc_comment"), (sizeof("doc_comment")), src->doc_comment, src->doc_comment_len + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
			add_assoc_null_ex(dst, ("doc_comment"), (sizeof("doc_comment")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("doc_comment"), (sizeof("doc_comment")), src->doc_comment, src->doc_comment_len + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

#endif
	#if defined(ZEND_ENGINE_2_2) && !defined(IS_UNICODE)
	
	if (src->ce) {
		
		
		add_assoc_stringl_ex(dst, ("ce"), (sizeof("ce")), src->ce->name, strlen(src->ce->name), 1);
	}
	else {
		
	add_assoc_null_ex(dst, ("ce"), (sizeof("ce")));
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce);
	done_count ++;

#endif


#endif

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_dasm_zend_class_entry */
	
		/* export:  	  	 	void 	xc_dasm_zend_class_entry( 		 		 		 		 		zval *dst, const zend_class_entry * const src 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dasm_zend_class_entry( 		 		 		 		 		zval *dst, const zend_class_entry * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_class_entry*)NULL)->type) + sizeof(((zend_class_entry*)NULL)->name) + sizeof(((zend_class_entry*)NULL)->name_length) + sizeof(((zend_class_entry*)NULL)->parent) + sizeof(((zend_class_entry*)NULL)->refcount) + sizeof(((zend_class_entry*)NULL)->constants_updated) + sizeof(((zend_class_entry*)NULL)->ce_flags) + sizeof(((zend_class_entry*)NULL)->function_table) + sizeof(((zend_class_entry*)NULL)->default_properties) + sizeof(((zend_class_entry*)NULL)->properties_info) + sizeof(((zend_class_entry*)NULL)->default_static_members) + sizeof(((zend_class_entry*)NULL)->static_members) + sizeof(((zend_class_entry*)NULL)->constants_table) + sizeof(((zend_class_entry*)NULL)->builtin_functions) + sizeof(((zend_class_entry*)NULL)->constructor) + sizeof(((zend_class_entry*)NULL)->destructor) + sizeof(((zend_class_entry*)NULL)->clone) + sizeof(((zend_class_entry*)NULL)->__get) + sizeof(((zend_class_entry*)NULL)->__set) + sizeof(((zend_class_entry*)NULL)->__unset) + sizeof(((zend_class_entry*)NULL)->__isset) + sizeof(((zend_class_entry*)NULL)->__call) + sizeof(((zend_class_entry*)NULL)->__tostring) + sizeof(((zend_class_entry*)NULL)->serialize_func) + sizeof(((zend_class_entry*)NULL)->unserialize_func) + sizeof(((zend_class_entry*)NULL)->iterator_funcs) + sizeof(((zend_class_entry*)NULL)->create_object) + sizeof(((zend_class_entry*)NULL)->get_iterator) + sizeof(((zend_class_entry*)NULL)->interface_gets_implemented) + sizeof(((zend_class_entry*)NULL)->serialize) + sizeof(((zend_class_entry*)NULL)->unserialize) + sizeof(((zend_class_entry*)NULL)->interfaces) + sizeof(((zend_class_entry*)NULL)->num_interfaces) + sizeof(((zend_class_entry*)NULL)->filename) + sizeof(((zend_class_entry*)NULL)->line_start) + sizeof(((zend_class_entry*)NULL)->line_end) + sizeof(((zend_class_entry*)NULL)->doc_comment) + sizeof(((zend_class_entry*)NULL)->doc_comment_len) + sizeof(((zend_class_entry*)NULL)->module)  ), assert_count = 39;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	
	/* DISPATCH(char,type) */

	
	
	
	
		add_assoc_long_ex(dst, ("type"), (sizeof("type")), src->type);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,name_length) */

	
	
	
	
		add_assoc_long_ex(dst, ("name_length"), (sizeof("name_length")), src->name_length);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("name"), (sizeof("name")), src->name, src->name_length + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name"), (sizeof("name")), src->name, src->name_length + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name"), (sizeof("name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name"), (sizeof("name")), src->name, src->name_length + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

	
		
	if (src->parent) {
		
		
		add_assoc_stringl_ex(dst, ("parent"), (sizeof("parent")), src->parent->name, strlen(src->parent->name), 1);
	}
	else {
		
	add_assoc_null_ex(dst, ("parent"), (sizeof("parent")));
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif


	
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(int,refcount) */

	
	
	
	
		add_assoc_long_ex(dst, ("refcount"), (sizeof("refcount")), src->refcount);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	

#else
	
	
	/* STRUCT_P(int,refcount) */

	if (src->refcount) {
		
		
	
	/* STRUCT_P_EX(int,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(int) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_int(
		
		
		
		
		zv,  src->refcount
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("refcount"), (sizeof("refcount")), zv);
	} while (0);
	
	
	

	}
	else {
		
		
	add_assoc_null_ex(dst, ("refcount"), (sizeof("refcount")));
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(zend_bool,constants_updated) */

	
	
	
	
		add_assoc_bool_ex(dst, ("constants_updated"), (sizeof("constants_updated")), src->constants_updated ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_updated);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,ce_flags) */

	
	
	
	
		add_assoc_long_ex(dst, ("ce_flags"), (sizeof("ce_flags")), src->ce_flags);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce_flags);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT(HashTable,default_properties,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_properties));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->default_properties,src->default_properties,default_properties,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_properties)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_HashTable_zval_ptr(
		
		
		
		
		zv, & src->default_properties
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("default_properties"), (sizeof("default_properties")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_properties);
	done_count ++;

#endif



	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->builtin_functions);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT(HashTable,properties_info,HashTable_zend_property_info) */

	assert(sizeof(HashTable) == sizeof(src->properties_info));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->properties_info,src->properties_info,properties_info,HashTable_zend_property_info,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->properties_info)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_HashTable_zend_property_info(
		
		
		
		
		zv, & src->properties_info
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("properties_info"), (sizeof("properties_info")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->properties_info);
	done_count ++;

#endif



#	ifdef ZEND_ENGINE_2_1
	
	
	/* STRUCT(HashTable,default_static_members,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_static_members));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->default_static_members,src->default_static_members,default_static_members,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_static_members)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_HashTable_zval_ptr(
		
		
		
		
		zv, & src->default_static_members
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("default_static_members"), (sizeof("default_static_members")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_static_members);
	done_count ++;

#endif



	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif


#	else
	
	
	/* STRUCT_P(HashTable,static_members,HashTable_zval_ptr) */

	if (src->static_members) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->static_members,src->static_members,static_members,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_members)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_HashTable_zval_ptr(
		
		
		
		
		zv,  src->static_members
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("static_members"), (sizeof("static_members")), zv);
	} while (0);
	
	
	

	}
	else {
		
		
	add_assoc_null_ex(dst, ("static_members"), (sizeof("static_members")));
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif



#	endif
	
	
	/* STRUCT(HashTable,constants_table,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->constants_table));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->constants_table,src->constants_table,constants_table,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->constants_table)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_HashTable_zval_ptr(
		
		
		
		
		zv, & src->constants_table
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("constants_table"), (sizeof("constants_table")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_table);
	done_count ++;

#endif




		
	
		if (src->num_interfaces) {
			/*
			zval *arr;
			ALLOC_INIT_ZVAL(arr);
			array_init(arr);
			for (i = 0; i < src->num_interfaces; i ++) {
				zval *zv;
				ALLOC_INIT_ZVAL(zv);
				ZVAL_STRING(src->num_interfaces);
			}
			add_assoc_zval_ex(dst, ("interfaces"), (sizeof("interfaces")), arr);
			*/
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interfaces);
	done_count ++;

#endif


		}
		else {
			
	
	add_assoc_null_ex(dst, (""), (sizeof("")));
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interfaces);
	done_count ++;

#endif



		}
	
	
		
	
	
	
	/* DISPATCH(zend_uint,num_interfaces) */

	
	
	
	
		add_assoc_long_ex(dst, ("num_interfaces"), (sizeof("num_interfaces")), src->num_interfaces);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_interfaces);
	done_count ++;

#endif



	


	
	/* PROC_STRING(filename) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->filename == NULL) {
		
		
			add_assoc_null_ex(dst, ("filename"), (sizeof("filename")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("filename"), (sizeof("filename")), src->filename, strlen(src->filename) + 1-1, 1);
				
	}
	
	
	
	
	
	

	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
	
		add_assoc_long_ex(dst, ("line_start"), (sizeof("line_start")), src->line_start);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
	
		add_assoc_long_ex(dst, ("line_end"), (sizeof("line_end")), src->line_end);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,doc_comment_len) */

	
	
	
	
		add_assoc_long_ex(dst, ("doc_comment_len"), (sizeof("doc_comment_len")), src->doc_comment_len);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
			add_assoc_null_ex(dst, ("doc_comment"), (sizeof("doc_comment")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("doc_comment"), (sizeof("doc_comment")), src->doc_comment, src->doc_comment_len + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
			add_assoc_null_ex(dst, ("doc_comment"), (sizeof("doc_comment")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("doc_comment"), (sizeof("doc_comment")), src->doc_comment, src->doc_comment_len + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
			add_assoc_null_ex(dst, ("doc_comment"), (sizeof("doc_comment")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("doc_comment"), (sizeof("doc_comment")), src->doc_comment, src->doc_comment_len + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->iterator_funcs);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->create_object);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_iterator);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interface_gets_implemented);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2_3
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_static_method);
	done_count ++;

#endif


#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize);
	done_count ++;

#endif


	/* deal with it inside xc_fix_method */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->destructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->clone);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__set);
	done_count ++;

#endif


/* should be >5.1 */
#ifdef ZEND_ENGINE_2_1
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__unset);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__isset);
	done_count ++;

#endif


# if defined(ZEND_ENGINE_2_2) || PHP_MAJOR_VERSION >= 6
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__tostring);
	done_count ++;

#endif


# endif
#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__call);
	done_count ++;

#endif


#ifdef ZEND_CALLSTATIC_FUNC_NAME
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__callstatic);
	done_count ++;

#endif


#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module);
	done_count ++;

#endif


#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_function_call);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_set);
	done_count ++;

#endif


#endif
		
	
	/* STRUCT(HashTable,function_table,HashTable_zend_function) */

	assert(sizeof(HashTable) == sizeof(src->function_table));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->function_table,src->function_table,function_table,HashTable_zend_function,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->function_table)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_HashTable_zend_function(
		
		
		
		
		zv, & src->function_table
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("function_table"), (sizeof("function_table")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_table);
	done_count ++;

#endif



	
	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dasm_znode */
	
 	  	 	static void inline 	xc_dasm_znode( 		 		 		 		 		zval *dst, const znode * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((znode*)NULL)->op_type) + sizeof(((znode*)NULL)->u)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(int,op_type) */

	
	
	
	
		add_assoc_long_ex(dst, ("op_type"), (sizeof("op_type")), src->op_type);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_type);
	done_count ++;

#endif



	


#ifdef IS_CV
#	define XCACHE_IS_CV IS_CV
#else
/* compatible with zend optimizer */
#	define XCACHE_IS_CV 16
#endif
	assert(src->op_type == IS_CONST ||
		src->op_type == IS_VAR ||
		src->op_type == XCACHE_IS_CV ||
		src->op_type == IS_TMP_VAR ||
		src->op_type == IS_UNUSED);
#undef XCACHE_IS_CV
		
	
	

	switch (src->op_type) {
		case IS_CONST:
			
	
	/* STRUCT(zval,u.constant) */

	assert(sizeof(zval) == sizeof(src->u.constant));
	
	
	
	/* STRUCT_P_EX(zval,dst->u.constant,src->u.constant,u.constant,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->u.constant)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_zval(
		
		
		
		
		zv, & src->u.constant
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("u.constant"), (sizeof("u.constant")), zv);
	} while (0);
	
	
	

	

			break;
		
		case IS_VAR:
		case IS_TMP_VAR:
#ifdef IS_CV
		case IS_CV:
#else
		case 16:
#endif
			
	
	/* DISPATCH(zend_uint,u.var) */

	
	
	
	
		add_assoc_long_ex(dst, ("u.var"), (sizeof("u.var")), src->u.var);
		
	
	

	

			
	
	/* DISPATCH(zend_uint,u.EA.type) */

	
	
	
	
		add_assoc_long_ex(dst, ("u.EA.type"), (sizeof("u.EA.type")), src->u.EA.type);
		
	
	

	

			break;
		case IS_UNUSED:
			
	
	/* DISPATCH(zend_uint,u.var) */

	
	
	
	
		add_assoc_long_ex(dst, ("u.var"), (sizeof("u.var")), src->u.var);
		
	
	

	

			
	
	/* DISPATCH(zend_uint,u.opline_num) */

	
	
	
	
		add_assoc_long_ex(dst, ("u.opline_num"), (sizeof("u.opline_num")), src->u.opline_num);
		
	
	

	

#ifndef ZEND_ENGINE_2
			
	
	/* DISPATCH(zend_uint,u.fetch_type) */

	
	
	
	
		add_assoc_long_ex(dst, ("u.fetch_type"), (sizeof("u.fetch_type")), src->u.fetch_type);
		
	
	

	

#endif
			
	
	/* DISPATCH(zend_uint,u.EA.type) */

	
	
	
	
		add_assoc_long_ex(dst, ("u.EA.type"), (sizeof("u.EA.type")), src->u.EA.type);
		
	
	

	

			break;
		
	}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->u);
	done_count ++;

#endif



		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dasm_zend_op */
	
		/* export:  	  	 	void 	xc_dasm_zend_op( 		 		 		 		 		zval *dst, const zend_op * const src 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dasm_zend_op( 		 		 		 		 		zval *dst, const zend_op * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op*)NULL)->handler) + sizeof(((zend_op*)NULL)->result) + sizeof(((zend_op*)NULL)->op1) + sizeof(((zend_op*)NULL)->op2) + sizeof(((zend_op*)NULL)->extended_value) + sizeof(((zend_op*)NULL)->lineno) + sizeof(((zend_op*)NULL)->opcode)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uchar,opcode) */

	
	
	
	
		add_assoc_long_ex(dst, ("opcode"), (sizeof("opcode")), src->opcode);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcode);
	done_count ++;

#endif



	

	
	
	/* STRUCT(znode,result) */

	assert(sizeof(znode) == sizeof(src->result));
	
	
	
	/* STRUCT_P_EX(znode,dst->result,src->result,result,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->result)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_znode(
		
		
		
		
		zv, & src->result
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("result"), (sizeof("result")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->result);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op1) */

	assert(sizeof(znode) == sizeof(src->op1));
	
	
	
	/* STRUCT_P_EX(znode,dst->op1,src->op1,op1,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op1)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_znode(
		
		
		
		
		zv, & src->op1
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("op1"), (sizeof("op1")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op1);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op2) */

	assert(sizeof(znode) == sizeof(src->op2));
	
	
	
	/* STRUCT_P_EX(znode,dst->op2,src->op2,op2,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op2)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_znode(
		
		
		
		
		zv, & src->op2
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("op2"), (sizeof("op2")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op2);
	done_count ++;

#endif



	
	
	/* DISPATCH(ulong,extended_value) */

	
	
	
	
		add_assoc_long_ex(dst, ("extended_value"), (sizeof("extended_value")), src->extended_value);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->extended_value);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,lineno) */

	
	
	
	
		add_assoc_long_ex(dst, ("lineno"), (sizeof("lineno")), src->lineno);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->lineno);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	
	/* DISPATCH(opcode_handler_t,handler) */

	/* is copying enough? */
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handler);
	done_count ++;

#endif


	

#endif

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dasm_zend_op_array */
	
		/* export:  	  	 	void 	xc_dasm_zend_op_array( 		 		 		 		 		zval *dst, const zend_op_array * const src 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dasm_zend_op_array( 		 		 		 		 		zval *dst, const zend_op_array * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op_array*)NULL)->type) + sizeof(((zend_op_array*)NULL)->function_name) + sizeof(((zend_op_array*)NULL)->scope) + sizeof(((zend_op_array*)NULL)->fn_flags) + sizeof(((zend_op_array*)NULL)->prototype) + sizeof(((zend_op_array*)NULL)->num_args) + sizeof(((zend_op_array*)NULL)->required_num_args) + sizeof(((zend_op_array*)NULL)->arg_info) + sizeof(((zend_op_array*)NULL)->pass_rest_by_reference) + sizeof(((zend_op_array*)NULL)->return_reference) + sizeof(((zend_op_array*)NULL)->refcount) + sizeof(((zend_op_array*)NULL)->opcodes) + sizeof(((zend_op_array*)NULL)->last) + sizeof(((zend_op_array*)NULL)->size) + sizeof(((zend_op_array*)NULL)->vars) + sizeof(((zend_op_array*)NULL)->last_var) + sizeof(((zend_op_array*)NULL)->size_var) + sizeof(((zend_op_array*)NULL)->T) + sizeof(((zend_op_array*)NULL)->brk_cont_array) + sizeof(((zend_op_array*)NULL)->last_brk_cont) + sizeof(((zend_op_array*)NULL)->current_brk_cont) + sizeof(((zend_op_array*)NULL)->try_catch_array) + sizeof(((zend_op_array*)NULL)->last_try_catch) + sizeof(((zend_op_array*)NULL)->static_variables) + sizeof(((zend_op_array*)NULL)->start_op) + sizeof(((zend_op_array*)NULL)->backpatch_count) + sizeof(((zend_op_array*)NULL)->done_pass_two) + sizeof(((zend_op_array*)NULL)->uses_this) + sizeof(((zend_op_array*)NULL)->filename) + sizeof(((zend_op_array*)NULL)->line_start) + sizeof(((zend_op_array*)NULL)->line_end) + sizeof(((zend_op_array*)NULL)->doc_comment) + sizeof(((zend_op_array*)NULL)->doc_comment_len) + sizeof(((zend_op_array*)NULL)->reserved)  ), assert_count = 34;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	do {
		zend_uint ii;
	int i;

	/* Common elements */
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
		add_assoc_long_ex(dst, ("type"), (sizeof("type")), src->type);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING(,function_name) */
 
	
	/* PROC_ZSTRING_N(,function_name,,strlen) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->function_name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("function_name"), (sizeof("function_name")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("function_name"), (sizeof("function_name")), src->function_name, xc_zstrlen_uchar (src->function_name) + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("function_name"), (sizeof("function_name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("function_name"), (sizeof("function_name")), src->function_name, xc_zstrlen_char (src->function_name) + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
			add_assoc_null_ex(dst, ("function_name"), (sizeof("function_name")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("function_name"), (sizeof("function_name")), src->function_name, xc_zstrlen_char (src->function_name) + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,fn_flags) */

	
	
	
	
		add_assoc_long_ex(dst, ("fn_flags"), (sizeof("fn_flags")), src->fn_flags);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->fn_flags);
	done_count ++;

#endif



	

	


	if (src->arg_info) {
		
	

		
			zval *arr;
			ALLOC_INIT_ZVAL(arr);
			array_init(arr);
			for (ii = 0; ii < src->num_args; ii ++) {
				zval *zv;

				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_zend_arg_info (zv, &(src->arg_info[ii]) TSRMLS_CC);
				add_next_index_zval(arr, zv);
			}
			add_assoc_zval_ex(dst, ("arg_info"), (sizeof("arg_info")), arr);
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif


		
	}
	else {
		
	
	add_assoc_null_ex(dst, (""), (sizeof("")));
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,num_args) */

	
	
	
	
		add_assoc_long_ex(dst, ("num_args"), (sizeof("num_args")), src->num_args);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,required_num_args) */

	
	
	
	
		add_assoc_long_ex(dst, ("required_num_args"), (sizeof("required_num_args")), src->required_num_args);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_rest_by_reference) */

	
	
	
	
		add_assoc_bool_ex(dst, ("pass_rest_by_reference"), (sizeof("pass_rest_by_reference")), src->pass_rest_by_reference ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_rest_by_reference);
	done_count ++;

#endif



	

#else
	if (src->arg_types) {
		
	
	
	
	/* allocate */
	
	
	
	
	

		
		do {
			zend_uint ii;
			int i;
			zval *zv;
			ALLOC_INIT_ZVAL(zv);
			array_init(zv);
			for (i = 0; i < src->arg_types[0]; i ++) {
				add_next_index_long(zv, src->arg_types[i + 1]);
			}
			add_assoc_zval_ex(dst, ("arg_types"), (sizeof("arg_types")), zv);
		} while (0);
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif


	}
	else {
		do {
			/* empty array */
			zval *zv;
			ALLOC_INIT_ZVAL(zv);
			array_init(zv);
			add_assoc_zval_ex(dst, ("arg_types"), (sizeof("arg_types")), zv);
		} while (0);
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif


		
	}
#endif
	
	
	/* DISPATCH(unsigned char,return_reference) */

	
	
	
	
		add_assoc_long_ex(dst, ("return_reference"), (sizeof("return_reference")), src->return_reference);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	/* END of common elements */
#ifdef IS_UNICODE
	#endif

	
	
	/* STRUCT_P(zend_uint,refcount) */

	if (src->refcount) {
		
		
	
	/* STRUCT_P_EX(zend_uint,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(zend_uint) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_zend_uint(
		
		
		
		
		zv,  src->refcount
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("refcount"), (sizeof("refcount")), zv);
	} while (0);
	
	
	

	}
	else {
		
		
	add_assoc_null_ex(dst, ("refcount"), (sizeof("refcount")));
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	
	

	
	


	if (src->opcodes) {
		
	

		
			zval *arr;
			ALLOC_INIT_ZVAL(arr);
			array_init(arr);
			for (ii = 0; ii < src->last; ii ++) {
				zval *zv;

				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_zend_op (zv, &(src->opcodes[ii]) TSRMLS_CC);
				add_next_index_zval(arr, zv);
			}
			add_assoc_zval_ex(dst, ("opcodes"), (sizeof("opcodes")), arr);
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif


		
	}
	else {
		
	
	add_assoc_null_ex(dst, (""), (sizeof("")));
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif



		
	}



	
	
	
	/* DISPATCH(zend_uint,last) */

	
	
	
	
		add_assoc_long_ex(dst, ("last"), (sizeof("last")), src->last);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,size) */

	
	
	
	
		add_assoc_long_ex(dst, ("size"), (sizeof("size")), src->size);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



	


#ifdef IS_CV
	
	if (src->vars) {
		
	

		
			zval *arr;
			ALLOC_INIT_ZVAL(arr);
			array_init(arr);
			for (i = 0; i < src->last_var; i ++) {
				zval *zv;

				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_zend_compiled_variable (zv, &(src->vars[i]) TSRMLS_CC);
				add_next_index_zval(arr, zv);
			}
			add_assoc_zval_ex(dst, ("vars"), (sizeof("vars")), arr);
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif


		
	}
	else {
		
	
	add_assoc_null_ex(dst, (""), (sizeof("")));
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_var) */

	
	
	
	
		add_assoc_long_ex(dst, ("last_var"), (sizeof("last_var")), src->last_var);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_var);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,size_var) */

	
	
	
	
		add_assoc_long_ex(dst, ("size_var"), (sizeof("size_var")), src->size_var);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size_var);
	done_count ++;

#endif



	

#else
		
		
		
#endif

	
	
	/* DISPATCH(zend_uint,T) */

	
	
	
	
		add_assoc_long_ex(dst, ("T"), (sizeof("T")), src->T);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->T);
	done_count ++;

#endif



	


	


	if (src->brk_cont_array) {
		
	

		
			zval *arr;
			ALLOC_INIT_ZVAL(arr);
			array_init(arr);
			for (ii = 0; ii < src->last_brk_cont; ii ++) {
				zval *zv;

				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_zend_brk_cont_element (zv, &(src->brk_cont_array[ii]) TSRMLS_CC);
				add_next_index_zval(arr, zv);
			}
			add_assoc_zval_ex(dst, ("brk_cont_array"), (sizeof("brk_cont_array")), arr);
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	add_assoc_null_ex(dst, (""), (sizeof("")));
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,last_brk_cont) */

	
	
	
	
		add_assoc_long_ex(dst, ("last_brk_cont"), (sizeof("last_brk_cont")), src->last_brk_cont);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_brk_cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,current_brk_cont) */

	
	
	
	
		add_assoc_long_ex(dst, ("current_brk_cont"), (sizeof("current_brk_cont")), src->current_brk_cont);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->current_brk_cont);
	done_count ++;

#endif



	

#ifndef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_globals) */

	
	
	
	
		add_assoc_bool_ex(dst, ("uses_globals"), (sizeof("uses_globals")), src->uses_globals ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_globals);
	done_count ++;

#endif



	

#endif

#ifdef ZEND_ENGINE_2
	
	if (src->try_catch_array) {
		
	

		
			zval *arr;
			ALLOC_INIT_ZVAL(arr);
			array_init(arr);
			for (i = 0; i < src->last_try_catch; i ++) {
				zval *zv;

				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_zend_try_catch_element (zv, &(src->try_catch_array[i]) TSRMLS_CC);
				add_next_index_zval(arr, zv);
			}
			add_assoc_zval_ex(dst, ("try_catch_array"), (sizeof("try_catch_array")), arr);
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	add_assoc_null_ex(dst, (""), (sizeof("")));
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_try_catch) */

	
	
	
	
		add_assoc_long_ex(dst, ("last_try_catch"), (sizeof("last_try_catch")), src->last_try_catch);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_try_catch);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT_P(HashTable,static_variables,HashTable_zval_ptr) */

	if (src->static_variables) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->static_variables,src->static_variables,static_variables,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_variables)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_HashTable_zval_ptr(
		
		
		
		
		zv,  src->static_variables
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("static_variables"), (sizeof("static_variables")), zv);
	} while (0);
	
	
	

	}
	else {
		
		
	add_assoc_null_ex(dst, ("static_variables"), (sizeof("static_variables")));
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_variables);
	done_count ++;

#endif




	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start_op);
	done_count ++;

#endif


	
	
	/* DISPATCH(int,backpatch_count) */

	
	
	
	
		add_assoc_long_ex(dst, ("backpatch_count"), (sizeof("backpatch_count")), src->backpatch_count);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->backpatch_count);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(zend_bool,done_pass_two) */

	
	
	
	
		add_assoc_bool_ex(dst, ("done_pass_two"), (sizeof("done_pass_two")), src->done_pass_two ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->done_pass_two);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_this) */

	
	
	
	
		add_assoc_bool_ex(dst, ("uses_this"), (sizeof("uses_this")), src->uses_this ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_this);
	done_count ++;

#endif



	

#endif

	
	/* PROC_STRING(filename) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->filename == NULL) {
		
		
			add_assoc_null_ex(dst, ("filename"), (sizeof("filename")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("filename"), (sizeof("filename")), src->filename, strlen(src->filename) + 1-1, 1);
				
	}
	
	
	
	
	
	

#ifdef IS_UNICODE
	
		
	/* PROC_STRING(script_encoding) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->script_encoding);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->script_encoding == NULL) {
		
		
			add_assoc_null_ex(dst, ("script_encoding"), (sizeof("script_encoding")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("script_encoding"), (sizeof("script_encoding")), src->script_encoding, strlen(src->script_encoding) + 1-1, 1);
				
	}
	
	
	
	
	
	

	
#endif
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
	
		add_assoc_long_ex(dst, ("line_start"), (sizeof("line_start")), src->line_start);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
	
		add_assoc_long_ex(dst, ("line_end"), (sizeof("line_end")), src->line_end);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
	
		add_assoc_long_ex(dst, ("doc_comment_len"), (sizeof("doc_comment_len")), src->doc_comment_len);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
			add_assoc_null_ex(dst, ("doc_comment"), (sizeof("doc_comment")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("doc_comment"), (sizeof("doc_comment")), src->doc_comment, src->doc_comment_len + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
			add_assoc_null_ex(dst, ("doc_comment"), (sizeof("doc_comment")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("doc_comment"), (sizeof("doc_comment")), src->doc_comment, src->doc_comment_len + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
			add_assoc_null_ex(dst, ("doc_comment"), (sizeof("doc_comment")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("doc_comment"), (sizeof("doc_comment")), src->doc_comment, src->doc_comment_len + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

#endif

	/* reserved */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->reserved);
	done_count ++;

#endif


#if defined(HARDENING_PATCH) && HARDENING_PATCH
	
	
	/* DISPATCH(zend_bool,created_by_eval) */

	
	
	
	
		add_assoc_bool_ex(dst, ("created_by_eval"), (sizeof("created_by_eval")), src->created_by_eval ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->created_by_eval);
	done_count ++;

#endif



	

#endif
	} while (0);

#ifdef ZEND_ENGINE_2
		
			
				
	
	add_assoc_null_ex(dst, (""), (sizeof("")));
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->prototype);
	done_count ++;

#endif



			
	
#endif

	
#ifdef ZEND_ENGINE_2
		
	if (src->scope) {
		
		
		add_assoc_stringl_ex(dst, ("scope"), (sizeof("scope")), src->scope->name, strlen(src->scope->name), 1);
	}
	else {
		
	add_assoc_null_ex(dst, ("scope"), (sizeof("scope")));
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->scope);
	done_count ++;

#endif


#endif
	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_zend_op_array : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_zend_op_array : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_dasm_xc_constinfo_t */
	
 	  	 	static void inline 	xc_dasm_xc_constinfo_t( 		 		 		 		 		zval *dst, const xc_constinfo_t * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_constinfo_t*)NULL)->key) + sizeof(((xc_constinfo_t*)NULL)->key_size) + sizeof(((xc_constinfo_t*)NULL)->constant)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
		add_assoc_long_ex(dst, ("key_size"), (sizeof("key_size")), src->key_size);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
		add_assoc_long_ex(dst, ("type"), (sizeof("type")), src->type);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("key"), (sizeof("key")), src->key, src->key_size-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("key"), (sizeof("key")), src->key, src->key_size-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("key"), (sizeof("key")), src->key, src->key_size-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

	
	
	
	/* STRUCT(zend_constant,constant) */

	assert(sizeof(zend_constant) == sizeof(src->constant));
	
	
	
	/* STRUCT_P_EX(zend_constant,dst->constant,src->constant,constant,,&) */

	
	

	
	assert(sizeof(zend_constant) == sizeof((& src->constant)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_zend_constant(
		
		
		
		
		zv, & src->constant
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("constant"), (sizeof("constant")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constant);
	done_count ++;

#endif




		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_dasm_xc_funcinfo_t */
	
		/* export:  	  	 	void 	xc_dasm_xc_funcinfo_t( 		 		 		 		 		zval *dst, const xc_funcinfo_t * const src 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dasm_xc_funcinfo_t( 		 		 		 		 		zval *dst, const xc_funcinfo_t * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_funcinfo_t*)NULL)->key) + sizeof(((xc_funcinfo_t*)NULL)->key_size) + sizeof(((xc_funcinfo_t*)NULL)->func)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
		add_assoc_long_ex(dst, ("key_size"), (sizeof("key_size")), src->key_size);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
		add_assoc_long_ex(dst, ("type"), (sizeof("type")), src->type);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("key"), (sizeof("key")), src->key, src->key_size-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("key"), (sizeof("key")), src->key, src->key_size-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("key"), (sizeof("key")), src->key, src->key_size-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

	
	
	
	/* STRUCT(zend_function,func) */

	assert(sizeof(zend_function) == sizeof(src->func));
	
	
	
	/* STRUCT_P_EX(zend_function,dst->func,src->func,func,,&) */

	
	

	
	assert(sizeof(zend_function) == sizeof((& src->func)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_zend_function(
		
		
		
		
		zv, & src->func
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("func"), (sizeof("func")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->func);
	done_count ++;

#endif




		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dasm_xc_classinfo_t */
	
		/* export:  	  	 	void 	xc_dasm_xc_classinfo_t( 		 		 		 		 		zval *dst, const xc_classinfo_t * const src 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dasm_xc_classinfo_t( 		 		 		 		 		zval *dst, const xc_classinfo_t * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_classinfo_t*)NULL)->key) + sizeof(((xc_classinfo_t*)NULL)->key_size) + sizeof(((xc_classinfo_t*)NULL)->cest) + sizeof(((xc_classinfo_t*)NULL)->oplineno)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
		add_assoc_long_ex(dst, ("key_size"), (sizeof("key_size")), src->key_size);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
		add_assoc_long_ex(dst, ("type"), (sizeof("type")), src->type);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("key"), (sizeof("key")), src->key, src->key_size-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("key"), (sizeof("key")), src->key, src->key_size-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("key"), (sizeof("key")), src->key, src->key_size-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

	
#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT_P(zend_class_entry,cest) */

	if (src->cest) {
		
		
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof(( src->cest)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_zend_class_entry(
		
		
		
		
		zv,  src->cest
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("cest"), (sizeof("cest")), zv);
	} while (0);
	
	
	

	}
	else {
		
		
	add_assoc_null_ex(dst, ("cest"), (sizeof("cest")));
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#else
	
	
	/* STRUCT(zend_class_entry,cest) */

	assert(sizeof(zend_class_entry) == sizeof(src->cest));
	
	
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,,&) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof((& src->cest)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_zend_class_entry(
		
		
		
		
		zv, & src->cest
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("cest"), (sizeof("cest")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(int,oplineno) */

	
	
	
	
		add_assoc_long_ex(dst, ("oplineno"), (sizeof("oplineno")), src->oplineno);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->oplineno);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2_1

	
	

/* {{{ xc_dasm_xc_autoglobal_t */
	
 	  	 	static void inline 	xc_dasm_xc_autoglobal_t( 		 		 		 		 		zval *dst, const xc_autoglobal_t * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_autoglobal_t*)NULL)->key) + sizeof(((xc_autoglobal_t*)NULL)->key_len)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_len) */

	
	
	
	
		add_assoc_long_ex(dst, ("key_len"), (sizeof("key_len")), src->key_len);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_len);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
		add_assoc_long_ex(dst, ("type"), (sizeof("type")), src->type);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	/* PROC_ZSTRING_L(type,key,key_len) */
 
	
	/* PROC_ZSTRING_N(type,key,key_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("key"), (sizeof("key")), src->key, src->key_len + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("key"), (sizeof("key")), src->key, src->key_len + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
			add_assoc_null_ex(dst, ("key"), (sizeof("key")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("key"), (sizeof("key")), src->key, src->key_len + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_dasm_xc_entry_data_php_t */
	
 	  	 	static void inline 	xc_dasm_xc_entry_data_php_t( 		 		 		 		 		zval *dst, const xc_entry_data_php_t * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_php_t*)NULL)->sourcesize) + sizeof(((xc_entry_data_php_t*)NULL)->device) + sizeof(((xc_entry_data_php_t*)NULL)->inode) + sizeof(((xc_entry_data_php_t*)NULL)->mtime) + sizeof(((xc_entry_data_php_t*)NULL)->op_array) + sizeof(((xc_entry_data_php_t*)NULL)->constinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->constinfos) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfos) + sizeof(((xc_entry_data_php_t*)NULL)->classinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->classinfos) + sizeof(((xc_entry_data_php_t*)NULL)->have_early_binding) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobal_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobals)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			zend_uint i;

#ifdef HAVE_INODE
	
	
	/* DISPATCH(int,device) */

	
	
	
	
		add_assoc_long_ex(dst, ("device"), (sizeof("device")), src->device);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->device);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,inode) */

	
	
	
	
		add_assoc_long_ex(dst, ("inode"), (sizeof("inode")), src->inode);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inode);
	done_count ++;

#endif



	

#endif
	
	
	/* DISPATCH(size_t,sourcesize) */

	
	
	
	
		add_assoc_long_ex(dst, ("sourcesize"), (sizeof("sourcesize")), src->sourcesize);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->sourcesize);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(time_t,mtime) */

	
	
	
	
		add_assoc_long_ex(dst, ("mtime"), (sizeof("mtime")), src->mtime);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->mtime);
	done_count ++;

#endif



	


	
	
	/* STRUCT_P(zend_op_array,op_array) */

	if (src->op_array) {
		
		
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof(( src->op_array)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_zend_op_array(
		
		
		
		
		zv,  src->op_array
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("op_array"), (sizeof("op_array")), zv);
	} while (0);
	
	
	

	}
	else {
		
		
	add_assoc_null_ex(dst, ("op_array"), (sizeof("op_array")));
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_array);
	done_count ++;

#endif




#ifdef HAVE_XCACHE_CONSTANT
	
	
	/* DISPATCH(zend_uint,constinfo_cnt) */

	
	
	
	
		add_assoc_long_ex(dst, ("constinfo_cnt"), (sizeof("constinfo_cnt")), src->constinfo_cnt);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->constinfos) {
		
	

		
			zval *arr;
			ALLOC_INIT_ZVAL(arr);
			array_init(arr);
			for (i = 0; i < src->constinfo_cnt; i ++) {
				zval *zv;

				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_xc_constinfo_t (zv, &(src->constinfos[i]) TSRMLS_CC);
				add_next_index_zval(arr, zv);
			}
			add_assoc_zval_ex(dst, ("constinfos"), (sizeof("constinfos")), arr);
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	add_assoc_null_ex(dst, (""), (sizeof("")));
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif



		
	}

#endif

	
	
	/* DISPATCH(zend_uint,funcinfo_cnt) */

	
	
	
	
		add_assoc_long_ex(dst, ("funcinfo_cnt"), (sizeof("funcinfo_cnt")), src->funcinfo_cnt);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->funcinfos) {
		
	

		
			zval *arr;
			ALLOC_INIT_ZVAL(arr);
			array_init(arr);
			for (i = 0; i < src->funcinfo_cnt; i ++) {
				zval *zv;

				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_xc_funcinfo_t (zv, &(src->funcinfos[i]) TSRMLS_CC);
				add_next_index_zval(arr, zv);
			}
			add_assoc_zval_ex(dst, ("funcinfos"), (sizeof("funcinfos")), arr);
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	add_assoc_null_ex(dst, (""), (sizeof("")));
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif



		
	}


	
	
	/* DISPATCH(zend_uint,classinfo_cnt) */

	
	
	
	
		add_assoc_long_ex(dst, ("classinfo_cnt"), (sizeof("classinfo_cnt")), src->classinfo_cnt);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfo_cnt);
	done_count ++;

#endif



	

	
	
	if (src->classinfos) {
		
	

		
			zval *arr;
			ALLOC_INIT_ZVAL(arr);
			array_init(arr);
			for (i = 0; i < src->classinfo_cnt; i ++) {
				zval *zv;

				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_xc_classinfo_t (zv, &(src->classinfos[i]) TSRMLS_CC);
				add_next_index_zval(arr, zv);
			}
			add_assoc_zval_ex(dst, ("classinfos"), (sizeof("classinfos")), arr);
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	add_assoc_null_ex(dst, (""), (sizeof("")));
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif



		
	}

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,autoglobal_cnt) */

	
	
	
	
		add_assoc_long_ex(dst, ("autoglobal_cnt"), (sizeof("autoglobal_cnt")), src->autoglobal_cnt);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobal_cnt);
	done_count ++;

#endif



	

	
	if (src->autoglobals) {
		
	

		
			zval *arr;
			ALLOC_INIT_ZVAL(arr);
			array_init(arr);
			for (i = 0; i < src->autoglobal_cnt; i ++) {
				zval *zv;

				ALLOC_INIT_ZVAL(zv);
				array_init(zv);
				xc_dasm_xc_autoglobal_t (zv, &(src->autoglobals[i]) TSRMLS_CC);
				add_next_index_zval(arr, zv);
			}
			add_assoc_zval_ex(dst, ("autoglobals"), (sizeof("autoglobals")), arr);
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif


		
	}
	else {
		
	
	add_assoc_null_ex(dst, (""), (sizeof("")));
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif



		
	}

#endif
	
	
	/* DISPATCH(zend_bool,have_early_binding) */

	
	
	
	
		add_assoc_bool_ex(dst, ("have_early_binding"), (sizeof("have_early_binding")), src->have_early_binding ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_early_binding);
	done_count ++;

#endif



	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dasm_xc_entry_data_var_t */
	
 	  	 	static void inline 	xc_dasm_xc_entry_data_var_t( 		 		 		 		 		zval *dst, const xc_entry_data_var_t * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_var_t*)NULL)->value)  ), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	
	/* STRUCT_P_EX(zval_ptr,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof((& src->value)[0]));
	
	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_zval_ptr(
		
		
		
		
		zv, & src->value
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("value"), (sizeof("value")), zv);
	} while (0);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_dasm_xc_entry_t */
	
		/* export:  	  	 	void 	xc_dasm_xc_entry_t( 		 		 		 		 		zval *dst, const xc_entry_t * const src 		 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_dasm_xc_entry_t( 		 		 		 		 		zval *dst, const xc_entry_t * const src 		 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_t*)NULL)->type) + sizeof(((xc_entry_t*)NULL)->hvalue) + sizeof(((xc_entry_t*)NULL)->next) + sizeof(((xc_entry_t*)NULL)->cache) + sizeof(((xc_entry_t*)NULL)->size) + sizeof(((xc_entry_t*)NULL)->refcount) + sizeof(((xc_entry_t*)NULL)->hits) + sizeof(((xc_entry_t*)NULL)->ctime) + sizeof(((xc_entry_t*)NULL)->atime) + sizeof(((xc_entry_t*)NULL)->dtime) + sizeof(((xc_entry_t*)NULL)->ttl) + sizeof(((xc_entry_t*)NULL)->name) + sizeof(((xc_entry_t*)NULL)->data) + sizeof(((xc_entry_t*)NULL)->have_references)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	
	/* DISPATCH(xc_entry_type_t,type) */

	
	
	
	
		add_assoc_long_ex(dst, ("type"), (sizeof("type")), src->type);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(size_t,size) */

	
	
	
	
		add_assoc_long_ex(dst, ("size"), (sizeof("size")), src->size);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(xc_hash_value_t,hvalue) */

	
	
	
	
		add_assoc_long_ex(dst, ("hvalue"), (sizeof("hvalue")), src->hvalue);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hvalue);
	done_count ++;

#endif



	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cache);
	done_count ++;

#endif


	/* skip */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->next);
	done_count ++;

#endif



	
	
	/* DISPATCH(long,refcount) */

	
	
	
	
		add_assoc_long_ex(dst, ("refcount"), (sizeof("refcount")), src->refcount);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(time_t,ctime) */

	
	
	
	
		add_assoc_long_ex(dst, ("ctime"), (sizeof("ctime")), src->ctime);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ctime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,atime) */

	
	
	
	
		add_assoc_long_ex(dst, ("atime"), (sizeof("atime")), src->atime);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->atime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,dtime) */

	
	
	
	
		add_assoc_long_ex(dst, ("dtime"), (sizeof("dtime")), src->dtime);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->dtime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(long,ttl) */

	
	
	
	
		add_assoc_long_ex(dst, ("ttl"), (sizeof("ttl")), src->ttl);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ttl);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_ulong,hits) */

	
	
	
	
		add_assoc_long_ex(dst, ("hits"), (sizeof("hits")), src->hits);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hits);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,name_type) */

	
	
	
	
		add_assoc_long_ex(dst, ("name_type"), (sizeof("name_type")), src->name_type);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_type);
	done_count ++;

#endif



	

#endif
		
	
	

#ifdef IS_UNICODE
		if (src->name_type == IS_UNICODE) {
			
	
	/* DISPATCH(int32_t,name.ustr.len) */

	
	
	
	
		add_assoc_long_ex(dst, ("name.ustr.len"), (sizeof("name.ustr.len")), src->name.ustr.len);
		
	
	

	

		}
		else {
			
	
	/* DISPATCH(int,name.str.len) */

	
	
	
	
		add_assoc_long_ex(dst, ("name.str.len"), (sizeof("name.str.len")), src->name.str.len);
		
	
	

	

		}
#else
		
	
	/* DISPATCH(int,name.str.len) */

	
	
	
	
		add_assoc_long_ex(dst, ("name.str.len"), (sizeof("name.str.len")), src->name.str.len);
		
	
	

	

#endif
		
#ifdef IS_UNICODE
			
	/* PROC_ZSTRING_L(name_type,name.uni.val,name.uni.len) */
 
	
	/* PROC_ZSTRING_N(name_type,name.uni.val,name.uni.len,len) */

#ifdef IS_UNICODE
	
	
	
		if (src->name_type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name.uni.val) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name.uni.val"), (sizeof("name.uni.val")));
		
	}
	else {
		
		
		
		
		
		
				
					add_assoc_unicodel_ex(dst, ("name.uni.val"), (sizeof("name.uni.val")), src->name.uni.val, src->name.uni.len + 1-1, 1);
					
				
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name.uni.val) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name.uni.val"), (sizeof("name.uni.val")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name.uni.val"), (sizeof("name.uni.val")), src->name.uni.val, src->name.uni.len + 1-1, 1);
				
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name.uni.val) == NULL) {
		
		
			add_assoc_null_ex(dst, ("name.uni.val"), (sizeof("name.uni.val")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name.uni.val"), (sizeof("name.uni.val")), src->name.uni.val, src->name.uni.len + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
	

#else
			
	/* PROC_STRING_L(name.str.val,name.str.len) */
 
	/* PROC_STRING_N(name.str.val,name.str.len + 1) */
 
	
	
	
	
	
	
	
	if (src->name.str.val == NULL) {
		
		
			add_assoc_null_ex(dst, ("name.str.val"), (sizeof("name.str.val")));
		
	}
	else {
		
		
		
		
		
		
				 					add_assoc_stringl_ex(dst, ("name.str.val"), (sizeof("name.str.val")), src->name.str.val, src->name.str.len + 1-1, 1);
				
	}
	
	
	
	
	
	

#endif
		
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		
	
	

		switch (src->type) {
		case XC_TYPE_PHP:
			
	
	/* STRUCT_P(xc_entry_data_php_t,data.php) */

	if (src->data.php) {
		
		
	
	/* STRUCT_P_EX(xc_entry_data_php_t,dst->data.php,src->data.php,data.php,) */

	
	

	
	assert(sizeof(xc_entry_data_php_t) == sizeof(( src->data.php)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_xc_entry_data_php_t(
		
		
		
		
		zv,  src->data.php
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("data.php"), (sizeof("data.php")), zv);
	} while (0);
	
	
	

	}
	else {
		
		
	add_assoc_null_ex(dst, ("data.php"), (sizeof("data.php")));
	

	}
	

			break;
		case XC_TYPE_VAR:
			
	
	/* STRUCT_P(xc_entry_data_var_t,data.var) */

	if (src->data.var) {
		
		
	
	/* STRUCT_P_EX(xc_entry_data_var_t,dst->data.var,src->data.var,data.var,) */

	
	

	
	assert(sizeof(xc_entry_data_var_t) == sizeof(( src->data.var)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	do {
		zval *zv;
		ALLOC_INIT_ZVAL(zv);
		array_init(zv);
	
	xc_dasm_xc_entry_data_var_t(
		
		
		
		
		zv,  src->data.var
		
		TSRMLS_CC
	);
	
		add_assoc_zval_ex(dst, ("data.var"), (sizeof("data.var")), zv);
	} while (0);
	
	
	

	}
	else {
		
		
	add_assoc_null_ex(dst, ("data.var"), (sizeof("data.var")));
	

	}
	

			break;
		default:
			assert(0);
		}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->data);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,have_references) */

	
	
	
	
		add_assoc_bool_ex(dst, ("have_references"), (sizeof("have_references")), src->have_references ? 1 : 0);
		
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_references);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_dasm_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_dasm_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif /* HAVE_XCACHE_DISASSEMBLER */
#ifdef HAVE_XCACHE_ASSEMBLER
  /* {{{ Pre-declare */
 	  	 	void 	xc_asm_zval( 		 		 		 		 		 		zval *dst, const zval * const src 		TSRMLS_DC 	); 	
 	  	 	static void inline 	xc_asm_zval_ptr( 		 		 		 		 		 		zval_ptr *dst, const zval_ptr * const src 		TSRMLS_DC 	); 	
 	  	 	void 	xc_asm_zend_op_array( 		 		 		 		 		 		zend_op_array *dst, const zend_op_array * const src 		TSRMLS_DC 	); 	
 	  	 	void 	xc_asm_zend_class_entry( 		 		 		 		 		 		zend_class_entry *dst, const zend_class_entry * const src 		TSRMLS_DC 	); 	
#ifdef HAVE_XCACHE_CONSTANT
 	  	 	static void inline 	xc_asm_zend_constant( 		 		 		 		 		 		zend_constant *dst, const zend_constant * const src 		TSRMLS_DC 	); 	
#endif
 	  	 	void 	xc_asm_zend_function( 		 		 		 		 		 		zend_function *dst, const zend_function * const src 		TSRMLS_DC 	); 	
 	  	 	void 	xc_asm_xc_entry_t( 		 		 		 		 		 		xc_entry_t *dst, const xc_entry_t * const src 		TSRMLS_DC 	); 	
#ifdef ZEND_ENGINE_2
 	  	 	static void inline 	xc_asm_zend_property_info( 		 		 		 		 		 		zend_property_info *dst, const zend_property_info * const src 		TSRMLS_DC 	); 	
#endif
/* }}} */
#ifdef IS_CV

	
	

/* {{{ xc_asm_zend_compiled_variable */
	
 	  	 	static void inline 	xc_asm_zend_compiled_variable( 		 		 		 		 		 		zend_compiled_variable *dst, const zend_compiled_variable * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_compiled_variable*)NULL)->name) + sizeof(((zend_compiled_variable*)NULL)->name_len) + sizeof(((zend_compiled_variable*)NULL)->hash_value)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(int,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,hash_value) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hash_value);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zend_compiled_variable : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_asm_zend_uint */
	
 	  	 	static void inline 	xc_asm_zend_uint( 		 		 		 		 		 		zend_uint *dst, const zend_uint * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_uint), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zend_uint : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifndef ZEND_ENGINE_2

	
	

/* {{{ xc_asm_int */
	
 	  	 	static void inline 	xc_asm_int( 		 		 		 		 		 		int *dst, const int * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(int), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_int : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_asm_zend_try_catch_element */
	
 	  	 	static void inline 	xc_asm_zend_try_catch_element( 		 		 		 		 		 		zend_try_catch_element *dst, const zend_try_catch_element * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_try_catch_element*)NULL)->try_op) + sizeof(((zend_try_catch_element*)NULL)->catch_op)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(zend_uint,try_op) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_op);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,catch_op) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->catch_op);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zend_try_catch_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif /* ifdef ZEND_ENGINE_2 */

	
	

/* {{{ xc_asm_zend_brk_cont_element */
	
 	  	 	static void inline 	xc_asm_zend_brk_cont_element( 		 		 		 		 		 		zend_brk_cont_element *dst, const zend_brk_cont_element * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_brk_cont_element*)NULL)->start) + sizeof(((zend_brk_cont_element*)NULL)->cont) + sizeof(((zend_brk_cont_element*)NULL)->brk) + sizeof(((zend_brk_cont_element*)NULL)->parent)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
#ifdef ZEND_ENGINE_2_2
#ifndef IS_UNICODE
	
	
	/* DISPATCH(int,start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start);
	done_count ++;

#endif



	

#endif
#endif
	
	
	/* DISPATCH(int,cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,brk) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,parent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zend_brk_cont_element : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	
	

/* {{{ xc_asm_HashTable_zval_ptr */
	
 	  	 	static void inline 	xc_asm_HashTable_zval_ptr( 		 		 		 		 		 		HashTable *dst, const HashTable * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		dst->canary = zend_hash_canary; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			
			
			if (sizeof(void *) == sizeof(zval_ptr)) {
				
								
	
	/* STRUCT_P_EX(zval_ptr,pnew->pData,((zval_ptr*)b->pData),,, ) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof((  ((zval_ptr*)b->pData))[0]));
	
	
	xc_asm_zval_ptr(
		
		
		
		
		
		  pnew->pData,   ((zval_ptr*)b->pData)
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zval_ptr,pnew->pData,((zval_ptr*)b->pData),,) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof(( ((zval_ptr*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_zval_ptr(
		
		
		
		
		
		 pnew->pData,  ((zval_ptr*)b->pData)
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_HashTable_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	
	

/* {{{ xc_asm_HashTable_zend_constant */
	
 	  	 	static void inline 	xc_asm_HashTable_zend_constant( 		 		 		 		 		 		HashTable *dst, const HashTable * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		dst->canary = zend_hash_canary; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			
			
			if (sizeof(void *) == sizeof(zend_constant)) {
				
								
	
	/* STRUCT_P_EX(zend_constant,pnew->pData,((zend_constant*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_constant) == sizeof((  ((zend_constant*)b->pData))[0]));
	
	
	xc_asm_zend_constant(
		
		
		
		
		
		  pnew->pData,   ((zend_constant*)b->pData)
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_constant,pnew->pData,((zend_constant*)b->pData),,) */

	
	

	
	assert(sizeof(zend_constant) == sizeof(( ((zend_constant*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_zend_constant(
		
		
		
		
		
		 pnew->pData,  ((zend_constant*)b->pData)
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_HashTable_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	
	

/* {{{ xc_asm_HashTable_zend_function */
	
		/* export:  	  	 	void 	xc_asm_HashTable_zend_function( 		 		 		 		 		 		HashTable *dst, const HashTable * const src 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_asm_HashTable_zend_function( 		 		 		 		 		 		HashTable *dst, const HashTable * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		dst->canary = zend_hash_canary; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			
			
			if (sizeof(void *) == sizeof(zend_function)) {
				
								
	
	/* STRUCT_P_EX(zend_function,pnew->pData,((zend_function*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_function) == sizeof((  ((zend_function*)b->pData))[0]));
	
	
	xc_asm_zend_function(
		
		
		
		
		
		  pnew->pData,   ((zend_function*)b->pData)
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_function,pnew->pData,((zend_function*)b->pData),,) */

	
	

	
	assert(sizeof(zend_function) == sizeof(( ((zend_function*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_zend_function(
		
		
		
		
		
		 pnew->pData,  ((zend_function*)b->pData)
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_HashTable_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef ZEND_ENGINE_2

	
	
	

/* {{{ xc_asm_HashTable_zend_property_info */
	
 	  	 	static void inline 	xc_asm_HashTable_zend_property_info( 		 		 		 		 		 		HashTable *dst, const HashTable * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((HashTable*)NULL)->nTableSize) + sizeof(((HashTable*)NULL)->nTableMask) + sizeof(((HashTable*)NULL)->nNumOfElements) + sizeof(((HashTable*)NULL)->nNextFreeElement) + sizeof(((HashTable*)NULL)->pInternalPointer) + sizeof(((HashTable*)NULL)->pListHead) + sizeof(((HashTable*)NULL)->pListTail) + sizeof(((HashTable*)NULL)->arBuckets) + sizeof(((HashTable*)NULL)->pDestructor) + sizeof(((HashTable*)NULL)->persistent) + sizeof(((HashTable*)NULL)->nApplyCount) + sizeof(((HashTable*)NULL)->bApplyProtection)  ), assert_count = 12;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
		
	

				
				Bucket *b, *pnew = NULL, *prev = NULL;
		zend_bool first = 1;
				
		

#if defined(HARDENING_PATCH_HASH_PROTECT) && HARDENING_PATCH_HASH_PROTECT
		dst->canary = zend_hash_canary; 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->canary);
	done_count ++;

#endif


#endif
		
	
	/* DISPATCH(uint,nTableSize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableSize);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nTableMask) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nTableMask);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(uint,nNumOfElements) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNumOfElements);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(ulong,nNextFreeElement) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nNextFreeElement);
	done_count ++;

#endif



	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pInternalPointer);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListHead);
	done_count ++;

#endif


		
	
	
	
	/* allocate */
	
	
	
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arBuckets);
	done_count ++;

#endif


		
	
	


		for (b = src->pListHead; b != NULL; b = b->pListNext) {
			

			
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			
			
			if (sizeof(void *) == sizeof(zend_property_info)) {
				
								
	
	/* STRUCT_P_EX(zend_property_info,pnew->pData,((zend_property_info*)b->pData),,, ) */

	
	

	
	assert(sizeof(zend_property_info) == sizeof((  ((zend_property_info*)b->pData))[0]));
	
	
	xc_asm_zend_property_info(
		
		
		
		
		
		  pnew->pData,   ((zend_property_info*)b->pData)
		TSRMLS_CC
	);
	
	
	

			}
			else {
				
	
	/* STRUCT_P_EX(zend_property_info,pnew->pData,((zend_property_info*)b->pData),,) */

	
	

	
	assert(sizeof(zend_property_info) == sizeof(( ((zend_property_info*)b->pData))[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_zend_property_info(
		
		
		
		
		
		 pnew->pData,  ((zend_property_info*)b->pData)
		TSRMLS_CC
	);
	
	
	

				
			}

			if (first) {
				
				first = 0;
			}

			
			prev = pnew;
		}
		
	
	

		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pListTail);
	done_count ++;

#endif


		 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pDestructor);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,persistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->persistent);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
		
	
	/* DISPATCH(zend_bool,unicode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unicode);
	done_count ++;

#endif



	

#endif
		
	
	/* DISPATCH(unsigned char,nApplyCount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->nApplyCount);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_bool,bApplyProtection) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->bApplyProtection);
	done_count ++;

#endif



	

#if ZEND_DEBUG
		
	
	/* DISPATCH(int,inconsistent) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inconsistent);
	done_count ++;

#endif



	

#endif
				
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_HashTable_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif

	
	

/* {{{ xc_asm_zval */
	
		/* export:  	  	 	void 	xc_asm_zval( 		 		 		 		 		 		zval *dst, const zval * const src 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_asm_zval( 		 		 		 		 		 		zval *dst, const zval * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zval*)NULL)->value) + sizeof(((zval*)NULL)->refcount) + sizeof(((zval*)NULL)->type) + sizeof(((zval*)NULL)->is_ref)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
				/* Variable information */
		
	
	

		switch (src->type & ~IS_CONSTANT_INDEX) {
			case IS_LONG:
			case IS_RESOURCE:
			case IS_BOOL:
				
	
	/* DISPATCH(long,value.lval) */

	
	
	
	
	

	

				break;
			case IS_DOUBLE:
				
	
	/* DISPATCH(double,value.dval) */

	
	
	
	
	

	

				break;
			case IS_NULL:
				
				break;

			case IS_CONSTANT:
#ifdef IS_UNICODE
				if (UG(unicode)) {
					goto proc_unicode;
				}
#endif
			case IS_STRING:
#ifdef FLAG_IS_BC
			case FLAG_IS_BC:
#endif
				
	
	/* DISPATCH(int,value.str.len) */

	
	
	
	
	

	

				
	/* PROC_STRING_L(value.str.val,value.str.len) */
 
	/* PROC_STRING_N(value.str.val,value.str.len + 1) */
 
	
	
	
	
	
	
	
	if (src->value.str.val == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

				break;
#ifdef IS_UNICODE
			case IS_UNICODE:
proc_unicode:
				
	
	/* DISPATCH(int32_t,value.uni.len) */

	
	
	
	
	

	

				
	/* PROC_ZSTRING_L(1,value.uni.val,value.uni.len) */
 
	
	/* PROC_ZSTRING_N(1,value.uni.val,value.uni.len,len) */

#ifdef IS_UNICODE
	
	
	
	
	
	
	
	
	
	
	if (ZSTR_U(src->value.uni.val) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

	
#else
	
	
	
	
	
	
	
	
	
	if (ZSTR_S(src->value.uni.val) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

				break;
#endif

			case IS_ARRAY:
			case IS_CONSTANT_ARRAY:
				
	
	/* STRUCT_P(HashTable,value.ht,HashTable_zval_ptr) */

	if (src->value.ht) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->value.ht,src->value.ht,value.ht,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->value.ht)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_HashTable_zval_ptr(
		
		
		
		
		
		 dst->value.ht,  src->value.ht
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

				break;

			case IS_OBJECT:
				
				#ifndef ZEND_ENGINE_2
				
	
	/* STRUCT_P(zend_class_entry,value.obj.ce) */

	if (src->value.obj.ce) {
		
		
	
	/* STRUCT_P_EX(zend_class_entry,dst->value.obj.ce,src->value.obj.ce,value.obj.ce,) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof(( src->value.obj.ce)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_zend_class_entry(
		
		
		
		
		
		 dst->value.obj.ce,  src->value.obj.ce
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

				
	
	/* STRUCT_P(HashTable,value.obj.properties,HashTable_zval_ptr) */

	if (src->value.obj.properties) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->value.obj.properties,src->value.obj.properties,value.obj.properties,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->value.obj.properties)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_HashTable_zval_ptr(
		
		
		
		
		
		 dst->value.obj.properties,  src->value.obj.properties
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

#endif
				break;

			default:
				assert(0);
		}
		
	
	

		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif


		
	
	/* DISPATCH(zval_data_type,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_uchar,is_ref) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->is_ref);
	done_count ++;

#endif



	

		
	
	/* DISPATCH(zend_ushort,refcount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_3
		
	
	/* DISPATCH(zend_uchar,idx_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->idx_type);
	done_count ++;

#endif



	

#endif
	
		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zval : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_asm_zval_ptr */
	
 	  	 	static void inline 	xc_asm_zval_ptr( 		 		 		 		 		 		zval_ptr *dst, const zval_ptr * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zval_ptr), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
		do {
			
			
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			
			
			
	
	/* STRUCT_P_EX(zval,dst[0],src[0],[0],, ) */

	
	

	
	assert(sizeof(zval) == sizeof((  src[0])[0]));
	
	
	xc_asm_zval(
		
		
		
		
		
		  dst[0],   src[0]
		TSRMLS_CC
	);
	
	
	

			
		} while (0);
	
	
#ifndef NDEBUG
		
	done_size += sizeof(zval_ptr);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zval_ptr : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_asm_zend_arg_info */
	
 	  	 	static void inline 	xc_asm_zend_arg_info( 		 		 		 		 		 		zend_arg_info *dst, const zend_arg_info * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_arg_info*)NULL)->name) + sizeof(((zend_arg_info*)NULL)->name_len) + sizeof(((zend_arg_info*)NULL)->class_name) + sizeof(((zend_arg_info*)NULL)->class_name_len) + sizeof(((zend_arg_info*)NULL)->array_type_hint) + sizeof(((zend_arg_info*)NULL)->allow_null) + sizeof(((zend_arg_info*)NULL)->pass_by_reference) + sizeof(((zend_arg_info*)NULL)->return_reference) + sizeof(((zend_arg_info*)NULL)->required_num_args)  ), assert_count = 9;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(zend_uint,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_uint,class_name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,class_name,class_name_len) */
 
	
	/* PROC_ZSTRING_N(,class_name,class_name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->class_name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->class_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->class_name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(zend_bool,array_type_hint) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->array_type_hint);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,allow_null) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->allow_null);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_by_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_by_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,return_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,required_num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zend_arg_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif
#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_asm_zend_constant */
	
 	  	 	static void inline 	xc_asm_zend_constant( 		 		 		 		 		 		zend_constant *dst, const zend_constant * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_constant*)NULL)->value) + sizeof(((zend_constant*)NULL)->flags) + sizeof(((zend_constant*)NULL)->name) + sizeof(((zend_constant*)NULL)->name_len) + sizeof(((zend_constant*)NULL)->module_number)  ), assert_count = 5;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* STRUCT(zval,value) */

	assert(sizeof(zval) == sizeof(src->value));
	
	
	
	/* STRUCT_P_EX(zval,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->value)[0]));
	
	
	xc_asm_zval(
		
		
		
		
		
		& dst->value, & src->value
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



	
	
	/* DISPATCH(int,flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,name_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_len);
	done_count ++;

#endif



	

	
	
	
	/* PROC_ZSTRING_L(,name,name_len) */
 
	
	/* PROC_ZSTRING_N(,name,name_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	
	/* DISPATCH(int,module_number) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module_number);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zend_constant : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_asm_zend_function */
	
		/* export:  	  	 	void 	xc_asm_zend_function( 		 		 		 		 		 		zend_function *dst, const zend_function * const src 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_asm_zend_function( 		 		 		 		 		 		zend_function *dst, const zend_function * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = sizeof(zend_function), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	

	switch (src->type) {
	case ZEND_INTERNAL_FUNCTION:
	case ZEND_OVERLOADED_FUNCTION:
		
		break;

	case ZEND_USER_FUNCTION:
	case ZEND_EVAL_CODE:
		
		
	
	/* STRUCT(zend_op_array,op_array) */

	assert(sizeof(zend_op_array) == sizeof(src->op_array));
	
	
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,,&) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof((& src->op_array)[0]));
	
	
	xc_asm_zend_op_array(
		
		
		
		
		
		& dst->op_array, & src->op_array
		TSRMLS_CC
	);
	
	
	

	

		break;

	default:
		assert(0);
	}
	
	
	

	
#ifndef NDEBUG
		
	done_size += sizeof(src[0]);
	done_count ++;

#endif


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zend_function : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2

	
	

/* {{{ xc_asm_zend_property_info */
	
 	  	 	static void inline 	xc_asm_zend_property_info( 		 		 		 		 		 		zend_property_info *dst, const zend_property_info * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_property_info*)NULL)->flags) + sizeof(((zend_property_info*)NULL)->name) + sizeof(((zend_property_info*)NULL)->name_length) + sizeof(((zend_property_info*)NULL)->h) + sizeof(((zend_property_info*)NULL)->doc_comment) + sizeof(((zend_property_info*)NULL)->doc_comment_len) + sizeof(((zend_property_info*)NULL)->ce)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	/* DISPATCH(zend_uint,flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->flags);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,name_length) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	/* DISPATCH(ulong,h) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->h);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif
	#if defined(ZEND_ENGINE_2_2) && !defined(IS_UNICODE)
	
	if (src->ce) {
		
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce);
	done_count ++;

#endif


#endif

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zend_property_info : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_asm_zend_class_entry */
	
		/* export:  	  	 	void 	xc_asm_zend_class_entry( 		 		 		 		 		 		zend_class_entry *dst, const zend_class_entry * const src 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_asm_zend_class_entry( 		 		 		 		 		 		zend_class_entry *dst, const zend_class_entry * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_class_entry*)NULL)->type) + sizeof(((zend_class_entry*)NULL)->name) + sizeof(((zend_class_entry*)NULL)->name_length) + sizeof(((zend_class_entry*)NULL)->parent) + sizeof(((zend_class_entry*)NULL)->refcount) + sizeof(((zend_class_entry*)NULL)->constants_updated) + sizeof(((zend_class_entry*)NULL)->ce_flags) + sizeof(((zend_class_entry*)NULL)->function_table) + sizeof(((zend_class_entry*)NULL)->default_properties) + sizeof(((zend_class_entry*)NULL)->properties_info) + sizeof(((zend_class_entry*)NULL)->default_static_members) + sizeof(((zend_class_entry*)NULL)->static_members) + sizeof(((zend_class_entry*)NULL)->constants_table) + sizeof(((zend_class_entry*)NULL)->builtin_functions) + sizeof(((zend_class_entry*)NULL)->constructor) + sizeof(((zend_class_entry*)NULL)->destructor) + sizeof(((zend_class_entry*)NULL)->clone) + sizeof(((zend_class_entry*)NULL)->__get) + sizeof(((zend_class_entry*)NULL)->__set) + sizeof(((zend_class_entry*)NULL)->__unset) + sizeof(((zend_class_entry*)NULL)->__isset) + sizeof(((zend_class_entry*)NULL)->__call) + sizeof(((zend_class_entry*)NULL)->__tostring) + sizeof(((zend_class_entry*)NULL)->serialize_func) + sizeof(((zend_class_entry*)NULL)->unserialize_func) + sizeof(((zend_class_entry*)NULL)->iterator_funcs) + sizeof(((zend_class_entry*)NULL)->create_object) + sizeof(((zend_class_entry*)NULL)->get_iterator) + sizeof(((zend_class_entry*)NULL)->interface_gets_implemented) + sizeof(((zend_class_entry*)NULL)->serialize) + sizeof(((zend_class_entry*)NULL)->unserialize) + sizeof(((zend_class_entry*)NULL)->interfaces) + sizeof(((zend_class_entry*)NULL)->num_interfaces) + sizeof(((zend_class_entry*)NULL)->filename) + sizeof(((zend_class_entry*)NULL)->line_start) + sizeof(((zend_class_entry*)NULL)->line_end) + sizeof(((zend_class_entry*)NULL)->doc_comment) + sizeof(((zend_class_entry*)NULL)->doc_comment_len) + sizeof(((zend_class_entry*)NULL)->module)  ), assert_count = 39;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	
	/* DISPATCH(char,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,name_length) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_length);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,name,name_length) */
 
	
	/* PROC_ZSTRING_N(,name,name_length,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
		
	if (src->parent) {
		
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->parent);
	done_count ++;

#endif


	
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(int,refcount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	

#else
	
	
	/* STRUCT_P(int,refcount) */

	if (src->refcount) {
		
		
	
	/* STRUCT_P_EX(int,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(int) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_int(
		
		
		
		
		
		 dst->refcount,  src->refcount
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(zend_bool,constants_updated) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_updated);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,ce_flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ce_flags);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT(HashTable,default_properties,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_properties));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->default_properties,src->default_properties,default_properties,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_properties)[0]));
	
	
	xc_asm_HashTable_zval_ptr(
		
		
		
		
		
		& dst->default_properties, & src->default_properties
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_properties);
	done_count ++;

#endif



	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->builtin_functions);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT(HashTable,properties_info,HashTable_zend_property_info) */

	assert(sizeof(HashTable) == sizeof(src->properties_info));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->properties_info,src->properties_info,properties_info,HashTable_zend_property_info,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->properties_info)[0]));
	
	
	xc_asm_HashTable_zend_property_info(
		
		
		
		
		
		& dst->properties_info, & src->properties_info
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->properties_info);
	done_count ++;

#endif



#	ifdef ZEND_ENGINE_2_1
	
	
	/* STRUCT(HashTable,default_static_members,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->default_static_members));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->default_static_members,src->default_static_members,default_static_members,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->default_static_members)[0]));
	
	
	xc_asm_HashTable_zval_ptr(
		
		
		
		
		
		& dst->default_static_members, & src->default_static_members
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->default_static_members);
	done_count ++;

#endif



	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif


#	else
	
	
	/* STRUCT_P(HashTable,static_members,HashTable_zval_ptr) */

	if (src->static_members) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->static_members,src->static_members,static_members,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_members)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_HashTable_zval_ptr(
		
		
		
		
		
		 dst->static_members,  src->static_members
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_members);
	done_count ++;

#endif



#	endif
	
	
	/* STRUCT(HashTable,constants_table,HashTable_zval_ptr) */

	assert(sizeof(HashTable) == sizeof(src->constants_table));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->constants_table,src->constants_table,constants_table,HashTable_zval_ptr,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->constants_table)[0]));
	
	
	xc_asm_HashTable_zval_ptr(
		
		
		
		
		
		& dst->constants_table, & src->constants_table
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constants_table);
	done_count ++;

#endif




		
	
	
		
			
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interfaces);
	done_count ++;

#endif


		
	
	
	
	/* DISPATCH(zend_uint,num_interfaces) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_interfaces);
	done_count ++;

#endif



	


	
	/* PROC_STRING(filename) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->filename == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize_func);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->iterator_funcs);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->create_object);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_iterator);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->interface_gets_implemented);
	done_count ++;

#endif


#ifdef ZEND_ENGINE_2_3
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->get_static_method);
	done_count ++;

#endif


#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->serialize);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->unserialize);
	done_count ++;

#endif


	/* deal with it inside xc_fix_method */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->destructor);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->clone);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__set);
	done_count ++;

#endif


/* should be >5.1 */
#ifdef ZEND_ENGINE_2_1
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__unset);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__isset);
	done_count ++;

#endif


# if defined(ZEND_ENGINE_2_2) || PHP_MAJOR_VERSION >= 6
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__tostring);
	done_count ++;

#endif


# endif
#endif
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__call);
	done_count ++;

#endif


#ifdef ZEND_CALLSTATIC_FUNC_NAME
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->__callstatic);
	done_count ++;

#endif


#endif
	/* # NOT DONE */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->module);
	done_count ++;

#endif


#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_function_call);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_get);
	done_count ++;

#endif


	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handle_property_set);
	done_count ++;

#endif


#endif
		
	
	/* STRUCT(HashTable,function_table,HashTable_zend_function) */

	assert(sizeof(HashTable) == sizeof(src->function_table));
	
	
	
	/* STRUCT_P_EX(HashTable,dst->function_table,src->function_table,function_table,HashTable_zend_function,&) */

	
	

	
	assert(sizeof(HashTable) == sizeof((& src->function_table)[0]));
	
	
	xc_asm_HashTable_zend_function(
		
		
		
		
		
		& dst->function_table, & src->function_table
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_table);
	done_count ++;

#endif



	
	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zend_class_entry : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_asm_znode */
	
 	  	 	static void inline 	xc_asm_znode( 		 		 		 		 		 		znode *dst, const znode * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((znode*)NULL)->op_type) + sizeof(((znode*)NULL)->u)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(int,op_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_type);
	done_count ++;

#endif



	


#ifdef IS_CV
#	define XCACHE_IS_CV IS_CV
#else
/* compatible with zend optimizer */
#	define XCACHE_IS_CV 16
#endif
	assert(src->op_type == IS_CONST ||
		src->op_type == IS_VAR ||
		src->op_type == XCACHE_IS_CV ||
		src->op_type == IS_TMP_VAR ||
		src->op_type == IS_UNUSED);
#undef XCACHE_IS_CV
		
	
	

	switch (src->op_type) {
		case IS_CONST:
			
	
	/* STRUCT(zval,u.constant) */

	assert(sizeof(zval) == sizeof(src->u.constant));
	
	
	
	/* STRUCT_P_EX(zval,dst->u.constant,src->u.constant,u.constant,,&) */

	
	

	
	assert(sizeof(zval) == sizeof((& src->u.constant)[0]));
	
	
	xc_asm_zval(
		
		
		
		
		
		& dst->u.constant, & src->u.constant
		TSRMLS_CC
	);
	
	
	

	

			break;
		
		case IS_VAR:
		case IS_TMP_VAR:
#ifdef IS_CV
		case IS_CV:
#else
		case 16:
#endif
			
	
	/* DISPATCH(zend_uint,u.var) */

	
	
	
	
	

	

			
	
	/* DISPATCH(zend_uint,u.EA.type) */

	
	
	
	
	

	

			break;
		case IS_UNUSED:
			
			
	
	/* DISPATCH(zend_uint,u.opline_num) */

	
	
	
	
	

	

#ifndef ZEND_ENGINE_2
			
	
	/* DISPATCH(zend_uint,u.fetch_type) */

	
	
	
	
	

	

#endif
			
	
	/* DISPATCH(zend_uint,u.EA.type) */

	
	
	
	
	

	

			break;
		
	}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->u);
	done_count ++;

#endif



		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_znode : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_asm_zend_op */
	
		/* export:  	  	 	void 	xc_asm_zend_op( 		 		 		 		 		 		zend_op *dst, const zend_op * const src 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_asm_zend_op( 		 		 		 		 		 		zend_op *dst, const zend_op * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op*)NULL)->handler) + sizeof(((zend_op*)NULL)->result) + sizeof(((zend_op*)NULL)->op1) + sizeof(((zend_op*)NULL)->op2) + sizeof(((zend_op*)NULL)->extended_value) + sizeof(((zend_op*)NULL)->lineno) + sizeof(((zend_op*)NULL)->opcode)  ), assert_count = 7;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uchar,opcode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcode);
	done_count ++;

#endif



	

	
	
	/* STRUCT(znode,result) */

	assert(sizeof(znode) == sizeof(src->result));
	
	
	
	/* STRUCT_P_EX(znode,dst->result,src->result,result,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->result)[0]));
	
	
	xc_asm_znode(
		
		
		
		
		
		& dst->result, & src->result
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->result);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op1) */

	assert(sizeof(znode) == sizeof(src->op1));
	
	
	
	/* STRUCT_P_EX(znode,dst->op1,src->op1,op1,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op1)[0]));
	
	
	xc_asm_znode(
		
		
		
		
		
		& dst->op1, & src->op1
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op1);
	done_count ++;

#endif



	
	
	/* STRUCT(znode,op2) */

	assert(sizeof(znode) == sizeof(src->op2));
	
	
	
	/* STRUCT_P_EX(znode,dst->op2,src->op2,op2,,&) */

	
	

	
	assert(sizeof(znode) == sizeof((& src->op2)[0]));
	
	
	xc_asm_znode(
		
		
		
		
		
		& dst->op2, & src->op2
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op2);
	done_count ++;

#endif



	
	
	/* DISPATCH(ulong,extended_value) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->extended_value);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(uint,lineno) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->lineno);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2_1
	
	
	
	/* DISPATCH(opcode_handler_t,handler) */

	/* is copying enough? */
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->handler);
	done_count ++;

#endif


	

#endif

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zend_op : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_asm_zend_op_array */
	
		/* export:  	  	 	void 	xc_asm_zend_op_array( 		 		 		 		 		 		zend_op_array *dst, const zend_op_array * const src 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_asm_zend_op_array( 		 		 		 		 		 		zend_op_array *dst, const zend_op_array * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((zend_op_array*)NULL)->type) + sizeof(((zend_op_array*)NULL)->function_name) + sizeof(((zend_op_array*)NULL)->scope) + sizeof(((zend_op_array*)NULL)->fn_flags) + sizeof(((zend_op_array*)NULL)->prototype) + sizeof(((zend_op_array*)NULL)->num_args) + sizeof(((zend_op_array*)NULL)->required_num_args) + sizeof(((zend_op_array*)NULL)->arg_info) + sizeof(((zend_op_array*)NULL)->pass_rest_by_reference) + sizeof(((zend_op_array*)NULL)->return_reference) + sizeof(((zend_op_array*)NULL)->refcount) + sizeof(((zend_op_array*)NULL)->opcodes) + sizeof(((zend_op_array*)NULL)->last) + sizeof(((zend_op_array*)NULL)->size) + sizeof(((zend_op_array*)NULL)->vars) + sizeof(((zend_op_array*)NULL)->last_var) + sizeof(((zend_op_array*)NULL)->size_var) + sizeof(((zend_op_array*)NULL)->T) + sizeof(((zend_op_array*)NULL)->brk_cont_array) + sizeof(((zend_op_array*)NULL)->last_brk_cont) + sizeof(((zend_op_array*)NULL)->current_brk_cont) + sizeof(((zend_op_array*)NULL)->try_catch_array) + sizeof(((zend_op_array*)NULL)->last_try_catch) + sizeof(((zend_op_array*)NULL)->static_variables) + sizeof(((zend_op_array*)NULL)->start_op) + sizeof(((zend_op_array*)NULL)->backpatch_count) + sizeof(((zend_op_array*)NULL)->done_pass_two) + sizeof(((zend_op_array*)NULL)->uses_this) + sizeof(((zend_op_array*)NULL)->filename) + sizeof(((zend_op_array*)NULL)->line_start) + sizeof(((zend_op_array*)NULL)->line_end) + sizeof(((zend_op_array*)NULL)->doc_comment) + sizeof(((zend_op_array*)NULL)->doc_comment_len) + sizeof(((zend_op_array*)NULL)->reserved)  ), assert_count = 34;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	do {
		zend_uint ii;
	int i;

	/* Common elements */
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING(,function_name) */
 
	
	/* PROC_ZSTRING_N(,function_name,,strlen) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->function_name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->function_name);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->function_name) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,fn_flags) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->fn_flags);
	done_count ++;

#endif



	

	


	if (src->arg_info) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (ii = 0; ii < src->num_args; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_arg_info,arg_info[i],) */

	assert(sizeof(zend_arg_info) == sizeof(src->arg_info[ii]));
	
	
	
	/* STRUCT_P_EX(zend_arg_info,dst->arg_info[i],src->arg_info[i],arg_info[i],,&) */

	
	

	
	assert(sizeof(zend_arg_info) == sizeof((& src->arg_info[ii])[0]));
	
	
	xc_asm_zend_arg_info(
		
		
		
		
		
		& dst->arg_info[ii], & src->arg_info[ii]
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_info);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,required_num_args) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->required_num_args);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_bool,pass_rest_by_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->pass_rest_by_reference);
	done_count ++;

#endif



	

#else
	if (src->arg_types) {
		
	
	
	
	/* allocate */
	
	
	
	
	

		
		
		
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif


	}
	else {
		
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->arg_types);
	done_count ++;

#endif



		
	}
#endif
	
	
	/* DISPATCH(unsigned char,return_reference) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->return_reference);
	done_count ++;

#endif



	

	/* END of common elements */
#ifdef IS_UNICODE
	#endif

	
	
	/* STRUCT_P(zend_uint,refcount) */

	if (src->refcount) {
		
		
	
	/* STRUCT_P_EX(zend_uint,dst->refcount,src->refcount,refcount,) */

	
	

	
	assert(sizeof(zend_uint) == sizeof(( src->refcount)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_zend_uint(
		
		
		
		
		
		 dst->refcount,  src->refcount
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	
	

	
	


	if (src->opcodes) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (ii = 0; ii < src->last; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_op,opcodes[i],) */

	assert(sizeof(zend_op) == sizeof(src->opcodes[ii]));
	
	
	
	/* STRUCT_P_EX(zend_op,dst->opcodes[i],src->opcodes[i],opcodes[i],,&) */

	
	

	
	assert(sizeof(zend_op) == sizeof((& src->opcodes[ii])[0]));
	
	
	xc_asm_zend_op(
		
		
		
		
		
		& dst->opcodes[ii], & src->opcodes[ii]
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->opcodes);
	done_count ++;

#endif



		
	}



	
	
	
	/* DISPATCH(zend_uint,last) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



	


#ifdef IS_CV
	
	if (src->vars) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->last_var; i ++) {
				
	
	

					
					
	
	/* STRUCT(zend_compiled_variable,vars[i],) */

	assert(sizeof(zend_compiled_variable) == sizeof(src->vars[i]));
	
	
	
	/* STRUCT_P_EX(zend_compiled_variable,dst->vars[i],src->vars[i],vars[i],,&) */

	
	

	
	assert(sizeof(zend_compiled_variable) == sizeof((& src->vars[i])[0]));
	
	
	xc_asm_zend_compiled_variable(
		
		
		
		
		
		& dst->vars[i], & src->vars[i]
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->vars);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_var) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_var);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,size_var) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size_var);
	done_count ++;

#endif



	

#else
		
#endif

	
	
	/* DISPATCH(zend_uint,T) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->T);
	done_count ++;

#endif



	


	


	if (src->brk_cont_array) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (ii = 0; ii < src->last_brk_cont; ii ++) {
				
	
	

					
					
	
	/* STRUCT(zend_brk_cont_element,brk_cont_array[i],) */

	assert(sizeof(zend_brk_cont_element) == sizeof(src->brk_cont_array[ii]));
	
	
	
	/* STRUCT_P_EX(zend_brk_cont_element,dst->brk_cont_array[i],src->brk_cont_array[i],brk_cont_array[i],,&) */

	
	

	
	assert(sizeof(zend_brk_cont_element) == sizeof((& src->brk_cont_array[ii])[0]));
	
	
	xc_asm_zend_brk_cont_element(
		
		
		
		
		
		& dst->brk_cont_array[ii], & src->brk_cont_array[ii]
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->brk_cont_array);
	done_count ++;

#endif



		
	}



	
	
	/* DISPATCH(zend_uint,last_brk_cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_brk_cont);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,current_brk_cont) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->current_brk_cont);
	done_count ++;

#endif



	

#ifndef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_globals) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_globals);
	done_count ++;

#endif



	

#endif

#ifdef ZEND_ENGINE_2
	
	if (src->try_catch_array) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->last_try_catch; i ++) {
				
	
	

					
					
	
	/* STRUCT(zend_try_catch_element,try_catch_array[i],) */

	assert(sizeof(zend_try_catch_element) == sizeof(src->try_catch_array[i]));
	
	
	
	/* STRUCT_P_EX(zend_try_catch_element,dst->try_catch_array[i],src->try_catch_array[i],try_catch_array[i],,&) */

	
	

	
	assert(sizeof(zend_try_catch_element) == sizeof((& src->try_catch_array[i])[0]));
	
	
	xc_asm_zend_try_catch_element(
		
		
		
		
		
		& dst->try_catch_array[i], & src->try_catch_array[i]
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->try_catch_array);
	done_count ++;

#endif



		
	}

	
	
	/* DISPATCH(int,last_try_catch) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->last_try_catch);
	done_count ++;

#endif



	

#endif

	
	
	/* STRUCT_P(HashTable,static_variables,HashTable_zval_ptr) */

	if (src->static_variables) {
		
		
	
	/* STRUCT_P_EX(HashTable,dst->static_variables,src->static_variables,static_variables,HashTable_zval_ptr) */

	
	

	
	assert(sizeof(HashTable) == sizeof(( src->static_variables)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_HashTable_zval_ptr(
		
		
		
		
		
		 dst->static_variables,  src->static_variables
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->static_variables);
	done_count ++;

#endif




	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->start_op);
	done_count ++;

#endif


	
	
	/* DISPATCH(int,backpatch_count) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->backpatch_count);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(zend_bool,done_pass_two) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->done_pass_two);
	done_count ++;

#endif



	

#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_bool,uses_this) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->uses_this);
	done_count ++;

#endif



	

#endif

	
	/* PROC_STRING(filename) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->filename);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->filename == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#ifdef IS_UNICODE
	
		
	/* PROC_STRING(script_encoding) */
 
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->script_encoding);
	done_count ++;

#endif


	
	
	
	
	
	
	
	if (src->script_encoding == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

	
#endif
#ifdef ZEND_ENGINE_2
	
	
	/* DISPATCH(zend_uint,line_start) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_start);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_uint,line_end) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->line_end);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,doc_comment_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment_len);
	done_count ++;

#endif



	

	
	/* PROC_ZSTRING_L(,doc_comment,doc_comment_len) */
 
	
	/* PROC_ZSTRING_N(,doc_comment,doc_comment_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
		if (UG(unicode)) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->doc_comment);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->doc_comment) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#endif

	/* reserved */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->reserved);
	done_count ++;

#endif


#if defined(HARDENING_PATCH) && HARDENING_PATCH
	
	
	/* DISPATCH(zend_bool,created_by_eval) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->created_by_eval);
	done_count ++;

#endif



	

#endif
	} while (0);

#ifdef ZEND_ENGINE_2
		
			
				
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->prototype);
	done_count ++;

#endif



			
	
#endif

	
#ifdef ZEND_ENGINE_2
		
	if (src->scope) {
		
		
		
	}
	else {
		
	
	

	}

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->scope);
	done_count ++;

#endif


#endif
	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_zend_op_array : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_zend_op_array : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#ifdef HAVE_XCACHE_CONSTANT

	
	

/* {{{ xc_asm_xc_constinfo_t */
	
 	  	 	static void inline 	xc_asm_xc_constinfo_t( 		 		 		 		 		 		xc_constinfo_t *dst, const xc_constinfo_t * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_constinfo_t*)NULL)->key) + sizeof(((xc_constinfo_t*)NULL)->key_size) + sizeof(((xc_constinfo_t*)NULL)->constant)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	/* STRUCT(zend_constant,constant) */

	assert(sizeof(zend_constant) == sizeof(src->constant));
	
	
	
	/* STRUCT_P_EX(zend_constant,dst->constant,src->constant,constant,,&) */

	
	

	
	assert(sizeof(zend_constant) == sizeof((& src->constant)[0]));
	
	
	xc_asm_zend_constant(
		
		
		
		
		
		& dst->constant, & src->constant
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constant);
	done_count ++;

#endif




		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_xc_constinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_asm_xc_funcinfo_t */
	
		/* export:  	  	 	void 	xc_asm_xc_funcinfo_t( 		 		 		 		 		 		xc_funcinfo_t *dst, const xc_funcinfo_t * const src 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_asm_xc_funcinfo_t( 		 		 		 		 		 		xc_funcinfo_t *dst, const xc_funcinfo_t * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_funcinfo_t*)NULL)->key) + sizeof(((xc_funcinfo_t*)NULL)->key_size) + sizeof(((xc_funcinfo_t*)NULL)->func)  ), assert_count = 3;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
	
	
	/* STRUCT(zend_function,func) */

	assert(sizeof(zend_function) == sizeof(src->func));
	
	
	
	/* STRUCT_P_EX(zend_function,dst->func,src->func,func,,&) */

	
	

	
	assert(sizeof(zend_function) == sizeof((& src->func)[0]));
	
	
	xc_asm_zend_function(
		
		
		
		
		
		& dst->func, & src->func
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->func);
	done_count ++;

#endif




		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_xc_funcinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_asm_xc_classinfo_t */
	
		/* export:  	  	 	void 	xc_asm_xc_classinfo_t( 		 		 		 		 		 		xc_classinfo_t *dst, const xc_classinfo_t * const src 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_asm_xc_classinfo_t( 		 		 		 		 		 		xc_classinfo_t *dst, const xc_classinfo_t * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_classinfo_t*)NULL)->key) + sizeof(((xc_classinfo_t*)NULL)->key_size) + sizeof(((xc_classinfo_t*)NULL)->cest) + sizeof(((xc_classinfo_t*)NULL)->oplineno)  ), assert_count = 4;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_size);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	
	/* PROC_ZSTRING_N(type,key,key_size) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	
#ifdef ZEND_ENGINE_2
	
	
	/* STRUCT_P(zend_class_entry,cest) */

	if (src->cest) {
		
		
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof(( src->cest)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_zend_class_entry(
		
		
		
		
		
		 dst->cest,  src->cest
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#else
	
	
	/* STRUCT(zend_class_entry,cest) */

	assert(sizeof(zend_class_entry) == sizeof(src->cest));
	
	
	
	/* STRUCT_P_EX(zend_class_entry,dst->cest,src->cest,cest,,&) */

	
	

	
	assert(sizeof(zend_class_entry) == sizeof((& src->cest)[0]));
	
	
	xc_asm_zend_class_entry(
		
		
		
		
		
		& dst->cest, & src->cest
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cest);
	done_count ++;

#endif



#endif
	
	
	/* DISPATCH(int,oplineno) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->oplineno);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_xc_classinfo_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#ifdef ZEND_ENGINE_2_1

	
	

/* {{{ xc_asm_xc_autoglobal_t */
	
 	  	 	static void inline 	xc_asm_xc_autoglobal_t( 		 		 		 		 		 		xc_autoglobal_t *dst, const xc_autoglobal_t * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_autoglobal_t*)NULL)->key) + sizeof(((xc_autoglobal_t*)NULL)->key_len)  ), assert_count = 2;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	/* DISPATCH(zend_uint,key_len) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key_len);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

#endif
	
		
	/* PROC_ZSTRING_L(type,key,key_len) */
 
	
	/* PROC_ZSTRING_N(type,key,key_len,len) */

#ifdef IS_UNICODE
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
		if (src->type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->key);
	done_count ++;

#endif


	
	
	
	
	
	
	
	
	if (ZSTR_S(src->key) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_xc_autoglobal_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	

#endif

	
	

/* {{{ xc_asm_xc_entry_data_php_t */
	
 	  	 	static void inline 	xc_asm_xc_entry_data_php_t( 		 		 		 		 		 		xc_entry_data_php_t *dst, const xc_entry_data_php_t * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_php_t*)NULL)->sourcesize) + sizeof(((xc_entry_data_php_t*)NULL)->device) + sizeof(((xc_entry_data_php_t*)NULL)->inode) + sizeof(((xc_entry_data_php_t*)NULL)->mtime) + sizeof(((xc_entry_data_php_t*)NULL)->op_array) + sizeof(((xc_entry_data_php_t*)NULL)->constinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->constinfos) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->funcinfos) + sizeof(((xc_entry_data_php_t*)NULL)->classinfo_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->classinfos) + sizeof(((xc_entry_data_php_t*)NULL)->have_early_binding) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobal_cnt) + sizeof(((xc_entry_data_php_t*)NULL)->autoglobals)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			zend_uint i;

#ifdef HAVE_INODE
	
	
	/* DISPATCH(int,device) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->device);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(int,inode) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->inode);
	done_count ++;

#endif



	

#endif
	
	
	/* DISPATCH(size_t,sourcesize) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->sourcesize);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(time_t,mtime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->mtime);
	done_count ++;

#endif



	


	
	
	/* STRUCT_P(zend_op_array,op_array) */

	if (src->op_array) {
		
		
	
	/* STRUCT_P_EX(zend_op_array,dst->op_array,src->op_array,op_array,) */

	
	

	
	assert(sizeof(zend_op_array) == sizeof(( src->op_array)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_zend_op_array(
		
		
		
		
		
		 dst->op_array,  src->op_array
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->op_array);
	done_count ++;

#endif




#ifdef HAVE_XCACHE_CONSTANT
	
	
	/* DISPATCH(zend_uint,constinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->constinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->constinfo_cnt; i ++) {
				
	
	

					
					
	
	/* STRUCT(xc_constinfo_t,constinfos[i],) */

	assert(sizeof(xc_constinfo_t) == sizeof(src->constinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_constinfo_t,dst->constinfos[i],src->constinfos[i],constinfos[i],,&) */

	
	

	
	assert(sizeof(xc_constinfo_t) == sizeof((& src->constinfos[i])[0]));
	
	
	xc_asm_xc_constinfo_t(
		
		
		
		
		
		& dst->constinfos[i], & src->constinfos[i]
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->constinfos);
	done_count ++;

#endif



		
	}

#endif

	
	
	/* DISPATCH(zend_uint,funcinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfo_cnt);
	done_count ++;

#endif



	

	
	if (src->funcinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->funcinfo_cnt; i ++) {
				
	
	

					
					
	
	/* STRUCT(xc_funcinfo_t,funcinfos[i],) */

	assert(sizeof(xc_funcinfo_t) == sizeof(src->funcinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_funcinfo_t,dst->funcinfos[i],src->funcinfos[i],funcinfos[i],,&) */

	
	

	
	assert(sizeof(xc_funcinfo_t) == sizeof((& src->funcinfos[i])[0]));
	
	
	xc_asm_xc_funcinfo_t(
		
		
		
		
		
		& dst->funcinfos[i], & src->funcinfos[i]
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->funcinfos);
	done_count ++;

#endif



		
	}


	
	
	/* DISPATCH(zend_uint,classinfo_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfo_cnt);
	done_count ++;

#endif



	

	
	
	if (src->classinfos) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->classinfo_cnt; i ++) {
				
	
	

					
		
	
					
	
	/* STRUCT(xc_classinfo_t,classinfos[i],) */

	assert(sizeof(xc_classinfo_t) == sizeof(src->classinfos[i]));
	
	
	
	/* STRUCT_P_EX(xc_classinfo_t,dst->classinfos[i],src->classinfos[i],classinfos[i],,&) */

	
	

	
	assert(sizeof(xc_classinfo_t) == sizeof((& src->classinfos[i])[0]));
	
	
	xc_asm_xc_classinfo_t(
		
		
		
		
		
		& dst->classinfos[i], & src->classinfos[i]
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->classinfos);
	done_count ++;

#endif



		
	}

#ifdef ZEND_ENGINE_2_1
	
	
	/* DISPATCH(zend_uint,autoglobal_cnt) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobal_cnt);
	done_count ++;

#endif



	

	
	if (src->autoglobals) {
		
	

		
			
	
	
	
	/* allocate */
	
	
	
	
	

			
			for (i = 0; i < src->autoglobal_cnt; i ++) {
				
	
	

					
		
	
					
	
	/* STRUCT(xc_autoglobal_t,autoglobals[i],) */

	assert(sizeof(xc_autoglobal_t) == sizeof(src->autoglobals[i]));
	
	
	
	/* STRUCT_P_EX(xc_autoglobal_t,dst->autoglobals[i],src->autoglobals[i],autoglobals[i],,&) */

	
	

	
	assert(sizeof(xc_autoglobal_t) == sizeof((& src->autoglobals[i])[0]));
	
	
	xc_asm_xc_autoglobal_t(
		
		
		
		
		
		& dst->autoglobals[i], & src->autoglobals[i]
		TSRMLS_CC
	);
	
	
	

	

				
	
	

			}
				
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif


		
	}
	else {
		
	
	
	

	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->autoglobals);
	done_count ++;

#endif



		
	}

#endif
	
	
	/* DISPATCH(zend_bool,have_early_binding) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_early_binding);
	done_count ++;

#endif



	

	

		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_xc_entry_data_php_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_asm_xc_entry_data_var_t */
	
 	  	 	static void inline 	xc_asm_xc_entry_data_var_t( 		 		 		 		 		 		xc_entry_data_var_t *dst, const xc_entry_data_var_t * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_data_var_t*)NULL)->value)  ), assert_count = 1;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
			
	
	
	/* STRUCT_P_EX(zval_ptr,dst->value,src->value,value,,&) */

	
	

	
	assert(sizeof(zval_ptr) == sizeof((& src->value)[0]));
	
	
	xc_asm_zval_ptr(
		
		
		
		
		
		& dst->value, & src->value
		TSRMLS_CC
	);
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->value);
	done_count ++;

#endif



		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_xc_entry_data_var_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


	
	

/* {{{ xc_asm_xc_entry_t */
	
		/* export:  	  	 	void 	xc_asm_xc_entry_t( 		 		 		 		 		 		xc_entry_t *dst, const xc_entry_t * const src 		TSRMLS_DC 	); 	 :export */
	
 	  	 	void 	xc_asm_xc_entry_t( 		 		 		 		 		 		xc_entry_t *dst, const xc_entry_t * const src 		TSRMLS_DC 	) 	
	{
		
		
		
#ifndef NDEBUG
		
			/* {{{ init assert */
			
			
						
				
			
			
				
			
			int assert_size = (  sizeof(((xc_entry_t*)NULL)->type) + sizeof(((xc_entry_t*)NULL)->hvalue) + sizeof(((xc_entry_t*)NULL)->next) + sizeof(((xc_entry_t*)NULL)->cache) + sizeof(((xc_entry_t*)NULL)->size) + sizeof(((xc_entry_t*)NULL)->refcount) + sizeof(((xc_entry_t*)NULL)->hits) + sizeof(((xc_entry_t*)NULL)->ctime) + sizeof(((xc_entry_t*)NULL)->atime) + sizeof(((xc_entry_t*)NULL)->dtime) + sizeof(((xc_entry_t*)NULL)->ttl) + sizeof(((xc_entry_t*)NULL)->name) + sizeof(((xc_entry_t*)NULL)->data) + sizeof(((xc_entry_t*)NULL)->have_references)  ), assert_count = 14;
			int done_size = 0, done_count = 0;
			/* }}} */
			
			
			do {
		
#endif


		

		
		
	
	
	
	/* DISPATCH(xc_entry_type_t,type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->type);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(size_t,size) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->size);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(xc_hash_value_t,hvalue) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hvalue);
	done_count ++;

#endif



	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->cache);
	done_count ++;

#endif


	/* skip */
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->next);
	done_count ++;

#endif



	
	
	/* DISPATCH(long,refcount) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->refcount);
	done_count ++;

#endif



	


	
	
	/* DISPATCH(time_t,ctime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ctime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,atime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->atime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(time_t,dtime) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->dtime);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(long,ttl) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->ttl);
	done_count ++;

#endif



	

	
	
	/* DISPATCH(zend_ulong,hits) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->hits);
	done_count ++;

#endif



	

#ifdef IS_UNICODE
	
	
	/* DISPATCH(zend_uchar,name_type) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name_type);
	done_count ++;

#endif



	

#endif
		
	
	

#ifdef IS_UNICODE
		if (src->name_type == IS_UNICODE) {
			
	
	/* DISPATCH(int32_t,name.ustr.len) */

	
	
	
	
	

	

		}
		else {
			
	
	/* DISPATCH(int,name.str.len) */

	
	
	
	
	

	

		}
#else
		
	
	/* DISPATCH(int,name.str.len) */

	
	
	
	
	

	

#endif
		
#ifdef IS_UNICODE
			
	/* PROC_ZSTRING_L(name_type,name.uni.val,name.uni.len) */
 
	
	/* PROC_ZSTRING_N(name_type,name.uni.val,name.uni.len,len) */

#ifdef IS_UNICODE
	
	
	
		if (src->name_type == IS_UNICODE) {
			
	
	
	
	
	
	
	
	if (ZSTR_U(src->name.uni.val) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
		else {
			
	
	
	
	
	
	
	
	if (ZSTR_S(src->name.uni.val) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

		}
	
#else
	
	
	
	
	
	
	
	
	
	if (ZSTR_S(src->name.uni.val) == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
	

#else
			
	/* PROC_STRING_L(name.str.val,name.str.len) */
 
	/* PROC_STRING_N(name.str.val,name.str.len + 1) */
 
	
	
	
	
	
	
	
	if (src->name.str.val == NULL) {
		
		
	}
	else {
		
		
		
		
		
		
	}
	
	
	
	
	
	

#endif
		
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->name);
	done_count ++;

#endif


	
		
	
	

		switch (src->type) {
		case XC_TYPE_PHP:
			
	
	/* STRUCT_P(xc_entry_data_php_t,data.php) */

	if (src->data.php) {
		
		
	
	/* STRUCT_P_EX(xc_entry_data_php_t,dst->data.php,src->data.php,data.php,) */

	
	

	
	assert(sizeof(xc_entry_data_php_t) == sizeof(( src->data.php)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_xc_entry_data_php_t(
		
		
		
		
		
		 dst->data.php,  src->data.php
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

			break;
		case XC_TYPE_VAR:
			
	
	/* STRUCT_P(xc_entry_data_var_t,data.var) */

	if (src->data.var) {
		
		
	
	/* STRUCT_P_EX(xc_entry_data_var_t,dst->data.var,src->data.var,data.var,) */

	
	

	
	assert(sizeof(xc_entry_data_var_t) == sizeof(( src->data.var)[0]));
	
	
	
	
	/* allocate */
	
	
	
	
	

	
	xc_asm_xc_entry_data_var_t(
		
		
		
		
		
		 dst->data.var,  src->data.var
		TSRMLS_CC
	);
	
	
	

	}
	else {
		
		
	
	

	}
	

			break;
		default:
			assert(0);
		}
	
	
	

	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->data);
	done_count ++;

#endif


		
	
	/* DISPATCH(zend_bool,have_references) */

	
	
	
	
	
	
	
#ifndef NDEBUG
		
	done_size += sizeof(src->have_references);
	done_count ++;

#endif



	


		
		
			
#ifndef NDEBUG
		
				/* {{{ check assert */
				if (done_count != assert_count) {
					fprintf(stderr
						, "count assertion failed at %s #%d xc_asm_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_count, assert_count, done_count - assert_count
						);
				}
				if (done_size != assert_size) {
					fprintf(stderr
						, "size assertion failed at %s #%d xc_asm_xc_entry_t : unexpected:%d - expecting:%d = %d != 0\n"
						, __FILE__, __LINE__
						, done_size, assert_size, done_size - assert_size
						);
				}
				if (done_count != assert_count || done_size != assert_size) {
					assert(0);
				}
				/* }}} */
			
#endif

			
				
				
				
			
		
		
		
#ifndef NDEBUG
		
			} while (0);
		
#endif

		
	}
/* }}} */
	


#endif /* HAVE_XCACHE_ASSEMBLER */


